#!/opt/perl/bin/perl
##############################################################################
# File: analyzer_elastic_json.pl
# Copyright 2014, Oracle and/or its affiliates. All rights reserved.
# $Id: analyzer_elastic_json.pl,v 1.0 2014/06/01 12:02:51 vivkumar Exp $
##############################################################################

use File::Find;
use File::Spec;
use HTML::TokeParser;
use Carp;
use Time::Local;
use Getopt::Long;
use Pod::Usage;
use HTML::Entities;
use LWP::UserAgent;
use XML::LibXML;
use IO::File;
use Pod::Usage;
use utf8;
use warnings;
use List::MoreUtils qw( any );
use English qw( -no_match_vars );
use Readonly;
use Try::Tiny;
use Pod::Usage;
use Cwd 'abs_path';
require Encode;
use Encode;

our $VERSION = 2014.06001;
my $folder                  = q();
my $library_partnumber      = q();
my $context                 = q();
my %library_data            = ();
my %library_data_docid      = ();
my %additional_data         = ();
my $indexcounter            = 0;
my $libraryurl              = 0;
my @html_files_from_library = ();
my %folderpnmapping         = ();

Readonly my $PARTNO => qr{ [[:alpha:]] \d{5} [-_] \d{1,2} }xms;

Readonly my $SHORTCUT_CHAR_LENGTH            => 3;
Readonly my $SHORTCUT_CHAR_LENGTH_FOUR       => 4;
Readonly my $SHORTCUT_CHAR_LENGTH_FIVE       => 5;
Readonly my $SHORTCUT_CHAR_LENGTH_SIX        => 6;
Readonly my $SHORTCUT_CHAR_LENGTH_SEVEN      => 7;
Readonly my $SHORTCUT_CHAR_LENGTH_EIGHT      => 8;
Readonly my $SHORTCUT_CHAR_LENGTH_NINE       => 9;
Readonly my $SHORTCUT_CHAR_LENGTH_TEN        => 10;
Readonly my $SHORTCUT_CHAR_LENGTH_SIXTY      => 60;
Readonly my $SHORTCUT_CHAR_LENGTH_FIFTY      => 50;
Readonly my $SHORTCUT_CHAR_LENGTH_TWOFIFTY   => 250;
Readonly my $SHORTCUT_CHAR_LENGTH_THIRTYFIVE => 35;
Readonly my $SHORTCUT_CHAR_LENGTH_NINENINE   => 99;
Readonly my $INITIAL                         => -1;
Readonly my $ADDER                           => 1900;
Readonly my $HUNDRED                         => 100;
Readonly my $ELEVEN                          => 11;
Readonly my $TWELVE                          => 12;
Readonly my $THIRTEEN                        => 13;
Readonly my $THREETWO                        => 32;
Readonly my $THREEEIGHT                      => 38;
Readonly my $THREEFOUR                       => 34;
Readonly my $THREENINE                       => 39;
Readonly my $SIXTWO                          => 62;
Readonly my $ONETWOSIX                       => 126;

my $server_elastic_search = q(stdoc.us.oracle.com);
my $pdb_url               = q(http://pdb.us.oracle.com/);
my %errors                = ();
my $whichpn               = "";
Getopt::Long::Configure('gnu_getopt');
GetOptions(
    'input|f=s' => \$folder,
    'lib|l=s'   => \$library_partnumber,
    'ctx|c=s'   => \$context,
    'url|u=s'   => \$libraryurl
);

$whichpn = $library_partnumber;
%subfolder_mappings = ();
&getFolderMappings ($whichpn);

my @administrator_pns = ("e49548", "e49573", "e50708", "e51676", "e51667", "e51666", "e50710", 
                         "e37261", "e40026", "e49574", "e49554", "e49599", "e49564", "e49601", 
                         "e49611", "e49613", "e49880", "e49622", "e49571", "e51077", "e49577", 
                         "e49595", "e50784", "e50989", "e50987", "e50988", "e50994", "e50990", 
                         "e50992", "e50993", "e51078", "e51153", "e39993", "e39998", "e40008", 
                         "e40009");

my @developer_pns     = ("e51622", "e38763", "e41172", "e51153", "e50784", 
                         "e37261", "e50335", "e50269");

my @user_pns          = ("e49617", "e41248", "e41354", "e41353", "e41352", "e49888",
                         "e49573", "e49574", "e49599", "e49564", "e49601", "e49611",
                         "e49613", "e49880", "e49622", "e49571", "e51077", "e49577",
                         "e49595", "e50784", "e49598", "e49596", "e49575", "e49527",
                         "e49576", "e49572", "e49562", "e49621", "e49887", "e49593",
                         "e49591", "e49592", "e49603", "e49547", "e49605", "e41049",
                         "e51098", "e49612", "e49610", "e49615", "e49618", "e49889",
                         "e49619", "e49620", "e49609", "e49570", "e49597", "e49594",
                         "e50655", "e51725", "e39964", "e40003", "e40005", "e40023",
                         "e40024", "e40007", "e39999", "e50656", "e39992", "e41356",
                         "e41357", "e41359", "e41358", "e50795", "e50713", "e51666",
                         "e49528", "e51213"
                         );

$context = lc $context;
if (length $libraryurl < 1) {
    print "Library URL was not provided\n";
}
else {
   $libraryurl =~ s/\/$//gis;
   $libraryurl = $libraryurl."/";
}

if ( length $folder < 1 || length $library_partnumber < 1 ) {
    print "----------------------------------------\n";
    print "Error: Incorrect Number of Arguments passed.\n";
    print "----------------------------------------\n";
    print "Usage:\n";
    print
        "perl analyzer_elastic_json.pl -f <FOLDER> -l <LIB PART NUMBER> --ctx <CONTEXT> -u <LIBRARY-URL>\n";
    print "----------------------------------------\n";
    exit;
}

my $snapshot_date = snapshot_time();

# take backup
# delete the index
# create a new index with the same name
# put mapping in a mapping file
# create the json data in a separate json file

my $elastic_index = <<"TOP";

curl -XPUT "${server_elastic_search}:9200/_snapshot/docsbackup/${context}_${snapshot_date}" -d '{
    "indices": "${context}",
    "ignore_unavailable": "true",
    "include_global_state": false
}'

VIVEK-SAMPLE

curl -XDELETE 'http://${server_elastic_search}:9200/${context}'

VIVEK-SAMPLE

curl -XPUT 'http://${server_elastic_search}:9200/${context}/'

VIVEK-SAMPLE

curl -XPUT "${server_elastic_search}:9200/${context}/_settings -d '{"index" : {"refresh_interval" : "-1", "number_of_replicas" : "0"}}'


TOP

my %index_type_mappings = ();

my @errorlines = ();

if ( !-d $folder ) {
    print "The Supplied folder path does not exist or is not a folder.\n";
    exit;
}
my @dir                  = ();
my @titlefiles           = ();
my %targets              = ();
my %duplicate_targets    = ();
my %images               = ();
my $time                 = time;
my $check_date           = localtime time;
my $count_chunks         = 0;
my $summary_count        = 0;
my $summary_not_written  = 0;
my @equivalence          = ();
my $stdocid              = q();
my $track                = 0;
my $ignore_track         = 0;
my @bar                  = ( '|', '/', '--', '|', '--', '\\', );
my $head                 = q();
my $body                 = q();
my @allids               = ();
my $lastidused           = ();
my $lastid_line_found    = q();
my @cleanedlines         = ();
my $chunk                = q();
my $wasadded             = q(0);
my $was_last_token_added = q(0);
my $total_targets        = 0;
my $total_images         = 0;
my $largeprint           = 0;
my $output_folder        = q(./output);
my $lastchunkwritten     = q();
my $isPSFT               = q();
my @headclasses
    = qw/msg glossaryentry glossaryterm glossterm titleinfigure titleinexample titleintable titleinequation itemizedlisttitle orderedlisttitle segmentedlisttitle variablelisttitle bridgehead titleinrefsubsect subhead1 subhead2 subhead3 titleinrefsubsect2 titleinrefsubsect3/;

my $chaptertitle      = q();
my $chunkclass        = q();
my $chunkheadlevel    = q();
my $last_id_found     = q();
my %booktitle         = ();
my %indexes           = ();
my %indexes_map       = ();
my %indexes_list      = ();
my $last_pn_processed = q();
my $parent            = q();
my $sect_found        = 0;
my %relation          = ();
my %topic_types_found = ();
my $file_release_date = q();
my $file_description  = q();
my $file_keywords     = q();
my $book_formats      = q();
$OUTPUT_AUTOFLUSH++;

find( { wanted => \&wanted, no_chdir => 1 }, $folder );
if ( scalar @dir < 1 ) {
    print "Error: no htm(l) file found.\n";
    exit;
}

if (-d $output_folder) {
    print "Output folder already present.";
    ##exit;
}

if ( !-d $output_folder ) {
    mkdir $output_folder;
}

my $jsonoutputfolder = $output_folder . q(/json);

if ( !-d $jsonoutputfolder ) {
    mkdir $jsonoutputfolder;
}

if ( -f $jsonoutputfolder . q(/) . $library_partnumber . q(.txt) ) {
    ##unlink $jsonoutputfolder . q(/) . $library_partnumber . q(.txt);
}

open my $CHUNK, '>>', $jsonoutputfolder . q(/) . $library_partnumber . q(-mapping.txt)
    or croak "Couldnt write to json output file: $OS_ERROR\n";
binmode $CHUNK, ':encoding(UTF-8)';
print {$CHUNK} $elastic_index . "\n\n";
print {$CHUNK} q(VIVEK-SAMPLE) . "\n\n";
close $CHUNK
    or carp "Cant close: $OS_ERROR";

build_data( $library_partnumber, $folder );

my $json_title_lib   = q();
my $json_release_lib = q();
if ( exists $library_data{$library_partnumber} ) {
    $json_title_lib = $library_data{$library_partnumber}[0];
    $json_release_lib
        = $library_data{$library_partnumber}[$SHORTCUT_CHAR_LENGTH_FIVE];
}
else { $json_title_lib = q(Not Available.); }
my $file_lang = q(en);
$isPSFT = q(0);
my $javadoc = q(0);
my $thisfilechunkcount = 0;
my $noid = q(0);

foreach my $file (@dir) {
    $noid = q(0);
    tidyhtml($file);
    $thisfilechunkcount = 0;
    $book_formats = q();
    $parent     = q();
    $sect_found = 0;
    %relation   = ();
    $file_lang = q(en);
    $file_release_date = q();
    $file_description  = q();
    $file_keywords     = q();
    $javadoc = q(0);

    $track++;
    my @file_content           = ();
    my $current_id             = q();
    my $last_id                = q();
    my $id_related_text        = q();
    my $equivalent_data_holder = q();
    $chaptertitle = q();
    @equivalence  = ();
    my $file_path = $file;
    $file_path =~ s/\\/\//gis;
    $file_path =~ s/(.*)\/(.*)/$1/gis;

    my $file_name_val = $file;
    $file_name_val =~ s/\\/\//gis;
    $file_name_val =~ s/(.*)\/(.*)/$2/gis;

    my $fname_processed = $file_path;
    if ( $fname_processed =~ m/(.*)\/(.*)/gis ) {
        $fname_processed = $2 . q(/) . $file_name_val;
    }

    # format Print
    ###########################################################################
   #my $log_print = "$track out of ".scalar (@dir)." files ".$bar[$track % 6];
    if ( $largeprint == 0 ) {
        $largeprint = length $fname_processed;
    }
    if ( $largeprint < length $fname_processed ) {
        $largeprint = length $fname_processed;
    }
    if ( length $fname_processed < $largeprint ) {
        my $space = q();
        for (
            my $spacer = 0;
            $spacer < (
                $SHORTCUT_CHAR_LENGTH_THIRTYFIVE - ( length $fname_processed )
            );
            $spacer++
            )
        {
            $space .= q( );
        }
        $fname_processed .= $space;
    }
    my $log_print = $track . q( out of ) . scalar @dir . q( files );
    printf qq[\r$log_print $fname_processed\t];
    ###########################################################################

    open my $FILE, '<', $file
        or croak "Error while reading $file: $OS_ERROR";
    #binmode $FILE, ':encoding(ISO-8859-1)';
    #binmode $FILE, ':encoding(UTF-8)';
    #binmode $FILE;
    @file_content = <$FILE>;
    close $FILE
        or carp "Can't close file: $OS_ERROR";

    # new way
    my $filejoined = join " ", @file_content;
    #print "Pass 012\n";

    if (   $filejoined =~ m{ <[?]xml [^>]+ encoding [[:space:]]* = [[:space:]]* ['"] utf -? 8 }xmsi
        || $filejoined =~ m{ <meta [^>]+ charset [[:space:]]* = [[:space:]]* utf -? 8 }xmsi )
    {
        $filejoined = Encode::decode( 'utf8', $filejoined );
    }

    #elsif (   $filejoined =~ m{ <[?]xml [^>]+ encoding [[:space:]]* = [[:space:]]* ['"] (.*?) ['"]}xmsi
    #    || $filejoined =~ m{ <meta [^>]+ charset [[:space:]]* = [[:space:]]* ['"] (.*?) ['"] }xmsi
    #) {
    #    my $charencodingfound = $1;
    #    $charencodingfound =~ s/^(\s)+//gis;
    #    $charencodingfound =~ s/(\s)+$//gis;
    #    $filejoined = decode( $1, $filejoined );
    #}
    #else {
    #    $filejoined = decode( 'utf8', $filejoined );
    #    # html/securityroles.htm
    #}

    if ( $filejoined =~ m/<title>(.*?)index(.*?)<\/title>/is) {
        next;
    }
    # exclude indexing if the file has noindex for robots
    if ( $filejoined =~ m/<meta(\s)*name(\s)*=(\s)*"robots"(\s)*content(\s)*=(\s)*"([^"]*)noindex([^"]*)"([^>]*)>/is) {
        $ignore_track++;
        next;
    }
    #print "Pass 01\n";
    # get the formats from the head elements
    if ($filejoined =~ m/<link(\s)*rel(\s)*=(\s)*"alternate"(\s)*href(\s)*=(\s)*"([^"]+)"(\s)*title="([^>]+)"(\s)*type(\s)*=(\s)*"application\/pdf"([^>]*)>/i) {
        $book_formats .= $7.";";
    }
    if ($filejoined =~ m/<link(\s)*rel(\s)*=(\s)*"alternate"(\s)*href(\s)*=(\s)*"([^"]+)"(\s)*title="([^>]+)"(\s)*type(\s)*=(\s)*"([^"]*)epub([^"]*)"([^>]*)>/i) {
        $book_formats .= $7.";";
    }
    if ($filejoined =~ m/<link(\s)*rel(\s)*=(\s)*"alternate"(\s)*href(\s)*=(\s)*"([^"]+)"(\s)*title="([^>]+)"(\s)*type(\s)*=(\s)*"([^"]*)mobi([^"]*)"([^>]*)>/i) {
        $book_formats .= $7.";";
    }
    #print "$book_formats\n";
    #print "Pass 013\n";

    ## First pass: Check if there are any name or id that can be used else we will need to get all the content and link it
    pos( $filejoined ) = 0;
    if ($filejoined !~ m/id(\s)*=(\s)*"([^"]+)"/gis && $filejoined !~ m/<a([^>]+)name(\s)*=(\s)*"([^"]+)"([^>]*)>/gis) {
        $filejoined =~ s/(<body([^>]*)>)/$1\n<a id="JUSTANID" \/>/is;
        $noid = q(1);
    }
    #print "Pass No id: $noid\n";

    ##if (length $book_formats > 0) {
    ##    my @ot_formats = split(/;/, $book_formats);
    ##    $book_formats = q();
    ##    my $bpath = $file;
    ##    $bpath =~ s/\\/\//gis;
    ##    $bpath =~ s/(.*)\/(.*)/$1/gis;
    ##    foreach my $fr (@ot_formats) {
    ##      if (length $fr > 0) {
    ##        #print "Add: $bpath, $fr";
    ##        $fr = calculate_path($bpath, $fr);
    ##        $fr =~ s/\\/\//gisxm;
    ##        if (length $fr > 0) {
    ##          $book_formats .= $fr.";";
    ##        }
    ##      }
    ##    }
    ##}
    
    pos( $filejoined ) = 0;
    #print "Pass 014\n";
    if ( $filejoined =~ m/<title>(.*?)(\s)API(.*?)<\/title>/is
        || $filejoined =~ m/<title>(.*?)xml\s*schema\s*(.*?)<\/title>/is
        || $filejoined =~ m/<title>(.*?)java(.*?)<\/title>/is
        || $filejoined =~ m/Generated(\s)by(\s)javadoc/is
    ) {
        $javadoc = q(1);
    }
    $filejoined = $filejoined;
    #print "Pass 014q\n";

    if ($javadoc eq q(1)) {
        #print "Pass 014-1\n";
        $filejoined =~ s/(<body([^>]*)>)(.*?)(<div(\s)*class(\s)*=(\s)*"header"([^>]*)>)/$1$4/is;
    }
    elsif ($filejoined !~ m/index[.]html$/) {
        #print "Pass 014-31\n";
        $filejoined =~ s/(<body([^>]*)>)(.*?)(<div([^>]*?)id(\s)*=(\s)*"header"([^>]*)>)/$1$4/is;
        #print "Pass 014-32\n";
        $filejoined =~ s/(<body([^>]*)>)(.*?)(<div([^>]*?)class(\s)*=(\s)*"header"([^>]*)>)/$1$4/is;
        #print "Pass 014-33\n";
        if ($filejoined =~ m/class(\s)*=(\s)*"ind"/is) {
                   $filejoined =~ s/(<div(\s)*class(\s)*=(\s)*"header"([^>]*)>)(.*?)(<div(\s)*class(\s)*=(\s)*"ind">)/$1$7/is;
        }
        #print "Pass 014-34\n";
    }
    #print "Pass 015\n";

    if ( $filejoined =~ m/lang="([^"]+)"/isxm ) {
        $file_lang = $1;
        if (lc $file_lang eq q(en-us)) {$file_lang = q(en);}
    }
    elsif ($filejoined =~ m/xml:lang="([^"]+)"/isxm) {
        $file_lang = $1;
        if (lc $file_lang eq q(en-us)) {$file_lang = q(en);}
    }

    #print "Pass 1\n";
    Readonly my $REGEXA => qr{ <div(\s)+class(\s)*=(\s)*"container"> }xms;
    Readonly my $REGEXB => qr{ <div(\s)+class(\s)*=(\s)*"bodyContent"> }xms;
    if ( $filejoined =~ m/$REGEXA(.*)$REGEXB/gis ) {
        $filejoined =~ s/($REGEXA)(.*)($REGEXB>)/$1$6/gis;
    }

    #print "Pass 2\n";
    $filejoined = $filejoined;
    if ( $filejoined =~ m/<A(\s)+NAME(\s)*=(\s)*"navbar_bottom">/gis ) {
        $filejoined =~ s/<A(\s)+NAME(\s)*=(\s)*"navbar_bottom">(.*)//gis;
    }

    Readonly my $REGEXC => qr{ <A(\s)+NAME(\s)*=(\s)*"navbar_top"> }xms;
    Readonly my $REGEXD =>
        qr{ <A(\s)+NAME(\s)*=(\s)*"skip-navbar_top"><\/A> }xms;
    $filejoined = $filejoined;
    if ( $filejoined =~ m/$REGEXC(.*)$REGEXD/gis ) {
        $filejoined =~ s/$REGEXC(.*)$REGEXD//gis;
    }

    $filejoined = $filejoined;
    if ( $filejoined =~ m/<div(\s)+class(\s)*=(\s)*"FootBookNav">/gis ) {
        $filejoined =~ s/<div(\s)+class(\s)*=(\s)*"FootBookNav">(.*)//gis;
    }

    Readonly my $REGEXEF =>
        qr{ <div(\s)+class(\s)*=(\s)*"zz-skip-header"> }xms;
    Readonly my $REGEXE =>
        qr{ $REGEXEF<a([^>]+)>Skip(\s)+Headers<\/a><\/div> }xms;
    $filejoined = $filejoined;
    while ( $filejoined =~ m/($REGEXE)/gis ) {
        $filejoined =~ s/\Q$1\E//gis;
    }

    if ( $filejoined =~ m/(<div(\s)+class(\s)*=(\s)*"footer">(.*?)<\/div>)/gis ) {
        $filejoined =~ s/\Q$1\E(.*)//gis;
    }
    if ( $filejoined =~ m/(<div(\s)+id(\s)*=(\s)*"footer">(.*?)<\/div>)/gis ) {
        $filejoined =~ s/\Q$1\E(.*)//gis;
    }
    if ( $filejoined =~ m/(<div(\s)+id(\s)*=(\s)*"RIGHTCOLUMN">(.*?)<\/div>)/gis ) {
        $filejoined =~ s/\Q$1\E(.*)//gis;
    }
    #print "Pass 3\n";

    $filejoined = $filejoined;
    while ( $filejoined
        =~ m/(<span(\s)+class(\s)*=(\s)*"icon">([^<]+)<\/span>)/gis )
    {
        $filejoined =~ s/\Q$1\E//gis;
    }
    Readonly my $REGEXFCLASS =>
        qr{ class[\s]*=[\s]*"simple[\s]*oac_no_warn" }xms;
    Readonly my $REGEXFA =>
        qr{ $REGEXFCLASS[\s]*summary=""[\s]*cellspacing="0" }xms;
    Readonly my $REGEXF =>
        qr{ <table[\s]*$REGEXFA[\s]*cellpadding="0"[\s]*width="150"> }xms;
    $filejoined = $filejoined;
    while ( $filejoined =~ m/($REGEXF(.*?)<\/table>)/gis ) {
        $filejoined =~ s/\Q$1\E//gis;
    }

    #print "Pass 4\n";

    Readonly my $REGEXG =>
        qr{ <table[\s]*$REGEXFA[\s]*cellpadding="0"[\s]*width="100"> }xms;
    $filejoined = $filejoined;
    while ( $filejoined =~ m/($REGEXG(.*?)<\/table>)/gis ) {
        $filejoined =~ s/\Q$1\E//gis;
    }

    Readonly my $REGEXH =>
        qr{ <table[\s]*$REGEXFA[\s]*cellpadding="0"[\s]*width="(.*?)"> }xms;
    $filejoined = $filejoined;
    while ( $filejoined =~ m/($REGEXH(.*?)<\/table>)/gis ) {
        $filejoined =~ s/\Q$1\E//gis;
    }

    Readonly my $REGEXI =>
        qr{ <p(\s)+class(\s)*=(\s)*"betadraftsubtitle"> }xms;
    $filejoined = $filejoined;
    while (
        $filejoined =~ m/(($REGEXI)(.*?)Beta[\s]*Draft:([^<]+)<\/p>)/gis )
    {
        $filejoined =~ s/\Q$1\E//gis;
    }
    #print "Pass 5\n";

    $filejoined = $filejoined;
    if ($filejoined =~ m/(<img(\s)+src(\s)*=(\s)*"([.\/]+)(dcommon)([^>]+)>)/i) {
      $filejoined =~ s/(<img(\s)+src(\s)*=(\s)*"([.\/]+)(dcommon)([^>]+)>)//gis;
    }
    #print "Pass 5.1\n";

    $filejoined = $filejoined;
    if ($filejoined =~ m/(<script([^>]*)>(.*?)<\/script>)/is) {
        $filejoined =~ s/(<script([^>]*)>(.*?)<\/script>)//gis;
    }
    #print "Pass 5.1.1\n";

    $filejoined = $filejoined;
    if ($filejoined =~ m/<div(\s)*([^>]*)id(\s)*=(\s)*"skiptocontent"([^>]*)>(.*?)<\/div>/gis){
        $filejoined =~ s/<div(\s)*([^>]*)id(\s)*=(\s)*"skiptocontent"([^>]*)>(.*?)<\/div>//gis;
    }
    #print "Pass 5.1.2\n";
    $filejoined = $filejoined;
    if ($filejoined =~ m/<div(\s)*([^>]*)class(\s)*=(\s)*"skiptocontent"([^>]*)>(.*?)<\/div>/gis){
        $filejoined =~ s/<div(\s)*([^>]*)class(\s)*=(\s)*"skiptocontent"([^>]*)>(.*?)<\/div>//gis;
    }
    #print "Pass 5.1.3\n";

    $filejoined = $filejoined;
    if ($filejoined =~ m/(<style([^>]*)>(.*?)<\/style>)/is) {
        $filejoined =~ s/(<style([^>]*)>(.*?)<\/style>)//gis;
    }
    #print "Pass 5.2\n";

    if ($isPSFT ne q(1) && $filejoined =~ m/<meta(.*?)content(\s)*=(\s)*"peoplesoft(.*)/is) {
        $isPSFT = q(1);
    }
    #print "Pass 5.3\n";

    if ($filejoined =~ m/<meta(.*?)name(\s)*=(\s)*"([^"]*)date([^"]*)"(\s)*content(\s)*=(\s)*"([^"]+)"/is)
    {
        $file_release_date = get_correct_date_format($9);
    }
    #print "Pass 5.3.1\n";
    if ($filejoined =~ m/<meta(.*?)name(\s)*=(\s)*"([^"]*)description([^"]*)"(\s)*content(\s)*=(\s)*"([^"]+)"/is)
    {
        $file_description = $9;
    }
    if ($filejoined =~ m/<meta(.*?)name(\s)*=(\s)*"([^"]*)keyword([^"]*)"(\s)*content(\s)*=(\s)*"([^"]+)"/is)
    {
        $file_keywords = $9;
    }

    $filejoined = $filejoined;
    if ($javadoc ne "1" && $filejoined =~ m/class(\s)*=(\s)*"bodyContent"/is){
        $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"container">)(.*)(<div(\s)+class(\s)*=(\s)*"bodyContent")>/$1$6/is;
    }
    #print "Pass 6\n";
    ##$filejoined = $filejoined;
    if ($javadoc eq  "1") {
      $filejoined =~ s/<A(\s)+NAME(\s)*=(\s)*"navbar_bottom">(.*)//is;    
    }
    #print "Pass 6.0.1\n";

    if ($javadoc ne "1" && $file !~ m/index.htm/ && $filejoined =~ m/<article/is) {
      $filejoined =~ s/<body([^>]*)>(.*?)<article([^>]*)>/<body><article>/is;
    }

    if ($javadoc ne "1" && $file !~ m/index.htm/ && $filejoined =~ m/<article>/is) {
      $filejoined = $filejoined;
      $filejoined =~ s/(.*?)<body>(.*?)<article>(.*?)<\/article>(.*?)/$1<body>$2<article>$3<\/article>/is;
    }
    #print "Pass 6.1\n";
    Readonly my $REGEXJA => qr{ align="right"(\s)+style="width:120px; }xms;
    Readonly my $REGEXJB => qr{ float:right;(\s)*padding-right:5px; }xms;
    Readonly my $REGEXJC =>
        qr{ padding-top:4px;(\s)*right:2px;(\s)*display:inline;" }xms;
    Readonly my $REGEXJ =>
        qr{ <div(\s)+$REGEXJA(\s)+$REGEXJB(\s)*$REGEXJC> }xms;
    Readonly my $REGEXPDF =>
        qr{ <a(\s)+href="[.][.]\/([[:upper:]]{5})[.]pdf">\[PDF\]<\/a> }xms;
    Readonly my $REGEXMOBI =>
        qr{ <a(\s)+href="[.][.]\/([[:upper:]]{5})[.]mobi">\[Mobi\]<\/a> }xms;
    Readonly my $REGEXEPUB =>
        qr{ <a href="[.][.]\/([[:upper:]]{5})[.]epub">\[ePub\]<\/a> }xms;
    Readonly my $REGEXK =>
        qr{ $REGEXPDF(\s)*$REGEXMOBI(\s)+$REGEXEPUB<br \/>([^<]+)<\/div> }xms;
    $filejoined = $filejoined;

    if ( $javadoc ne q(1) && $filejoined =~ m/<a href="[.][.]\/([[:upper:]]{5})[.]epub">\[ePub\]<\/a>/is) {
        $filejoined =~ s/$REGEXJ$REGEXK//is;
    }
    #print "Pass 7: \n";

    $filejoined = $filejoined;
    if ($noid ne "1" && $javadoc ne "1" && $file !~ m/index.html$/) {

            $filejoined =~ s/(<span(\s)*class(\s)*=(\s)*"page"([^>]*)>(.*?)<\/span>)//is;
            #print "Pass 7.1: \n";

            $filejoined =~ s/(<body([^>]*)>)(.*?)(<div(\s)*class(\s)*=(\s)*"pagePane([^"]*)">)/$1$4/gis;
            #print "Pass 7.1: \n";

            $filejoined =~ s/<img([^>]*)>//gis;
            #print "Pass 7.2: \n";

            $filejoined =~ s/(<body([^>]*)>)(.*?)(<div(\s)*class(\s)*=(\s)*"ContentPane">)/$1$4/gis;
            #print "Pass 7.3: \n";

            $filejoined =~ s/<div(\s)+align="right"(\s)+style="width:120px;(\s)+float:right;(\s)*padding-right:5px;(\s)*padding-top:4px;(\s)*right:2px;(\s)*display:inline;"><a(\s)+href="\.\.\/([A-Z]{5})\.pdf">\[PDF\]<\/a>(\s)*<a(\s)+href="\.\.\/([A-Z]{5}).mobi">\[Mobi\]<\/a>(\s)+<a href="\.\.\/([A-Z]{5})\.epub">\[ePub\]<\/a><br \/>([^<]+)<\/div>//gs;
            #print "Pass 7.4: \n";
            ####
            if ($filejoined =~ m/"articlewrapper"/gis) {
              $filejoined =~ s/(<body([^>]*)>)(.*?)(<div([^>]*)class(\s)*=(\s)*"articlewrapper"([^>]*)>)/$4/gis;
              $filejoined =~ s/(<div(\s)*id(\s)*=(\s)*"rightslider">)(.*?)(<div([^>]*)class(\s)*=(\s)*"articlewrapper"([^>]*)>)/$6/gis;
            }
            #print "Pass 7.4.1: \n";
            #print "Pass 7.5: \n";

            if ($filejoined =~ m/"panellists"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"panellists">(.*?)<\/div>)//is;
            }
            #print "Pass 7.6: \n";

            if ($filejoined =~ m/"breadcrumb"/is) {
              $filejoined =~ s/(<div(\s)+id(\s)*=(\s)*"breadcrumb">(.*?)<\/div>)//is;
            }
            #print "Pass 7.7: \n";

            if ($filejoined =~ m/"bookoptions"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"bookoptions">(.*?)<\/div>)//is;
            }
            #print "Pass 7.8: \n";

            if ($filejoined =~ m/"pagenav"/is) {
              $filejoined =~ s/(<div(\s)+id(\s)*=(\s)*"pagenav">(.*?)<\/div>)//is;
            }
            #print "Pass 7.9: \n";

            if ($filejoined =~ m/"contentarea(\s+)clearboth"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"contentarea(\s+)clearboth">(.*?)<\/div>)//is;
            }
            #print "Pass 7.10: \n";

            if ($filejoined =~ m/"altlogo"/is) {
              $filejoined =~ s/(<div(\s)+id(\s)*=(\s)*"altlogo">(.*?)<\/div>)//gis;
            }

            #print "Pass 7.11: \n";

            if ($filejoined =~ m/"searchheader"/is) {
              $filejoined =~ s/(<div(\s)+id(\s)*=(\s)*"searchheader">(.*?)<\/div>)//is;
            }
            #print "Pass 7.12: \n";

            if ($filejoined =~ m/"signInArea"/is) {
              $filejoined =~ s/(<div(\s)+id(\s)*=(\s)*"signInArea">(.*?)<\/div>)//is;
            }
            #print "Pass 7.13: \n";

            if ($filejoined =~ m/<header/is) {
              $filejoined =~ s/(<header(.*?)>(.*?)<\/header>)//is;
            }

            #print "Pass 7.14: \n";

            if ($filejoined =~ m/"tw-like"/is) {
              $filejoined =~ s/(<div(\s)+id="tw-like">(.*?)<\/div>)//is;
            }
            #print "Pass 7.15: \n";

            if ($filejoined =~ m/"g-plusone"/is) {
              $filejoined =~ s/(<div(\s)+class="g-plusone"(.*?)>(.*?)<\/div>)//is;
            }
            #print "Pass 7.16: \n";

            if ($filejoined =~ m/"go-like"/is) {
              $filejoined =~ s/(<div(\s)+id="go-like">(.*?)<\/div>)//is;
            }
            #print "Pass 7.17: \n";

            if ($filejoined =~ m/"fb-root"/is) {
              $filejoined =~ s/(<div(\s)+id="fb-root">(.*?)<\/div>)//is;
            }
            #print "Pass 7.18: \n";

            if ($filejoined =~ m/"fb-like"/is) {
              $filejoined =~ s/(<div(\s)+id="fb-like">(.*?)<\/div>)//is;
            }
            #print "Pass 7.19: \n";

            if ($filejoined =~ m/"SocialBar"/is) {
              $filejoined =~ s/(<div(\s)+class="SocialBar">(.*?)<\/div>)//is;
            }
            #print "Pass 7.20: \n";

            if ($filejoined =~ m/"footer-containerbook/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"footer-containerbook(.*?)">(.*?)<\/div>)//is;
            }
            #print "Pass 7.21: \n";

            if ($filejoined =~ m/"socialicons"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"socialicons">(.*?)<\/div>)//gis;
            }
            #print "Pass 7.22: \n";

            $filejoined =~ s/(<footer>(.*?)<\/footer>)//is;
            #print "Pass 7.23: \n";

            if ($filejoined =~ m/"socialicons"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"socialicons">(.*?)<\/div>)//gis;
            }

            if ($filejoined =~ m/"footer"/is) {
              $filejoined =~ s/<div(\s)*class(\s)*=(\s)*"footer">(.*)//is;
              #print "Pass 7.24: \n";

              $filejoined =~ s/<div(\s)*id(\s)*=(\s)*"footer">(.*)//is;
              #print "Pass 7.25: \n";
            }

            if ($filejoined =~ m/"booktitle"/is) {
              $filejoined =~ s/(<div(\s)+class(\s)*=(\s)*"booktitle">(.*?)<\/div>)//is;
            }

            #print "Pass 7.26: \n";
            if ($filejoined =~ m/<div(\s)*class(\s)*=(\s)*"ind">/is) {
                $filejoined =~ s/(<body([^>]+)>)(.*?)(<div(\s)*class(\s)*=(\s)*"ind">)/$1$4/is;
            }
            #print "Pass 7.27: \n";
            if ($isPSFT eq q(1)) {
              $filejoined =~ s/<span(\s)*class(\s)*=(\s)*"PTTRANSPARENT"(\s)*([^>]*)>(.*?)<\/([^>]+)>//gis;  
            }
            #print "Pass 7.28: \n";
    }

    ### Process tables first
    #print "Pass 8\n";


    if ( $noid ne "1" && $javadoc ne "1" && $filejoined =~ m/<table/gis ) {
        while ( $filejoined =~ m/((<table.*?>)(.*?)<\/table>)/gis ) {
            my $howmanytablechunks = 0;
            my $this_table = $1;
            my $original_table_content = $3;
            $original_table_content =~ s/<tbody([^>]*)>//gis;
            $original_table_content =~ s/<\/tbody>//gis;
            #print "Pass 8.1\n";
            # find the ids and append as a id before it
            my $backup_table = $original_table_content;
            my $idstreamtable
                = HTML::TokeParser->new( \$original_table_content )
                or carp "Couldn't parse: $OS_ERROR\n";
            my $tableid     = q();
            my $thishead    = q();
            my $thistabsumm = q();
            while ( my $tabletoken = $idstreamtable->get_token() ) {
                if ( $tabletoken->[0] eq q(S)
                    && exists $tabletoken->[2]{'id'} )
                {
                    if (   length $tableid > 0
                        && length $thishead > 0
                        && length $thistabsumm > 0 )
                    {
                        $thistabsumm =~ s/(\s)+$//gis;
                        my $stdocida  = $tableid;
                        my $chunkname = $file_name_val . q(-) . $stdocida;
                        $count_chunks++;
                        $howmanytablechunks++;
                        write_chunk(
                            $file,
                            $stdocida,
                            $file_path . q(/) . $chunkname,
                            q(),
                            q(<head>) . $thishead . q(</head>),
                            $thistabsumm,
                            q(),
                            q()
                        );
                        $tableid     = q();
                        $thishead    = q();
                        $thistabsumm = q();

                    }

                    # this is the head
                    $tableid  = $tabletoken->[2]{'id'};
                    $thishead = $idstreamtable->get_trimmed_text(
                        q(/) . $tabletoken->[1] );
                    $thishead =~ s/^( )+//gis;
                    $thishead =~ s/( )+$//gis;
                }
                elsif (
                    $tabletoken->[0] eq 'S'
                    && (   $tabletoken->[1] eq q(td)
                        || $tabletoken->[1] eq q(th) )
                    )
                {
                    my $thiscell = $idstreamtable->get_trimmed_text(q(/) . $tabletoken->[1] );

                    # check if the heading was just a numbering

                    if ($thishead =~ m/^[\s\d.]+$/i) {
                        $thishead = $thishead.q( ).$thiscell;
                    }
                    elsif (length($thishead) < 1) {
                        $thishead = $thiscell;
                    }
                    else {
                        $thistabsumm
                          .= $thiscell . q( );
                    }
                    #$thistabsumm
                    #    .= $thiscell . q( );
                }
            }
            if (   length $tableid > 0
                && length $thishead > 0
                && length $thistabsumm > 0 )
            {
                $thistabsumm =~ s/(\s)+$//gis;
                my $stdocida  = $tableid;
                my $chunkname = $file_name_val . q(-) . $stdocida;
                $count_chunks++;
                $howmanytablechunks++;
                write_chunk(
                    $file,
                    $stdocida,
                    $file_path . q(/) . $chunkname,
                    q(),
                    q(<head>) . $thishead . q(</head>),
                    $thistabsumm,
                    q(),
                    q()
                );
                $tableid     = q();
                $thishead    = q();
                $thistabsumm = q();
            }

            if ($howmanytablechunks > 0) {
              $filejoined =~ s/\Q$this_table\E//gis;
            }
        }
    }
    # remove the tablefootnote, titleintable
    #print "Pass 8.1\n";
    pos( $filejoined ) = 0;
    $filejoined
        =~ s/<p(\s)*class(\s)*=(\s)*"tablefootnote"([^>]*)>(.*?)<\/p>//gis;

    $filejoined = $filejoined;
    $filejoined
        =~ s/<p(\s)*class(\s)*=(\s)*"titleintable"([^>]*)>(.*?)<\/p>//gis;
    #print "Pass arranging\n";
    @file_content = arrange( $filejoined, $file );
    #print "Pass arrangws\n";
    @cleanedlines = @file_content;
    @file_content = ();

    $lastchunkwritten = q();
    my $countChunks = 0;
    for (
        my $linenumber = 0;
        $linenumber < scalar @cleanedlines;
        $linenumber++
        )
    {    
        #print "LINE: ".$cleanedlines[$linenumber]."\n";
        if ($cleanedlines[$linenumber] =~ m/<[h]/i) {
            #code
        }
        if ( $cleanedlines[$linenumber]
            =~ m/<script([^>]+)text\/javascript([^>]+)>/gis )
        {
            while ($linenumber < scalar @cleanedlines
                && $cleanedlines[$linenumber] !~ m/<\/script>/gis )
            {
                $linenumber++;
            }
            next;
        }
        if ( $cleanedlines[$linenumber] =~ m/<noscript>/gis ) {
            while ($linenumber < scalar @cleanedlines
                && $cleanedlines[$linenumber] !~ m/<\/noscript>/gis )
            {
                $linenumber++;
            }
            next;
        }
        if ( $cleanedlines[$linenumber] =~ m/<style>/gis ) {
            while ($linenumber < scalar @cleanedlines
                && $cleanedlines[$linenumber] !~ m/<\/style>/gis )
            {
                $linenumber++;
            }
            next;
        }
        $wasadded             = q();
        $was_last_token_added = q();
        my $idstream = HTML::TokeParser->new( \$cleanedlines[$linenumber] )
            or carp "Couldn't parse: $OS_ERROR";

        while ( my $token = $idstream->get_token() ) {

            if ((      $token->[0] eq 'S'
                    && $token->[2] && $token->[2]{'lang'}
                )
                )
            {
                $file_lang = $token->[2]{'lang'};
                if (lc $file_lang eq q(en-us)) {$file_lang = q(en);}
            }

            if ((      $token->[0] eq 'S'
                    && $token->[1] eq 'meta'
                    && $token->[2]{'name'}
                    && $token->[2]{'name'} eq 'generator'
                    && $token->[2]{'content'}
                    && lc( $token->[2]{'content'} ) =~ m/jadey/gis
                )
                )
            {
                $javadoc = q(1);
            }

            if ($isPSFT eq q(1) && $token->[1] =~ m/h[1-6]/isxm && $token->[2]{'class'} && $token->[2]{'class'} =~ m/PTPAGELETHEADER/)
            {
                #print "Pass initial\n";
                # head was found
                # 1. check if the previous collection has ID, head and body
                ( $head, $body, $chunkclass, $chunkheadlevel, @allids )
                    = beautify_chunk($chunk);
                if (   length $head < 1
                    || length $body < 1
                    || ( scalar @allids < 1 )
                    || verify_chunk( $head, $body ) == 0 )
                {
                    # continue to add
                    if ( $wasadded ne q(1) ) {
                        $chunk .= $cleanedlines[$linenumber] . "\n";
                        $wasadded = q(1);
                    }
                }
                elsif (length $head > 0
                    && length $body > 0
                    && ( scalar @allids ) > 0 )
                {
                    # chunk formed
                    my $stdocidb  = get_apt_id(@allids);
                    $lastidused = $stdocidb;
                    my $chunkname = $file_name_val . q(-) . $stdocid;
                    $count_chunks++;
                    write_chunk( $file, $stdocidb,
                        $file_path . q(/) . $chunkname,
                        $chunk, $head, $body, $chunkclass, $chunkheadlevel );
                    $head           = q();
                    $body           = q();
                    $chunkclass     = q();
                    $chunkheadlevel = q();
                    @allids         = ();
                    $chunk          = q();

                    if ( $wasadded ne q(1) ) {
                        $chunk    = $cleanedlines[$linenumber] . "\n";
                        $wasadded = q(1);
                    }
                }
            }
            elsif ($isPSFT ne q(1) && $javadoc ne q(1) && $token->[0] eq 'S'
                && ((   $token->[2]{'class'}
                        && ( any { $_ eq $token->[2]{'class'} } @headclasses )
                    )
                    || ( $token->[1] =~ m/h[1-6]/isxm )
                    || ( ( $token->[1] =~ m/dt/gis ) && $javadoc ne q(1) && $noid ne "1")
                    || ( $token->[1] =~ m/caption/isxm )
                    || ( ( $token->[1] =~ m/tr/isxm ) && $javadoc ne q(1) && $noid ne "1")
                )
                )
            {

                #print "Pass breaking $token->[1]\n";
                # head was found
                # 1. check if the previous collection has ID, head and body
                ( $head, $body, $chunkclass, $chunkheadlevel, @allids )
                    = beautify_chunk($chunk);
                if (   length $head < 1
                    || length $body < 1
                    || ( scalar @allids < 1 )
                    || verify_chunk( $head, $body ) == 0 )
                {
                    # continue to add
                    if ( $wasadded ne q(1) ) {
                        $chunk .= $cleanedlines[$linenumber] . "\n";
                        $wasadded = q(1);
                    }
                }
                elsif (length $head > 0
                    && length $body > 0
                    && ( scalar @allids ) > 0 )
                {
                    # chunk formed
                    my $stdocidb  = get_apt_id(@allids);
                    $lastidused = $stdocidb;
                    my $chunkname = $file_name_val . q(-) . $stdocid;
                    $count_chunks++;
                    write_chunk( $file, $stdocidb,
                        $file_path . q(/) . $chunkname,
                        $chunk, $head, $body, $chunkclass, $chunkheadlevel );
                    $head           = q();
                    $body           = q();
                    $chunkclass     = q();
                    $chunkheadlevel = q();
                    @allids         = ();
                    $chunk          = q();

                    if ( $wasadded ne q(1) ) {
                        $chunk    = $cleanedlines[$linenumber] . "\n";
                        $wasadded = q(1);
                    }
                }
            }
            elsif (
                (
                   (   $token->[0] eq q(S) && $token->[2]{'id'} && length $token->[2]{'id'} > 0 )
                   ||
                   (   $token->[0] eq q(S) && $token->[1] eq q(a) && $token->[2]{'name'} && length $token->[2]{'name'} > 0 )
                )
                &&
                (      !( $token->[1] eq 'th' )    # do not consider tables
                    && !( $token->[1] eq 'tr' )
                    && !( $token->[1] eq 'thead' )
                    && !( $token->[1] eq 'tbody' )
                )
                &&
                $isPSFT ne q(1)
                )
            {
                #print "Pass third one: $token->[1]; $cleanedlines[$linenumber]\n";
                # this could also mark the beginning
                if (   ($token->[2]{'id'} && $token->[2]{'id'} =~ m/BEGIN/isxm)
                    || ($token->[2]{'name'} && $token->[2]{'name'} =~ m/BEGIN/isxm) )
                {
                    $chunk = q();
                }
                if ( $token->[2]{'id'} && ($token->[2]{'id'} =~ m/sthref(\d)+/gsxm
                    || $token->[2]{'id'} =~ m/i(\d)+/gis) )
                {
                    if ( $wasadded ne q(1) ) {

                        $chunk .= $cleanedlines[$linenumber] . "\n";
                        $wasadded = q(1);
                    }
                }
                else {

                    $last_id_found = $token->[2]{'id'};
                    if ( $token->[2]{'name'} && length $token->[2]{'name'} > 0
                        && (! $token->[2]{'id'} || length $token->[2]{'id'} < 1) )
                    {
                        $last_id_found = $token->[2]{'name'};
                    }

                    ( $head, $body, $chunkclass, $chunkheadlevel, @allids )
                        = beautify_chunk($chunk);

                    if (   length $head < 1
                        || length $body < 1
                        || ( scalar @allids < 1 )
                        || verify_chunk( $head, $body ) == 0 )
                    {
                        # continue to add
                        if ( $wasadded ne q(1) ) {

                            $chunk .= $cleanedlines[$linenumber] . "\n";
                            $wasadded = q(1);
                        }
                    }
                    elsif (length $head > 0
                        && length $body > 0
                        && ( scalar @allids ) > 0 )
                    {
                        # chunk formed
                        my $stdocid   = get_apt_id(@allids);
                        $lastidused = $stdocid;
                        my $chunkname = $file_name_val . q(-) . $stdocid;
                        $count_chunks++;

                        write_chunk(
                            $file,                          $stdocid,
                            $file_path . q(/) . $chunkname, $chunk,
                            $head,                          $body,
                            $chunkclass,                    $chunkheadlevel
                        );
                        $head           = q();
                        $body           = q();
                        $chunk          = q();
                        $chunkclass     = q();
                        $chunkheadlevel = q();
                        @allids         = ();
                        if ( $wasadded ne q(1) ) {
                            $chunk    = $cleanedlines[$linenumber] . "\n";
                            $wasadded = q(1);
                        }
                    }
                }
            }
            else {
                # continue to add lines
                if ( $wasadded ne q(1) ) {
                    $chunk .= $cleanedlines[$linenumber] . "\n";
                    $wasadded = q(1);
                }
            }
        }
    }

 # here the holder will have the last ID and its text...so write it as a chunk
    if (scalar @allids < 1) {
        if (length ($lastidused) < 1) {
            push @allids, "JUSTANID";
        }
        else {
            push @allids, $lastidused;
        }
    }
    if ( length $chunk > 0 ) {
        ( $head, $body, $chunkclass, $chunkheadlevel, @allids )
            = beautify_chunk($chunk);

        if (   length $head > 0
            && length $body > 0
            && ( scalar @allids > 0 ) )
        {
            my $stdocid = get_apt_id(@allids);
            if ( length $stdocid < 1 ) {
                $stdocid = $last_id;
            }
            my $chunkname = $file_name_val . q(-) . $stdocid;
            $count_chunks++;

            write_chunk( $file, $stdocid, $file_path . q(/) . $chunkname,
                $chunk, $head, $body, $chunkclass, $chunkheadlevel );
            $chunk = q();
        }
    }

    $chunk = q();
    $lastidused = q();
    # was any chunk written?
    if ($thisfilechunkcount < 1) {
        # get all the text and write to the json
        my $idstreamlast = HTML::TokeParser->new( \$filejoined )
            or carp "Couldn't parse: $OS_ERROR\n";
        my $alltext = '';
        while ( my $stokenp = $idstreamlast->get_token() ) {
            $alltext
                .= q( ) . $idstreamlast->get_trimmed_text();
        }
        $alltext =~ s/^( )+//gis;
        $alltext =~ s/( )+$//gis;
        $alltext =~ s/( )+/ /gis;
        my $file_path_to_use = $file;
        $file_path_to_use =~ s/\\/\//gis;
        writejsonchunk( "JUSTANID", $chaptertitle, $alltext, $file_path_to_use."-abc", "", $file_path_to_use);
        ##writejsonchunk( $id_val, $cleanhead, $shortbody, $chunkpath, $classtarget, $cpth);
    }
    $filejoined   = q();
}

# Construct the Summary:
$time = time - $time;
$time = time_format_standard($time);
my $topic_type_summary = q();
foreach my $keys (keys %topic_types_found){
    $topic_type_summary = $topic_type_summary.$keys.q(:).$topic_types_found{$keys}.";";
}
my $summary
    = q(Summary for run on ) . $check_date . "                           \n";
$summary .= "--------------------------------------------\n";
$summary .= q(Total Files Scanned   : )
    . ( ( scalar @dir ) - $ignore_track ) . "\n";
$summary .= q(Total Files Ignored   : ) . $ignore_track . "\n";
$summary .= q(Total mappings        : ) . scalar (keys %index_type_mappings) . "\n";
$summary .= q(Total JSON Chunks     : ) . $summary_count . "\n";
$summary .= q(Topic Types found     : ) . $topic_type_summary . "\n";
$summary .= q(Total time Taken      : ) . $time . "\n";
$summary .= "=========================================================\n";
$summary .= "Errors: \n";

foreach my $el (@errorlines) {
    $summary .= $el . "\n";
}

open my $SUMMARY, '>', $output_folder . q(/log.txt)
    or croak "Cannot open log for writing: $OS_ERROR";
binmode $SUMMARY, ':encoding(UTF-8)';
print {$SUMMARY} $summary . "\n";
close $SUMMARY
    or carp "Can't close summary file: $OS_ERROR";

printf qq[\r$summary\t];

################################################################################

################################################################################
## sub-routines
################################################################################

##############################################################################
# Function: is_html_file
# Parameters:
#
# Description: Tests to see if the file passed is an htm[l] file and checks is it needs to
# be ignored
##############################################################################

sub is_html_file {
    my $filename = shift @_;
    if ( $filename =~ /[.]ht(m|ml)$/ ) {
        return 1;
    }
    return 0;

}

##############################################################################
# Function: is_ignored
# Parameters:
#
# Description: Test if a filename is a special one we don't index.
##############################################################################

sub is_ignored {
    my $filename = shift @_;
    return ( $filename =~ /index(.*)[.]ht(m|ml)$/isxm );
}

##############################################################################
# Function: wanted
# Parameters:
#
# Description: Traversal method to get the list of files to be processed
##############################################################################

sub wanted {
    my $temp = $_;
    if (   is_html_file($temp)
        && $temp !~ m/.*\/img_text\/.*/gis
        && $temp !~ m/.*\/images\/([^\/]*)/gis
        && $temp !~ m/.*\/image([^\/]*)\/([^\/]*)/gis
        && $temp !~ m/.*\/dcommon\/.*/gis
        && $temp !~ m/.*\/nav\/.*/gis )
    {
        @dir = ( @dir, $temp );
        if ( $temp =~ m/title.htm$/gis ) {
            push @titlefiles, $temp;
        }
    }
    return;
}

##############################################################################
# Function: is_xlink
# Parameters:
#
# Description: Test if an anchor represents an xlink.
##############################################################################

sub is_xlink {
    my $anchor = shift @_;
    return ( $anchor =~ /^[[:upper:]]{5}[\d]{3,5}$/isxm );
}

##############################################################################
# Function: scanify
# Parameters:
#
# Description: Extract the text from head for target file
##############################################################################

sub scanify {
    my $text_to_scan = shift;
    my $entiretext   = q();
    my $subtext      = q();
    my $sstream      = HTML::TokeParser->new( \$text_to_scan )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $stoken = $sstream->get_token() ) {
        if ( ( $stoken->[0] eq 'S' && $stoken->[1] eq 'head' ) ) {
            $subtext = $sstream->get_trimmed_text('/head');
        }
    }
    return limitheading($subtext);
}

##############################################################################
# Function: limitheading
# Parameters:
#
# Description: Make the heading only upto 250 characters
##############################################################################

sub limitheading {
    my $texttostrip = shift;
    if ( length $texttostrip > $SHORTCUT_CHAR_LENGTH_TWOFIFTY ) {
        my @wordsofheading = split / /, $texttostrip;
        $texttostrip = q();
        foreach my $word (@wordsofheading) {
            $texttostrip = $texttostrip . q( ) . $word;
            if ( length $texttostrip > $SHORTCUT_CHAR_LENGTH_TWOFIFTY ) {
                last;
            }
        }
        undef @wordsofheading;
    }
    return $texttostrip;
}

##############################################################################
# Function: limitbody
# Parameters:
#
# Description: Make the body only upto 250 characters
##############################################################################

sub limitbody {

    my $texttostrip = shift;
    $texttostrip =~ s/\Q&nbsp;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/\Q&#x00A0;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/\Q&#160;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/(\s)+/ /gis;

    if ( length $texttostrip > $SHORTCUT_CHAR_LENGTH_TWOFIFTY ) {
        my @wordsofheading = split / /, $texttostrip;
        $texttostrip = q();
        foreach my $word (@wordsofheading) {
            $texttostrip = $texttostrip . q( ) . $word;
            if ( length $texttostrip > $SHORTCUT_CHAR_LENGTH_TWOFIFTY ) {
                last;
            }
        }
        undef @wordsofheading;
    }

    return $texttostrip;
}

##############################################################################
# Function: normalize_spaces
# Parameters:
#
# Description: remove extra white spaces
##############################################################################

sub normalize_spaces {
    my $texttostrip = shift;
    $texttostrip =~ s/\Q&nbsp;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/\Q&#x00A0;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/\Q&#160;\E/ /gis;
    $texttostrip = $texttostrip;
    $texttostrip =~ s/(\s)+/ /gis;
    return $texttostrip;
}

##############################################################################
# Function: limittext
# Parameters:
#
# Description: Limits the text to certain number of characters
##############################################################################

sub limittext {

    my $texttostrip = shift;
    my $sizetolimit = shift;
    $texttostrip = normalize_spaces($texttostrip);

    if ( length $texttostrip > $sizetolimit ) {
        my @wordsofheading = split / /, $texttostrip;
        $texttostrip = q();
        foreach my $word (@wordsofheading) {
            $texttostrip = $texttostrip . q( ) . $word;
            if (length $texttostrip
                > ( $sizetolimit - $SHORTCUT_CHAR_LENGTH_FIFTY ) )
            {
                last;
            }
        }
        undef @wordsofheading;
    }
    return $texttostrip;
}

##############################################################################
# Function: time_format_standard
# Parameters:
#
# Description: Converts seconds to format HH::MM::SS
##############################################################################

sub time_format_standard {
    my $time_passed  = shift;
    my $seconds_here = $time_passed % $SHORTCUT_CHAR_LENGTH_SIXTY;

    my $minutes_here = int( $time_passed / $SHORTCUT_CHAR_LENGTH_SIXTY );
    my $hours_here   = int( $minutes_here / $SHORTCUT_CHAR_LENGTH_SIXTY );
    $minutes_here
        = $minutes_here - ( $hours_here * $SHORTCUT_CHAR_LENGTH_SIXTY );

    if ( $hours_here < $SHORTCUT_CHAR_LENGTH_TEN ) {
        $hours_here = q(0) . $hours_here;
    }
    if ( $minutes_here < $SHORTCUT_CHAR_LENGTH_TEN ) {
        $minutes_here = q(0) . $minutes_here;
    }
    if ( $seconds_here < $SHORTCUT_CHAR_LENGTH_TEN ) {
        $seconds_here = q(0) . $seconds_here;
    }

    my $ret_val = $hours_here . q(:) . $minutes_here . q(:) . $seconds_here;
    return $ret_val;
}

##############################################################################
# Function: correct_representation
# Parameters:
#
# Description: Converts the text appropriately.
##############################################################################

sub correct_representation {
    my $datacontent = shift;
    $datacontent = encode_entities($datacontent);
    $datacontent =~ s{(&[A-Za-z0-9.]+;)}{htmlents2numeric($1)}xmsge;
    $datacontent =~ s{(&[A-Za-z0-9.]+;)}{isoents2numeric($1)}xmsge;
    $datacontent =~ s{(&\#x?[A-fa-f\d]+;)}{numeric2ascii($1)}xmsge;
    return $datacontent;
}

##############################################################################
# Function: write_chunk
# Parameters:
#
# Description: write to a chunk file
##############################################################################

sub write_chunk {
    my $file_being_studied = shift;
    my $id_val             = shift;
    my $chunkpath          = shift;
    my $chunkcontent       = shift;
    my $headpart           = shift;
    my $bodypart           = shift;
    my $original_body      = $bodypart;
    my $chunkcl = shift;
    my $chunkhl = shift;
    $lastchunkwritten = $chunkpath;
    if ( length $id_val < 1 || length $headpart < 1 || length $bodypart < 1 )
    {
        #print "Returned at 3: $id_val : $headpart : $bodypart\n";
        return;
    }
    my $newidval = $id_val;
    $newidval =~ s/\r//gis;
    $newidval =~ s/\n//gis;
    if ( $newidval ne $id_val ) {
        #print "Returned at 4\n";
        return;
    }

    my $plainhead = removehtmlmarkupbody($headpart);
    $bodypart =~ s/^\Q$plainhead\E(\s)*//;
    ##$bodypart =~ s/^([\s\d.]+)//isxm;

    #$original_body =~ s/^([\s\d.]+)//i;
    #$original_body =~ s/^\Q$plainhead\E(\s)*//;

    if ( length $bodypart < 1 ) {
        $bodypart = $original_body;
    }

    $bodypart =~ s/</&lt;/gis;
    $bodypart =~ s/>/&gt;/gis;

    if ( $headpart =~ m/<head><\/head>/isxm ) {
        $headpart = q();
        my @pextractedwords = split / /, $bodypart;
        foreach my $wrds (@pextractedwords) {
            $headpart .= $wrds . q( );
            if ( length $headpart > $SHORTCUT_CHAR_LENGTH_NINENINE ) {
                last;
            }
        }
        $headpart = q(<head>) . $headpart . q(</head>);
    }
    my $cleanedcontent = $headpart."\n".$bodypart;
    &get_targets ($file_being_studied, $id_val, $cleanedcontent, "", $chunkpath, $chunkcl, $chunkhl);
    my $classtarget = q();
    my $targetpath = $file_being_studied."#".$id_val;
    if ($targets{$targetpath}{class}) {
        $classtarget = $targets{$targetpath}{class};
    }
    my $cpth = q();
    if ($targets{$targetpath}{path}) {
        $cpth = $targets{$targetpath}{path};
    }
    my $ctitle = q();
    if ($targets{$targetpath}{title}) {
        $ctitle = $targets{$targetpath}{title};
    }
    my $cleanhead = removehtmlmarkup($headpart);
    my $shortbody = removehtmlmarkupbodyall($original_body);
    #print "Now writing chunk: $id_val, $cleanhead, $shortbody, $chunkpath, $classtarget, $cpth\n";
    writejsonchunk( $id_val, $cleanhead, $shortbody, $chunkpath, $classtarget, $cpth);

    return;
}

##############################################################################
# Function: beautify_chunk
# Parameters:
#
# Description: scans the chunk and tries to filter it.
# to do: remove the a id tags from the filtered head .
# These ids should also be attached to the @allids
# to do: get the head from the element A title
##############################################################################

sub beautify_chunk {
    my $text_to_scan = shift;
    my $text_backup  = $text_to_scan;
    $text_to_scan =~ s/\r//gis;
    $text_to_scan =~ s/\n//gis;
    my $was_head_found    = q();
    my $was_section_found = q();
    my $relevant_head     = q();
    my $relevanttext      = q();
    my $text_before_head  = q();
    my $id_lines          = q();
    my @extract_ids       = ();
    my $class             = q();
    my $headlevel         = q();

    my $wasrelevancefound = q();

    if (   $text_to_scan =~ m/<h[1-6]([^>]*)>/gis
        || $text_to_scan =~ m/<\/h[1-6]>/gis )
    {
        $wasrelevancefound = q(1);
    }
    elsif (   $text_to_scan =~ m/<dt(\s)*([^>]*)>/gis
           && $text_to_scan =~ m/<\/dt>/gis
          )
    {
        $wasrelevancefound = q(1);
    }
    elsif ($text_to_scan =~ m/<tr([^>]*)>/gis
        && $text_to_scan =~ m/<\/tr>/gis )
    {
        $wasrelevancefound = q(1);
    }
    else {
        foreach my $searchclass (@headclasses) {
            if ( $text_to_scan =~ m/"$searchclass"/gis ) {
                $wasrelevancefound = q(1);
                $class             = $searchclass;
            }
        }
    }
    if (( $wasrelevancefound eq q(0) )
        && (   $text_to_scan =~ m/<tr([^>]+)>/gis
            && $text_to_scan !~ m/<\/tr>/gis )
        )
    {
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }
    if (( $wasrelevancefound eq q(0) )
        && (   $text_to_scan =~ m/<\/p>/gis
            || $text_to_scan =~ m/<p ([^>]+)>/gis )
        )
    {
        $wasrelevancefound = q(1);

        # special case when the id was found and the head is not there but <p>
        # available.
        # make the first 100 characters as head.
        my $idstreamp = HTML::TokeParser->new( \$text_backup )
            or carp "Couldn't parse: $OS_ERROR\n";
        while ( my $idtokenp = $idstreamp->get_token() ) {
            if (   ( $idtokenp->[0] eq 'S' && $idtokenp->[1] eq 'a' )
                || ( length $idtokenp->[2]{'id'} > 0 ) )
            {
                my $secidp   = $idtokenp->[2]{'id'};
                my $secnamep = $idtokenp->[2]{'name'};

                if ( length $secnamep > 0 ) {
                    $secidp = $secnamep;
                }
                if ( length $secidp > 0 ) {
                    push @extract_ids, $secidp;
                }
            }
        }
        if ( scalar @extract_ids < 1 ) {
            return ( $relevant_head, $relevanttext, $class,
                $headlevel, @extract_ids );
        }
        my $extractedplaintextp = q();
        if ( $text_backup =~ m/<title>(.*)<\/title>/isxm ) {
            $relevant_head = $1;
            if ( length $relevant_head > 0 ) {
                $relevant_head = q(<head>) . $relevant_head . q(</head>);
            }
        }
        my $sectionstreamp = HTML::TokeParser->new( \$text_backup )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $stokenp = $sectionstreamp->get_token() ) {
            $extractedplaintextp
                .= q( ) . $sectionstreamp->get_trimmed_text();
        }

        # get the first
        if ( length $extractedplaintextp <= $HUNDRED ) {
            $relevant_head = q(<head>) . $extractedplaintextp . q(</head>);
            $relevanttext  = $extractedplaintextp;
            $headlevel     = q();
        }
        else {
            $relevant_head = q();
            $relevanttext  = $extractedplaintextp;
            my @extractedwords = split / /, $extractedplaintextp;
            foreach my $wrds (@extractedwords) {
                $relevant_head .= $wrds . q( );
                if ( length $relevant_head > $SHORTCUT_CHAR_LENGTH_NINENINE )
                {
                    last;
                }
            }
            $relevant_head =~ s/^(\s)+//gis;
            $relevant_head =~ s/(\s)+$//gis;
            $relevanttext =~ s/^(\s)+//gis;
            $relevanttext =~ s/(\s)+$//gis;

            if ( length $relevant_head > 0 ) {
                $relevant_head = q(<head>) . $relevant_head . q(</head>);
            }
        }
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    @extract_ids = ();

    if ( $wasrelevancefound eq q(0) ) {
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    if (   $text_to_scan =~ m/<\/h[1-6]>/is
        && $text_to_scan !~ m/<h[1-6]>.*<\/h[1-6]>/is )
    {
        if ( $text_to_scan =~ m/<\/h([1-6])>/is ) {
            $text_to_scan = q(<h) . $1 . q(>) . $text_to_scan;
        }
    }
    ##if ( $text_to_scan =~ m/(<h[1-6]>(.*)<\/h[1-6]>)(.*)(<h[1-6]>)(.*)/gis )
    ##{
    ##    $text_to_scan = $1 . $3;
    ##}

    my $idstream = HTML::TokeParser->new( \$text_backup )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $idtoken = $idstream->get_token() ) {
        if (   ( $idtoken->[0] eq 'S' && $idtoken->[1] eq 'a' )
            || ( $idtoken->[2]{'id'} && length $idtoken->[2]{'id'} > 0 ) )
        {
            my $secid   = $idtoken->[2]{'id'};
            my $secname = $idtoken->[2]{'name'};

            if ( $secname && length $secname > 0 ) {
                $secid = $secname;
            }
            if ( $secid && length $secid > 0 ) {
                push @extract_ids, $secid;
            }
        }
        if (   $idtoken
            && $idtoken->[2]{'class'}
            && length $idtoken->[2]{'class'} > 0 )
        {
            my $sectval = $idtoken->[2]{'class'};
            if ( $sectval =~ m/^sect([\d])$/gis ) {
                $sect_found = $1;
            }
        }
    }
    if ( scalar @extract_ids < 1 ) {
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    if ( $text_to_scan =~ m/(.*?)(<h[1-6])(.*)/is ) {

        $text_before_head = $1;
        my $headtag = $2;
        $headlevel = $headtag;
        $headlevel =~ s/<h//gis;
        my $headclose = $headtag;
        $headclose =~ s/\</<\//gis;

        # extract the head
        $id_lines = $text_before_head;
        $text_to_scan =~ s/(.*)$headtag//is;

        # from the idLines extract all the IDs
        # done.
        # remove till the closing h
        my $indexresult = index $text_to_scan, '>';

        # check for classes
        $class = substr $text_to_scan, 0,
            $indexresult;    # might be like class=""
        $class =~ s/^\s//gis;
        $class =~ s/\s$//gis;

        if ( $class =~ m/class(\s)*=(\s)*(")*([^"]+)(")*/gis ) {
            $class = $4;
            $class =~ s/\"//gis;
            $class =~ s/^\s//gis;
            $class =~ s/\s$//gis;
        }

        $text_to_scan = substr $text_to_scan, $indexresult + 1;
        if ( $text_to_scan =~ m/(.*)$headclose>(.*)/is ) {
            $relevant_head = $1;
            $relevanttext  = $2;
            $relevant_head =~ s/^(\s)+//gis;
            $relevant_head =~ s/(\s)+$//gis;
            if ( length $relevant_head > 0 ) {
                $relevant_head  = q(<head>) . $relevant_head . q(</head>);
                $was_head_found = q(1);

                # now streamify to get the text out of section
                my $extractedplaintext = q();
                my $sectionstream = HTML::TokeParser->new( \$text_backup )
                    or carp "Couldn't parse: $OS_ERROR";
                while ( my $stoken = $sectionstream->get_token() ) {
                    if ( ( $stoken->[0] eq 'S' && $stoken->[1] eq 'img' ) ) {
                        ##$extractedplaintext .= $stoken->[2]{'alt'};
                    }
                    elsif ( ( $stoken->[0] eq 'S' && $stoken->[1] eq 'pre' ) )
                    {
                        $extractedplaintext
                            .= $stoken->[$SHORTCUT_CHAR_LENGTH_FOUR]
                            . $sectionstream->get_trimmed_text('/pre')
                            . q(</pre>);
                    }
                    else {
                        $extractedplaintext
                            .= q( ) . $sectionstream->get_trimmed_text();
                    }
                }
                if ( length $extractedplaintext > 0 ) {
                    $relevanttext = $extractedplaintext;
                    $relevanttext =~ s/(\s)+/ /gis;
                    $relevanttext =~ s/^\s//gis;
                    $relevanttext =~ s/\s$//gis;
                    $was_section_found = q(1);
                }
            }
        }
    }
    if (   $was_head_found eq q(1)
        && $was_section_found eq q(1)
        && ( scalar @extract_ids > 0 ) )
    {
        my $headstream = HTML::TokeParser->new( \$relevant_head )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $htoken = $headstream->get_token() ) {
            if (( $htoken->[0] eq 'S' || $htoken->[0] eq 'E' )
                && (   $htoken->[1] eq 'a'
                    || $htoken->[1] eq 'dt'
                    || $htoken->[1] eq 'p'
                    || $htoken->[1] eq 'dd'
                    || $htoken->[1] eq 'th'
                    || $htoken->[1] eq 'td'
                    || $htoken->[1] eq 'tr' )
                )
            {
                if ( $htoken->[$SHORTCUT_CHAR_LENGTH_FOUR] ) {
                    my $atag = $htoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
                    $relevant_head =~ s/\Q$atag\E//gis;
                }
            }
        }
        $relevant_head = removeunclosedtags($relevant_head);
        $relevant_head =~ s/\r//gis;
        $relevant_head =~ s/\n//gis;
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    #############################################################################
    # if I reach here, I need to see for other ways
    # of grabbing the head and section.
    # caption, dt, th
    #############################################################################
    my @boldheads = ( 'caption', 'dt' );
    $text_to_scan      = $text_backup;
    $was_head_found    = q();
    $relevant_head     = q();
    $relevanttext      = q();
    $was_section_found = q();

    if (   $text_to_scan =~ m/<\/dt>/gis
        && $text_to_scan !~ m/<dt>.*<\/dt>/gis )
    {
        $text_to_scan = q(<dt>) . $text_to_scan;
    }
    if (   $text_to_scan =~ m/<\/tr>/gis
        && $text_to_scan !~ m/<tr>.*<\/tr>/gis )
    {
        $text_to_scan = q(<tr>) . $text_to_scan;
    }

    if ( $text_to_scan =~ m/(<dt>(.*)<\/dt>)(.*)(<dt>)(.*)/gis ) {
        $text_to_scan = $1 . $3;
    }
    if ( $text_to_scan =~ m/(<tr>(.*)<\/tr>)(.*)(<tr>)(.*)/gis ) {
        $text_to_scan = $1 . $3;
    }

    # create the token stream

    my $idstreamm = HTML::TokeParser->new( \$text_to_scan )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $idtoken = $idstreamm->get_token() ) {
        if ($idtoken->[0] eq 'S'
            && (   $idtoken->[1] eq 'caption'
                || $idtoken->[1] eq 'dt'
                || $idtoken->[1] eq 'tr' )
            )
        {
            my $headtoken = $idtoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
            my $headtag   = $idtoken->[1];
            #print "Pass i am in : $headtoken; $headtag\n"; 
            if ( $text_to_scan =~ m/(.*)$headtoken(.*)/is ) {
                my $indexresult = index $text_to_scan, $headtoken;
                $id_lines = substr $text_to_scan, 0, $indexresult;
                $relevant_head = substr $text_to_scan,
                    $indexresult + ( length $headtoken ),
                    length $text_to_scan;

                $indexresult = index lc ($relevant_head), '</' . $headtag . '>';
                my $rhead = substr $relevant_head, 0, $indexresult;
                if ($indexresult >= 0){
                    $relevanttext = substr $relevant_head,
                        $indexresult
                        + ( length ($headtag) + $SHORTCUT_CHAR_LENGTH ),
                        length ($relevant_head);
                }
                $relevant_head = $rhead;
                $relevant_head =~ s/^(\s)+//gis;
                $relevant_head =~ s/(\s)+$//gis;
                if ( length $relevant_head > 0 ) {
                    $relevant_head  = q(<head>) . $relevant_head . q(</head>);
                    $was_head_found = q(1);
                    if ( length $relevanttext > 0 ) {
                        my $extractedplaintext = q();
                        my $sectionstream
                            = HTML::TokeParser->new( \$text_backup )
                            or carp "Couldn't parse: $OS_ERROR";
                        while ( my $stoken = $sectionstream->get_token() ) {
                            if ((      $stoken->[0] eq 'S'
                                    && $stoken->[1] eq 'img'
                                )
                                )
                            {
                                ##$extractedplaintext .= $stoken->[2]{'alt'};
                            }
                            elsif (
                                (      $stoken->[0] eq 'S'
                                    && $stoken->[1] eq 'pre'
                                )
                                )
                            {
                                $extractedplaintext
                                    .= $stoken->[$SHORTCUT_CHAR_LENGTH_FOUR]
                                    . $sectionstream->get_trimmed_text('/pre')
                                    . q(</pre>);
                            }
                            else {
                                $extractedplaintext .= q( )
                                    . $sectionstream->get_trimmed_text();
                            }
                        }
                        if ( length $extractedplaintext > 0 ) {
                            $relevanttext = $extractedplaintext;
                            $relevanttext =~ s/(\s)+/ /gis;
                            $relevanttext =~ s/^\s//gis;
                            $relevanttext =~ s/\s$//gis;
                            $was_section_found = q(1);
                        }
                    }
                }
            }
        }
        if ( $was_head_found eq q(1) && $was_section_found eq q(1) ) {
            last;
        }
    }
    ##print "Pass IDENTIFIVXATION: $was_head_found ($relevant_head):$was_section_found:".(scalar @extract_ids)."\n";
    if (   $was_head_found eq q(1)
        && $was_section_found eq q(1)
        && ( scalar @extract_ids > 0 ) )
    {
        my $headstream = HTML::TokeParser->new( \$relevant_head )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $htoken = $headstream->get_token() ) {
            if (( $htoken->[0] eq 'S' || $htoken->[0] eq 'E' )
                && (   $htoken->[1] eq 'a'
                    || $htoken->[1] eq 'dt'
                    || $htoken->[1] eq 'p'
                    || $htoken->[1] eq 'dd'
                    || $htoken->[1] eq 'th'
                    || $htoken->[1] eq 'td'
                    || $htoken->[1] eq 'tr' )
                )
            {
                my $atag = $htoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
                if ($atag) {$relevant_head =~ s/\Q$atag\E//gis;}
            }
        }
        $relevant_head = removeunclosedtags($relevant_head);
        $relevant_head =~ s/\r//gis;
        $relevant_head =~ s/\n//gis;
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    #############################################################################
    # if I reach here, I need to see for other ways
    # of grabbing the head and section
    #############################################################################
    $text_to_scan      = $text_backup;
    $was_head_found    = q();
    $relevant_head     = q();
    $relevanttext      = q();
    $was_section_found = q();

    # create the token stream
    my $idstreamn = HTML::TokeParser->new( \$text_to_scan )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $idtoken = $idstreamn->get_token() ) {

        #if(($idtoken->[0] eq 'S' && $idtoken->[1] eq 'p' )
        if ( ( $idtoken->[0] eq 'S' ) ) {
            if ( $idtoken->[2]{'class'}
                && any { $_ eq $idtoken->[2]{'class'} } @headclasses )
            {
                $class = $idtoken->[2]{'class'};
                my $headtoken        = $idtoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
                my $tagbeinganalysed = $idtoken->[1];
                if ( $text_to_scan =~ m/.*$headtoken.*/ ) {
                    my $indexresult = index $text_to_scan, $headtoken;
                    $id_lines = substr $text_backup, 0, $indexresult;
                    $relevant_head = substr $text_backup,
                        $indexresult + ( length $headtoken ),
                        length $text_backup;
                    #$id_lines = $1;
                    #$relevant_head = $2;

                    $indexresult = index $relevant_head,
                        '</' . $tagbeinganalysed . '>';
                    my $rhead = substr $relevant_head, 0, $indexresult;
                    if ($indexresult > 0) {
                        $relevanttext = substr $relevant_head, ($indexresult + length ($tagbeinganalysed) + $SHORTCUT_CHAR_LENGTH), length $relevant_head;
                    }
                    $relevant_head = $rhead;
                    $relevant_head =~ s/^(\s)+//gis;
                    $relevant_head =~ s/(\s)+$//gis;
                    if ( length $relevant_head > 0 ) {
                        $relevant_head
                            = q(<head>) . $relevant_head . q(</head>);
                        $was_head_found = q(1);
                        if ( length $relevanttext > 0 ) {
                            my $extractedplaintext = q();
                            my $sectionstream
                                = HTML::TokeParser->new( \$text_backup )
                                or carp "Couldn't parse: $OS_ERROR";
                            while ( my $stoken = $sectionstream->get_token() )
                            {
                                if ((      $stoken->[0] eq 'S'
                                        && $stoken->[1] eq 'img'
                                    )
                                    )
                                {
                                    ## $extractedplaintext .= $stoken->[2]{'alt'};
                                }
                                elsif (
                                    (      $stoken->[0] eq 'S'
                                        && $stoken->[1] eq 'pre'
                                    )
                                    )
                                {
                                    $extractedplaintext
                                        .= $stoken
                                        ->[$SHORTCUT_CHAR_LENGTH_FOUR]
                                        . $sectionstream->get_trimmed_text(
                                        '/pre')
                                        . q(</pre>);
                                }
                                else {
                                    $extractedplaintext .= q( )
                                        . $sectionstream->get_trimmed_text();
                                }
                            }
                            if ( length $extractedplaintext > 0 ) {
                                $relevanttext = $extractedplaintext;
                                $relevanttext =~ s/(\s)+/ /gis;
                                $relevanttext =~ s/^\s//gis;
                                $relevanttext =~ s/\s$//gis;
                                $was_section_found = q(1);
                            }
                        }
                    }

                }
            }
        }

        if ( $was_head_found eq q(1) && $was_section_found eq q(1) ) {
            last;
        }
    }

    if (   $was_head_found eq q(1)
        && $was_section_found eq q(1)
        && ( scalar @extract_ids > 0 ) )
    {
        my $headstream = HTML::TokeParser->new( \$relevant_head )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $htoken = $headstream->get_token() ) {
            if (( $htoken->[0] eq 'S' || $htoken->[0] eq 'E' )
                && (   $htoken->[1] eq 'a'
                    || $htoken->[1] eq 'dt'
                    || $htoken->[1] eq 'p'
                    || $htoken->[1] eq 'dd'
                    || $htoken->[1] eq 'th'
                    || $htoken->[1] eq 'td'
                    || $htoken->[1] eq 'tr' )
                )
            {
                my $atag = $htoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
                if ($atag) {
                    $relevant_head =~ s/\Q$atag\E//gis;
                }
            }
        }
        $relevant_head = removeunclosedtags($relevant_head);
        $relevant_head =~ s/\r//gis;
        $relevant_head =~ s/\n//gis;
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }

    #############################################################################
    # if I reach here, I need to see for other ways
    # of grabbing the head and section.
    # a title
    #############################################################################
    $text_to_scan      = $text_backup;
    $was_head_found    = q();
    $relevant_head     = q();
    $relevanttext      = q();
    $was_section_found = q();

    # create the token stream
    my $idstreamg = HTML::TokeParser->new( \$text_to_scan )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $idtoken = $idstreamg->get_token() ) {
        if ($idtoken->[0] eq 'S'
            && (   $idtoken->[1] eq 'a'
                && $idtoken->[2]{'title'}
                && length $idtoken->[2]{'title'} > 0 )
            )
        {
            my $headtoken = $idtoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
            my $headtag   = $idtoken->[1];
            if ( $text_to_scan =~ m/(.*)\Q$headtoken\E(.*)/gis ) {

                $id_lines      = $1 . $headtoken;
                $relevanttext  = $2;
                $relevant_head = $idtoken->[2]{'title'};
                $relevant_head =~ s/^(\s)+//gis;
                $relevant_head =~ s/(\s)+$//gis;

                if ( length $relevant_head > 0 ) {
                    $relevant_head  = q(<head>) . $relevant_head . q(</head>);
                    $was_head_found = q(1);
                    if ( length $relevanttext > 0 ) {
                        my $extractedplaintext = q();
                        my $sectionstream
                            = HTML::TokeParser->new( \$text_backup )
                            or carp "Couldn't parse: $OS_ERROR";
                        while ( my $stoken = $sectionstream->get_token() ) {
                            if ((      $stoken->[0] eq 'S'
                                    && $stoken->[1] eq 'img'
                                )
                                )
                            {
                                ##$extractedplaintext .= $stoken->[2]{'alt'};
                            }
                            elsif (
                                (      $stoken->[0] eq 'S'
                                    && $stoken->[1] eq 'pre'
                                )
                                )
                            {
                                $extractedplaintext
                                    .= $stoken->[$SHORTCUT_CHAR_LENGTH_FOUR]
                                    . $sectionstream->get_trimmed_text('/pre')
                                    . q(</pre>);
                            }
                            else {
                                $extractedplaintext .= q( )
                                    . $sectionstream->get_trimmed_text();
                            }
                        }
                        if ( length $extractedplaintext > 0 ) {
                            $relevanttext = $extractedplaintext;
                            $relevanttext =~ s/(\s)+/ /gis;
                            $relevanttext =~ s/^\s//gis;
                            $relevanttext =~ s/\s$//gis;
                            $was_section_found = q(1);
                        }
                    }
                }
            }
        }
        if ( $was_head_found eq q(1) && $was_section_found eq q(1) ) {
            last;
        }
    }

    if (   $was_head_found eq q(1)
        && $was_section_found eq q(1)
        && ( scalar @extract_ids > 0 ) )
    {
        my $headstream = HTML::TokeParser->new( \$relevant_head )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $htoken = $headstream->get_token() ) {
            if (( $htoken->[0] eq 'S' || $htoken->[0] eq 'E' )
                && (   $htoken->[1] eq 'a'
                    || $htoken->[1] eq 'p'
                    || $htoken->[1] eq 'dt'
                    || $htoken->[1] eq 'dd'
                    || $htoken->[1] eq 'th'
                    || $htoken->[1] eq 'td'
                    || $htoken->[1] eq 'tr' )
                )
            {
                my $atag = $htoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
                $relevant_head =~ s/\Q$atag\E//gis;
            }
        }
        $relevant_head = removeunclosedtags($relevant_head);
        $relevant_head =~ s/\r//gis;
        $relevant_head =~ s/\n//gis;
        return ( $relevant_head, $relevanttext, $class, $headlevel,
            @extract_ids );
    }
}

##############################################################################
# Function: get_apt_id
# Parameters:
#
# Description: Determines the best ID to use for chunk name
##############################################################################

sub get_apt_id {
    my @idlist = @_;
    my $testid = q();
    foreach my $inid (@idlist) {
        if ( is_xlink($inid) ) {
            $testid = $inid;
        }
    }
    if ( length $testid < 1 ) {
        foreach my $inid (@idlist) {
            if ( $inid =~ m/([[:upper:]])+/gsxm ) {
                $testid = $inid;
            }
        }
        if ( length $testid < 1 ) {
            $testid = $idlist[ scalar @idlist - 1 ];
        }
    }
    return $testid;
}

##############################################################################
# Function: remove_duplicate
# Parameters:
#
# Description: removes the duplicates from a collection
##############################################################################

sub remove_duplicate {
    my @idlist  = @_;
    my @newlist = ();
    foreach my $checkid (@idlist) {
        if ( !any { $_ eq $checkid } @newlist ) {
            push @newlist, $checkid;
        }
    }
    undef @idlist;
    return @newlist;
}

##############################################################################
# Function: grab
# Parameters:
#
# Description: Grab the relevant section
##############################################################################

sub grab {
    my $unformattedlines = shift;
    $unformattedlines =~ s/\r//gis;
    $unformattedlines =~ s/\n/######vivek####/gis;
    if ( $unformattedlines =~ m/<a(\s)+/gis ) {
        $unformattedlines =~ s/(<a(\s)+)/######vivek####$1/gis;
    }
    $unformattedlines =~ s/(\n)+/\n/gis;
    my @cleanedlinestemp = split /######vivek####/, $unformattedlines;
    my $headfoundat      = $INITIAL;
    my $headwas          = q();
    my $idfoundat        = $INITIAL;
    my $idwas            = q();
    for (
        my $linenumber = 0;
        $linenumber < scalar @cleanedlinestemp;
        $linenumber++
        )
    {
        my $idstreamd
            = HTML::TokeParser->new( \$cleanedlinestemp[$linenumber] )
            or carp "Couldn't parse: $OS_ERROR";
        while ( my $token = $idstreamd->get_token() ) {
            if ($token->[0] eq 'S'
                && (   ( any { $_ eq $token->[2]{'class'} } @headclasses )
                    || ( $token->[1] =~ m/h[1-6]/isxm )
                    || ( $token->[1] =~ m/dt/gis )
                    || ( $token->[1] =~ m/caption/isxm ) )
                )
            {
                $headfoundat = $linenumber;
                $headwas     = $token->[$SHORTCUT_CHAR_LENGTH_FOUR];
            }
            elsif (
                (      $token->[0] eq 'S'
                    && $token->[1] eq 'a'
                    && length $token->[2]{'id'} > 0
                )
                )
            {
                $idwas     = $token->[$SHORTCUT_CHAR_LENGTH_FOUR];
                $idfoundat = $linenumber;
                if (   $idfoundat > $headfoundat
                    && $headfoundat > $INITIAL
                    && $idfoundat > $INITIAL )
                {
                    $cleanedlinestemp[$headfoundat]
                        =~ s/$headwas/$idwas$headwas/gis;
                    $cleanedlinestemp[$linenumber] =~ s/$idwas//gis;
                }
            }
        }
    }
    undef $unformattedlines;
    return @cleanedlinestemp;
}

##############################################################################
# Function: arrange
# Parameters:
#
# Description: Arranges the HTML before the processing starts
# so that all the relevant lines are on one line
##############################################################################

sub arrange {
    my $unformattedlines = shift;
    my $whatfile         = shift;
    $unformattedlines =~ s/\r//gis;

    # get the chapter title
    my $idstreamv = HTML::TokeParser->new( \$unformattedlines )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $tokenh = $idstreamv->get_token() ) {

        if ( ( $tokenh->[0] eq 'S' ) && ( $tokenh->[1] =~ m/title/isxm ) ) {
            $chaptertitle = $idstreamv->get_trimmed_text('/title');
            last;
        }
        if ( ( $tokenh->[0] eq 'S' ) && ( $tokenh->[1] =~ m/h1/isxm ) ) {
            $chaptertitle = $idstreamv->get_trimmed_text('/h1');
            last;
        }
    }

    # if the title.htm was there, then its title is the book title
    if ( $whatfile =~ m/title.htm$/gis && length $chaptertitle > 0 ) {
        my $testpn = &extractpartnumber($whatfile);
        if ( length $testpn > 0 ) {
            if ( !exists $booktitle{$testpn} ) {
                $booktitle{$testpn} = $chaptertitle;
            }
        }
    }

   if ($chaptertitle =~ m/index/i){
         $indexes_list {$whatfile} = "";
   }

    $unformattedlines =~ s/<head>(.*)<\/head>//gis;
    # GLOBALLY REPLACE STARTING TAG WITH A NEW LINE

    $unformattedlines =~ s/(<[[:alpha:]])/\n$1/gis;
    # bring all dt and dd in one line
    #pos( $unformattedlines ) = 0;
    #print "Start\n";
    if ($javadoc ne q(1)) {
        my $tempbuffer = $unformattedlines; 
        while ($unformattedlines =~ m/(<dl([^>]*)>(.*?)<\/dl>)/gisxm) {
            my $exact = $1;
            my $newone = $exact;
            $newone =~ s/\r//gis;
            $newone =~ s/\n//gis;
            $newone =~ s/(<dt)/\n$1/gis;
            $newone =~ s/(<dl)/\n$1/gis;
            $newone =~ s/(<\/dl)/\n$1/gis;
            $tempbuffer =~ s/\Q$exact\E/$newone/gis;
        }
        $unformattedlines = $tempbuffer;
        undef $tempbuffer;
    }
    $unformattedlines =~ s/\r//gis;
    my @cleanedlinestemp = split /\n/, $unformattedlines;
    undef $idstreamv;
    my @newcleaned = q();
    foreach (@cleanedlinestemp) {
        my $cl = $_;
        $cl =~ s/\n/ /gis;
        if ( length $cl > 0 ) {
            push @newcleaned, $cl;
        }
    }
    @cleanedlinestemp = ();
    undef $unformattedlines;

    return @newcleaned;
}

##############################################################################
# Function: removeunclosedtags
# Parameters:
#
# Description:Rremove unclosed HTML tags
##############################################################################

sub removeunclosedtags {
    my $dirtytext = shift;
    my $idstreamb = HTML::TokeParser->new( \$dirtytext )
        or carp "Couldn't parse: $OS_ERROR";
    while ( my $dtoken = $idstreamb->get_token() ) {
        if ( $dtoken->[0] eq 'S' ) {
            my $atag      = $dtoken->[$SHORTCUT_CHAR_LENGTH_FOUR];
            my $close_tag = q(</) . $dtoken->[1] . q(>);
            if (   $dirtytext =~ m/\Q$atag\E/isxm
                && $dirtytext !~ m/\Q$close_tag\E/isxm )
            {
                $dirtytext =~ s/\Q$atag\E//isxm;
            }
        }
        elsif ( $dtoken->[0] eq 'E' ) {
            my $atag    = q(</) . $dtoken->[1] . q(>);
            my $opentag = q(<) . $dtoken->[1];
            if (   $dirtytext =~ m/\Q$atag\E/isxm
                && $dirtytext !~ m/\Q$opentag\E/isxm )
            {
                $dirtytext =~ s/\Q$atag\E//isxm;
            }
        }
    }
    return $dirtytext;
}

##############################################################################
# Function: removehtmlmarkup
# Parameters:
#
# Description:Remove html markup
##############################################################################

sub removehtmlmarkup {
    my $dirtytext = shift;
    my $idstreame = HTML::TokeParser->new( \$dirtytext )
        or carp "Couldn't parse: $OS_ERROR";
    my $plaintext = q();
    while ( my $setoken = $idstreame->get_token() ) {
        $plaintext .= q( ) . $idstreame->get_trimmed_text();
    }
    $plaintext =~ s/(\s)+/ /gis;
    $plaintext =~ s/(\s)+$//gis;
    $plaintext =~ s/^(\s)+//gis;
    return $plaintext;
}

##############################################################################
# Function: removehtmlmarkupbody
# Parameters:
#
# Description: Remove html markup from body
##############################################################################

sub removehtmlmarkupbody {
    my $dirtytext = shift;
    my $idstreamw = HTML::TokeParser->new( \$dirtytext )
        or carp "Couldn't parse: $OS_ERROR";
    my $plaintext = q();
    while ( my $setoken = $idstreamw->get_token() ) {
        $plaintext .= q( ) . $idstreamw->get_trimmed_text();
    }
    if ( length $plaintext < 2 ) {
        $idstreamw = HTML::TokeParser->new( \$dirtytext )
            or carp "Couldn't parse: $OS_ERROR";
        $plaintext = $idstreamw->get_trimmed_text();
    }
    $plaintext =~ s/(\s)+/ /gis;
    $plaintext =~ s/(\s)+$//gis;
    $plaintext =~ s/^(\s)+//gis;
    return $plaintext;
}

##############################################################################
# Function: removehtmlmarkupbodyall
# Parameters:
#
# Description: Remove all html markups from body
##############################################################################

sub removehtmlmarkupbodyall {
    my $dirtytext = shift;
    $dirtytext =~ s/<([^>]+)>//gis;
    return $dirtytext;
}

##############################################################################
# Function: extractpartnumber
# Parameters:
#
# Description: Extract pn or identifier like Doc ID
##############################################################################

sub extractpartnumber {
    my $take_path   = shift;
    my $extractedpn = q();
    ##print "PATH: $take_path\n";
    if ( $take_path =~ m/([[:alpha:]](\d){5}([_-](\d){1,2})*)\//gis ) {
        $extractedpn = $1;
        $extractedpn = lc $extractedpn;
        $extractedpn =~ s/_(.*)//gis;
        $extractedpn =~ s/-(.*)//gis;
    }
    elsif ( $take_path =~ m/(([[:upper:]]){5})\//gs ) {
        $extractedpn = $1;
    }
    elsif ( $take_path =~ m/(([0-9]+)-([0-9]+))\//gis ) {
        $extractedpn = $1;
    }
    else {
      my @pathparts = split (/\//, $take_path);
      foreach my $ppaths (@pathparts) {
        if ($ppaths =~ m/^[a-z]{4}$/){
          $extractedpn = $ppaths;
        }
      }
    }

    if (length $extractedpn < 1) {
        $extractedpn = q(generic);
    }
    return $extractedpn;
}

##############################################################################
# Function: erroradd
# Parameters:
#
# Description: Aggregate error lines
##############################################################################

sub erroradd {
    my $el = shift;
    push @errorlines, $el;
    return;
}

##############################################################################
# Function: writejsonchunk
# Parameters:
#
# Description: Create a file with JSON values
##############################################################################

sub writejsonchunk {
    my $id_json      = shift;
    my $title_json   = shift;
    my $summary_json = shift;
    my $url_json     = shift;
    my $class_passed = shift;
    my $cpth         = shift;
    my $ctitle       = $chaptertitle;
    $thisfilechunkcount++;
    $id_json =~ s/JUSTANID//i;
    if ( length $context < 1 ) { $context = q(all); }
    $url_json =~ s/((.*)[.]ht(m|ml))-(.*)/$1/gis;
    if ( $url_json =~ /^[.][\/]/sxm ) {
        $url_json =~ s/[.]\///sxm;
    }
    $url_json =~ s/^\///sxm;
    #$title_json =~ s/^([.0-9]+)(\s)+//is;
    $indexcounter++;

    my $folderpn = q();
    my $testpn   = extractpartnumber($url_json);
    if ($testpn eq q(generic)) {
        my $pnpath = $url_json;
        $pnpath =~ s/(.*)\/(.*)/$1/gis;
        if (exists $folderpnmapping{$pnpath}) {
          $folderpn = $folderpnmapping{$pnpath};
          ##print "Got $folderpn\n";
        }
    }
    elastic_search_mapping($testpn);
    my $book_url = $url_json;
    if ( $book_url =~ m/(.*)\/(.*)/sxm ) {
        $book_url =~ s/(.*)\/(.*)/$1\//gis;
        if ( -f $book_url . q(toc.htm) ) {
            $book_url = $book_url . q(toc.htm);
        }
        elsif ( -f $book_url . q(package-summary.html) ) {
            $book_url = $book_url . q(package-summary.html);
        }
        elsif ( -f $book_url . q(index.htm) ) {
            $book_url = $book_url . q(package-summary.html);
        }
        elsif ( exists $folderpnmapping{$book_url} && -f $book_url . q(index.html) ) {
            $book_url = $book_url . q(index.html);
        }
        else {
            $book_url = $url_json;
        }
    }

    my $json_title_book   = q();
    my $json_release_book = q();
    my $complete_pn       = q();

    if (length $folderpn > 0 && exists $library_data{$folderpn}) {
        $testpn = $folderpn;
        ##print "folderpn: $folderpn\n";
        $json_title_book = $library_data{$folderpn}[0];
        $json_release_book
            = $library_data{$folderpn}[$SHORTCUT_CHAR_LENGTH_FIVE];
        $complete_pn = $library_data{$folderpn}[$SHORTCUT_CHAR_LENGTH_SEVEN];
    }
    elsif ( exists $library_data_docid{$testpn} ) {
        ##print "Found in 1\n";
        $json_title_book = $library_data_docid{$testpn}[0];
        ##$testpn = $library_data_docid{$testpn}[$SHORTCUT_CHAR_LENGTH_SEVEN];
        $json_release_book
            = $library_data_docid{$testpn}[$SHORTCUT_CHAR_LENGTH_FIVE];
        $complete_pn = $library_data_docid{$testpn}[$SHORTCUT_CHAR_LENGTH_SEVEN];

    }
    elsif ( exists $library_data{$testpn} ) {
        ##print "Found in 2\n";
        $json_title_book = $library_data{$testpn}[0];
        ##$testpn = $library_data{$testpn}[$SHORTCUT_CHAR_LENGTH_SEVEN];
        $json_release_book
            = $library_data{$testpn}[$SHORTCUT_CHAR_LENGTH_FIVE];
        $complete_pn = $library_data{$testpn}[$SHORTCUT_CHAR_LENGTH_SEVEN];
    }
    elsif ( exists $additional_data{$testpn} ) {
        ##print "Found in 3\n";
        $json_title_book = $additional_data{$testpn}[2];
        $complete_pn          = $additional_data{$testpn}[0];
        $json_release_book
            = $additional_data{$testpn}[$SHORTCUT_CHAR_LENGTH_FOUR];
    }
    else { $json_title_book = q(Not Available.); $complete_pn = q(); $json_release_book = q();}

    if (length $json_title_book < 1 && exists $additional_data{$testpn} && exists $additional_data{$testpn}[2]) {
        $json_title_book = $additional_data{$testpn}[2];
    }
    if (length $complete_pn < 1 && exists $additional_data{$testpn} && exists $additional_data{$testpn}[0]) {
        $complete_pn          = $additional_data{$testpn}[0];
    }
    if (length $json_release_book < 1 && exists $additional_data{$testpn} && exists $additional_data{$testpn}[$SHORTCUT_CHAR_LENGTH_FOUR]) {
        $json_release_book          = $additional_data{$testpn}[$SHORTCUT_CHAR_LENGTH_FOUR] ;
    }
    my $index_container = $url_json;
    $index_container =~ s/(.*)\/(.*)/$2/gsxm;


    #$url_json = $libraryurl.$url_json;

    if ( length $book_url > 0 ) {
        #$book_url = $libraryurl . $book_url;
        #$book_url = $libraryurl . $book_url;
    }
    my $ul_json
        = q(http://www.oracle.com/pls/topic/lookup?ctx=)
        . $context . q(&id=)
        . $id_json;
    if (length $id_json < 1) {
        $ul_json
          = q(http://www.oracle.com/pls/topic/lookup?ctx=)
          . $context;
    }

    my $type_to_take = $testpn;
    if ( $type_to_take =~ m/([[:alpha:]](\d){5}([_-](\d){1,2})*)\//sxm ) {
       $type_to_take = lc $type_to_take;
    }
    $type_to_take =~ s/_(.*)//sxm;
    $type_to_take =~ s/-(.*)//sxm;

    if ( length $title_json > 0 && length $summary_json > 0 ) {
        my $json = q();
        if (length $id_json > 0) {
          $json
            = q(curl -XPUT 'http://)
            . $server_elastic_search
            . q(:9200/)
            . $context . q(/)
            . $type_to_take . q(/)
            . $indexcounter . q(-)
            . $index_container . q(#)
            . $id_json
            . "/_create' -d '{\n";
        }
        else {
          $json
            = q(curl -XPUT 'http://)
            . $server_elastic_search
            . q(:9200/)
            . $context . q(/)
            . $type_to_take . q(/)
            . $indexcounter . q(-)
            . $index_container
            . "/_create' -d '{\n";
        }
        $json_release_lib =~ s/<i>//gis;
        $json_release_lib =~ s/<\/i>//gis;

        $title_json         = jsonify ($title_json);
        $summary_json       = jsonify ($summary_json);
        $json_title_book    = jsonify ($json_title_book);
        $json_release_lib   = jsonify ($json_release_lib);
        $file_lang          = jsonify ($file_lang);
        $ctitle             = jsonify ($ctitle);
        $file_release_date  = jsonify ($file_release_date);
        $file_description   = jsonify ($file_description);
        $file_keywords      = jsonify ($file_keywords);

        my $computed_topic_type = q();
        if ($javadoc eq q(1)) {
            $computed_topic_type = "API";
        }
        else {
           $computed_topic_type = &get_topic_type(encode_entities_html($title_json), $json_title_book, $url_json, $class_passed); 
        }

        my $url_json_value = $url_json;
        if (length $id_json > 0) {
            $url_json_value = $url_json_value . q(#) . $id_json;
        }

        my $sublibpn      = "";
        my $releasenumber = "";
        my $categoryname      = "";

        my $folderV = $url_json_value;
        $folderV =~ s/(.*)://gis;
        $folderV =~ s/"//gis;
        $folderV =~ s/^(\s)+//gis;
        $folderV =~ s/(\s)+$//gis;
        $folderV =~ s/\\/\//gis;
        # first check with the complete path
        if (length $folderV > 0) {
           my @paths = split(/[\/]/, $folderV);
           for (my $p = 0; $p < (scalar @paths); $p++) {
             my $constructpath = q();
             for (my $pin = 0; $pin < $p; $pin++) {
               $constructpath .= q(/).$paths[$pin];
             }
             $constructpath =~ s/^\///gis;
             if (exists $subfolder_mappings{$constructpath}) {
               $sublibpn      = $subfolder_mappings{$constructpath}{'PART_NUMBER'};
               $releasenumber = $subfolder_mappings{$constructpath}{'RELEASE_NUMBER'};
               $categoryname      = $subfolder_mappings{$constructpath}{'CATEGORY_NAME'};
               #last; 
            }
           }
        }

        if ((length $sublibpn) < 1) {
            if (exists $subfolder_mappings{$whichpn}){
              $sublibpn      = $subfolder_mappings{$whichpn}{'PART_NUMBER'};
              $releasenumber = $subfolder_mappings{$whichpn}{'RELEASE_NUMBER'};
              $categoryname      = $subfolder_mappings{$whichpn}{'CATEGORY_NAME'};
            }
            else {
              $sublibpn      = $whichpn;
              $releasenumber = "";
              $categoryname      = "";
            }
        }

        $categoryname =~ s/(\s)+//gis;
        $categoryname = lc ($categoryname);
        $json .= q("id": ") . $id_json . "\",\n";
        ##$json .= q("title": ") . encode_entities_html($title_json) . "\",\n";
        $json .= q("title": ") . $title_json . "\",\n";
        ##$json .= q("summary": "). encode_entities_html($summary_json) . "\",\n";
        $json .= q("summary": "). $summary_json . "\",\n";
        $json .= q("book_title": ") . $json_title_book . "\",\n";
        $json .= q("part_number": ") . $complete_pn . "\",\n";
        $json .= q("book_release": ") . $json_release_book . "\",\n";
        $json .= q("ul": ") . $ul_json . "\",\n";
        $json .= q("book_url": ") . $book_url . "\",\n";
        $json .= "\"topic_type\": \"".$computed_topic_type."\",\n";
        $json .= "\"roles\": \"".roles($cpth, $json_title_book, part_number, $url_json_value)."\",\n";
        $json .= q("lang": ") . $file_lang . "\",\n";
        ##$json .= q("page_title": ") . encode_entities_html($ctitle) . "\",\n";
        $json .= q("page_title": ") . $ctitle . "\",\n";
        $json .= q("release_date": ") . $file_release_date . "\",\n";
        $json .= q("formats": ") . $book_formats . "\",\n";
        $json .= q("libraryurl": ") . $libraryurl . "\",\n";
        $json .= q("url": ") . $url_json_value . "\",\n";
        $json .= q("description": ") . $file_description . "\",\n";
        $json .= q("mainpn": ") . $whichpn . "\",\n";
        $json .= q("librarypn": ") . $sublibpn . "\",\n";
        $json .= q("category_name": ") . $categoryname . "\",\n";
        $json .= q("release_number": ") . standardize_release_number($releasenumber) . "\",\n";
        $json .= q("keywords": ") . $file_keywords . "\"\n";
        $json .= "}'\n";

        open my $CHUNK, '>>',
            $jsonoutputfolder . q(/) . $library_partnumber . q(.txt)
            or croak "Cannot open for reading: $OS_ERROR ".$jsonoutputfolder . q(/) . $library_partnumber . q(.txt);
        binmode $CHUNK, ':encoding(UTF-8)';
        #binmode $CHUNK;
        print {$CHUNK} $json . "\n\n";
        print {$CHUNK} q(VIVEK-SAMPLE) . "\n\n";
        close $CHUNK
            or carp "Can't close : $OS_ERROR";
        $summary_count++;
    }
    return;
}
##############################################################################
#
# Function: jsonify
#
# escapes to the valid json text
##############################################################################

sub jsonify {
    my $jsondata = shift;
    $jsondata =~ s/^(\r)+//gis;
    $jsondata =~ s/^(\n)+//gis;
    $jsondata =~ s/(\r)+$//gis;
    $jsondata =~ s/(\n)+$//gis;
    $jsondata =~ s/(\r)*//gis;
    $jsondata =~ s/(\n)+/ /gis;
    $jsondata =~ s/&nbsp;/ /gis;
    $jsondata =~ s/\\/\\\\/gis;
    $jsondata =~ s/"/\\"/gis;
    return $jsondata;
}

##############################################################################
#
# Function: htmlents2numeric
#
# Receives an XML entity. If the entity is a named HTML characters entity,
# returns the hexadecimal numeric character entity. Best used like this:
# $content =~ s{(&[A-Za-z0-9.]+;)}{htmlents2numeric($1)}xmsge;
#
##############################################################################

sub htmlents2numeric {
    my $c = shift;
    return '&amp;'  if ( $c eq '&amp;' );     # ampersand
    return '&gt;'   if ( $c eq '&gt;' );      # greater-than sign
    return '&lt;'   if ( $c eq '&lt;' );      # less-than sign
    return '&quot;' if ( $c eq '&quot;' );    # quotation mark = APL quote
    return '&#x00A0;'
        if ( $c eq '&nbsp;' );    # no-break space = non-breaking space
    return '&#x00A1;' if ( $c eq '&iexcl;' );     # inverted exclamation mark
    return '&#x00A2;' if ( $c eq '&cent;' );      # cent sign
    return '&#x00A3;' if ( $c eq '&pound;' );     # pound sign
    return '&#x00A4;' if ( $c eq '&curren;' );    # currency sign
    return '&#x00A5;' if ( $c eq '&yen;' );       # yen sign = yuan sign
    return '&#x00A6;'
        if ( $c eq '&brvbar;' );    # broken bar = broken vertical bar
    return '&#x00A7;' if ( $c eq '&sect;' );   # section sign
    return '&#x00A8;' if ( $c eq '&uml;' );    # diaeresis = spacing diaeresis
    return '&#x00A9;' if ( $c eq '&copy;' );   # copyright sign
    return '&#x00AA;' if ( $c eq '&ordf;' );   # feminine ordinal indicator
    return '&#x00AB;'
        if ( $c eq '&laquo;' )
        ; # left-pointing double angle quotation mark = left pointing guillemet
    return '&#x00AC;' if ( $c eq '&not;' );    # not sign
    return '&#x00AD;'
        if ( $c eq '&shy;' );    # soft hyphen = discretionary hyphen
    return '&#x00AE;'
        if ( $c eq '&reg;' );   # registered sign = registered trade mark sign
    return '&#x00AF;'
        if ( $c eq '&macr;' )
        ;    # macron = spacing macron = overline = APL overbar
    return '&#x00B0;' if ( $c eq '&deg;' );    # degree sign
    return '&#x00B1;'
        if ( $c eq '&plusmn;' );    # plus-minus sign = plus-or-minus sign
    return '&#x00B2;'
        if ( $c eq '&sup2;' )
        ;    # superscript two = superscript digit two = squared
    return '&#x00B3;'
        if ( $c eq '&sup3;' )
        ;    # superscript three = superscript digit three = cubed
    return '&#x00B4;' if ( $c eq '&acute;' );  # acute accent = spacing acute
    return '&#x00B5;' if ( $c eq '&micro;' );  # micro sign
    return '&#x00B6;' if ( $c eq '&para;' );   # pilcrow sign = paragraph sign
    return '&#x00B7;'
        if ( $c eq '&middot;' )
        ;    # middle dot = Georgian comma = Greek middle dot
    return '&#x00B8;' if ( $c eq '&cedil;' );    # cedilla = spacing cedilla
    return '&#x00B9;'
        if ( $c eq '&sup1;' );    # superscript one = superscript digit one
    return '&#x00BA;' if ( $c eq '&ordm;' );    # masculine ordinal indicator
    return '&#x00BB;'
        if ( $c eq '&raquo;' )
        ; # right-pointing double angle quotation mark = right pointing guillemet
    return '&#x00BC;'
        if ( $c eq '&frac14;' )
        ;    # vulgar fraction one quarter = fraction one quarter
    return '&#x00BD;'
        if ( $c eq '&frac12;' )
        ;    # vulgar fraction one half = fraction one half
    return '&#x00BE;'
        if ( $c eq '&frac34;' )
        ;    # vulgar fraction three quarters = fraction three quarters
    return '&#x00BF;'
        if ( $c eq '&iquest;' )
        ;    # inverted question mark = turned question mark
    return '&#x00C0;'
        if ( $c eq '&Agrave;' )
        ;   # latin capital letter A with grave = latin capital letter A grave
    return '&#x00C1;'
        if ( $c eq '&Aacute;' );    # latin capital letter A with acute
    return '&#x00C2;'
        if ( $c eq '&Acirc;' );     # latin capital letter A with circumflex
    return '&#x00C3;'
        if ( $c eq '&Atilde;' );    # latin capital letter A with tilde
    return '&#x00C4;'
        if ( $c eq '&Auml;' );      # latin capital letter A with diaeresis
    return '&#x00C5;'
        if ( $c eq '&Aring;' )
        ; # latin capital letter A with ring above = latin capital letter A ring
    return '&#x00C6;'
        if ( $c eq '&AElig;' )
        ;    # latin capital letter AE = latin capital ligature AE
    return '&#x00C7;'
        if ( $c eq '&Ccedil;' );    # latin capital letter C with cedilla
    return '&#x00C8;'
        if ( $c eq '&Egrave;' );    # latin capital letter E with grave
    return '&#x00C9;'
        if ( $c eq '&Eacute;' );    # latin capital letter E with acute
    return '&#x00CA;'
        if ( $c eq '&Ecirc;' );     # latin capital letter E with circumflex
    return '&#x00CB;'
        if ( $c eq '&Euml;' );      # latin capital letter E with diaeresis
    return '&#x00CC;'
        if ( $c eq '&Igrave;' );    # latin capital letter I with grave
    return '&#x00CD;'
        if ( $c eq '&Iacute;' );    # latin capital letter I with acute
    return '&#x00CE;'
        if ( $c eq '&Icirc;' );     # latin capital letter I with circumflex
    return '&#x00CF;'
        if ( $c eq '&Iuml;' );      # latin capital letter I with diaeresis
    return '&#x00D0;' if ( $c eq '&ETH;' );    # latin capital letter ETH
    return '&#x00D1;'
        if ( $c eq '&Ntilde;' );    # latin capital letter N with tilde
    return '&#x00D2;'
        if ( $c eq '&Ograve;' );    # latin capital letter O with grave
    return '&#x00D3;'
        if ( $c eq '&Oacute;' );    # latin capital letter O with acute
    return '&#x00D4;'
        if ( $c eq '&Ocirc;' );     # latin capital letter O with circumflex
    return '&#x00D5;'
        if ( $c eq '&Otilde;' );    # latin capital letter O with tilde
    return '&#x00D6;'
        if ( $c eq '&Ouml;' );      # latin capital letter O with diaeresis
    return '&#x00D7;' if ( $c eq '&times;' );    # multiplication sign
    return '&#x00D8;'
        if ( $c eq '&Oslash;' )
        ;  # latin capital letter O with stroke = latin capital letter O slash
    return '&#x00D9;'
        if ( $c eq '&Ugrave;' );    # latin capital letter U with grave
    return '&#x00DA;'
        if ( $c eq '&Uacute;' );    # latin capital letter U with acute
    return '&#x00DB;'
        if ( $c eq '&Ucirc;' );     # latin capital letter U with circumflex
    return '&#x00DC;'
        if ( $c eq '&Uuml;' );      # latin capital letter U with diaeresis
    return '&#x00DD;'
        if ( $c eq '&Yacute;' );    # latin capital letter Y with acute
    return '&#x00DE;' if ( $c eq '&THORN;' );    # latin capital letter THORN
    return '&#x00DF;'
        if ( $c eq '&szlig;' );    # latin small letter sharp s = ess-zed
    return '&#x00E0;'
        if ( $c eq '&agrave;' )
        ;    # latin small letter a with grave = latin small letter a grave
    return '&#x00E1;'
        if ( $c eq '&aacute;' );    # latin small letter a with acute
    return '&#x00E2;'
        if ( $c eq '&acirc;' );     # latin small letter a with circumflex
    return '&#x00E3;'
        if ( $c eq '&atilde;' );    # latin small letter a with tilde
    return '&#x00E4;'
        if ( $c eq '&auml;' );      # latin small letter a with diaeresis
    return '&#x00E5;'
        if ( $c eq '&aring;' )
        ;   # latin small letter a with ring above = latin small letter a ring
    return '&#x00E6;'
        if ( $c eq '&aelig;' )
        ;    # latin small letter ae = latin small ligature ae
    return '&#x00E7;'
        if ( $c eq '&ccedil;' );    # latin small letter c with cedilla
    return '&#x00E8;'
        if ( $c eq '&egrave;' );    # latin small letter e with grave
    return '&#x00E9;'
        if ( $c eq '&eacute;' );    # latin small letter e with acute
    return '&#x00EA;'
        if ( $c eq '&ecirc;' );     # latin small letter e with circumflex
    return '&#x00EB;'
        if ( $c eq '&euml;' );      # latin small letter e with diaeresis
    return '&#x00EC;'
        if ( $c eq '&igrave;' );    # latin small letter i with grave
    return '&#x00ED;'
        if ( $c eq '&iacute;' );    # latin small letter i with acute
    return '&#x00EE;'
        if ( $c eq '&icirc;' );     # latin small letter i with circumflex
    return '&#x00EF;'
        if ( $c eq '&iuml;' );      # latin small letter i with diaeresis
    return '&#x00F0;' if ( $c eq '&eth;' );    # latin small letter eth
    return '&#x00F1;'
        if ( $c eq '&ntilde;' );    # latin small letter n with tilde
    return '&#x00F2;'
        if ( $c eq '&ograve;' );    # latin small letter o with grave
    return '&#x00F3;'
        if ( $c eq '&oacute;' );    # latin small letter o with acute
    return '&#x00F4;'
        if ( $c eq '&ocirc;' );     # latin small letter o with circumflex
    return '&#x00F5;'
        if ( $c eq '&otilde;' );    # latin small letter o with tilde
    return '&#x00F6;'
        if ( $c eq '&ouml;' );      # latin small letter o with diaeresis
    return '&#x00F7;' if ( $c eq '&divide;' );    # division sign
    return '&#x00F8;'
        if ( $c eq '&oslash;' )
        ;    # latin small letter o with stroke, = latin small letter o slash
    return '&#x00F9;'
        if ( $c eq '&ugrave;' );    # latin small letter u with grave
    return '&#x00FA;'
        if ( $c eq '&uacute;' );    # latin small letter u with acute
    return '&#x00FB;'
        if ( $c eq '&ucirc;' );     # latin small letter u with circumflex
    return '&#x00FC;'
        if ( $c eq '&uuml;' );      # latin small letter u with diaeresis
    return '&#x00FD;'
        if ( $c eq '&yacute;' );    # latin small letter y with acute
    return '&#x00FE;' if ( $c eq '&thorn;' );    # latin small letter thorn
    return '&#x00FF;'
        if ( $c eq '&yuml;' );    # latin small letter y with diaeresis
    return '&#x0152;' if ( $c eq '&OElig;' );    # latin capital ligature OE
    return '&#x0153;' if ( $c eq '&oelig;' );    # latin small ligature oe
    return '&#x0160;'
        if ( $c eq '&Scaron;' );    # latin capital letter S with caron
    return '&#x0161;'
        if ( $c eq '&scaron;' );    # latin small letter s with caron
    return '&#x0178;'
        if ( $c eq '&Yuml;' );      # latin capital letter Y with diaeresis
    return '&#x0192;'
        if ( $c eq '&fnof;' );   # latin small f with hook = function = florin
    return '&#x02C6;'
        if ( $c eq '&circ;' );    # modifier letter circumflex accent
    return '&#x02DC;' if ( $c eq '&tilde;' );   # small tilde
    return '&#x0391;' if ( $c eq '&Alpha;' );   # greek capital letter alpha
    return '&#x0392;' if ( $c eq '&Beta;' );    # greek capital letter beta
    return '&#x0393;' if ( $c eq '&Gamma;' );   # greek capital letter gamma
    return '&#x0394;' if ( $c eq '&Delta;' );   # greek capital letter delta
    return '&#x0395;' if ( $c eq '&Epsilon;' ); # greek capital letter epsilon
    return '&#x0396;' if ( $c eq '&Zeta;' );    # greek capital letter zeta
    return '&#x0397;' if ( $c eq '&Eta;' );     # greek capital letter eta
    return '&#x0398;' if ( $c eq '&Theta;' );   # greek capital letter theta
    return '&#x0399;' if ( $c eq '&Iota;' );    # greek capital letter iota
    return '&#x039A;' if ( $c eq '&Kappa;' );   # greek capital letter kappa
    return '&#x039B;' if ( $c eq '&Lambda;' );  # greek capital letter lambda
    return '&#x039C;' if ( $c eq '&Mu;' );      # greek capital letter mu
    return '&#x039D;' if ( $c eq '&Nu;' );      # greek capital letter nu
    return '&#x039E;' if ( $c eq '&Xi;' );      # greek capital letter xi
    return '&#x039F;' if ( $c eq '&Omicron;' ); # greek capital letter omicron
    return '&#x03A0;' if ( $c eq '&Pi;' );      # greek capital letter pi
    return '&#x03A1;' if ( $c eq '&Rho;' );     # greek capital letter rho
    return '&#x03A3;' if ( $c eq '&Sigma;' );   # greek capital letter sigma
    return '&#x03A4;' if ( $c eq '&Tau;' );     # greek capital letter tau
    return '&#x03A5;' if ( $c eq '&Upsilon;' ); # greek capital letter upsilon
    return '&#x03A6;' if ( $c eq '&Phi;' );     # greek capital letter phi
    return '&#x03A7;' if ( $c eq '&Chi;' );     # greek capital letter chi
    return '&#x03A8;' if ( $c eq '&Psi;' );     # greek capital letter psi
    return '&#x03A9;' if ( $c eq '&Omega;' );   # greek capital letter omega
    return '&#x03B1;' if ( $c eq '&alpha;' );   # greek small letter alpha
    return '&#x03B2;' if ( $c eq '&beta;' );    # greek small letter beta
    return '&#x03B3;' if ( $c eq '&gamma;' );   # greek small letter gamma
    return '&#x03B4;' if ( $c eq '&delta;' );   # greek small letter delta
    return '&#x03B5;' if ( $c eq '&epsilon;' ); # greek small letter epsilon
    return '&#x03B6;' if ( $c eq '&zeta;' );    # greek small letter zeta
    return '&#x03B7;' if ( $c eq '&eta;' );     # greek small letter eta
    return '&#x03B8;' if ( $c eq '&theta;' );   # greek small letter theta
    return '&#x03B9;' if ( $c eq '&iota;' );    # greek small letter iota
    return '&#x03BA;' if ( $c eq '&kappa;' );   # greek small letter kappa
    return '&#x03BB;' if ( $c eq '&lambda;' );  # greek small letter lambda
    return '&#x03BC;' if ( $c eq '&mu;' );      # greek small letter mu
    return '&#x03BD;' if ( $c eq '&nu;' );      # greek small letter nu
    return '&#x03BE;' if ( $c eq '&xi;' );      # greek small letter xi
    return '&#x03BF;' if ( $c eq '&omicron;' ); # greek small letter omicron
    return '&#x03C0;' if ( $c eq '&pi;' );      # greek small letter pi
    return '&#x03C1;' if ( $c eq '&rho;' );     # greek small letter rho
    return '&#x03C2;'
        if ( $c eq '&sigmaf;' );    # greek small letter final sigma
    return '&#x03C3;' if ( $c eq '&sigma;' );     # greek small letter sigma
    return '&#x03C4;' if ( $c eq '&tau;' );       # greek small letter tau
    return '&#x03C5;' if ( $c eq '&upsilon;' );   # greek small letter upsilon
    return '&#x03C6;' if ( $c eq '&phi;' );       # greek small letter phi
    return '&#x03C7;' if ( $c eq '&chi;' );       # greek small letter chi
    return '&#x03C8;' if ( $c eq '&psi;' );       # greek small letter psi
    return '&#x03C9;' if ( $c eq '&omega;' );     # greek small letter omega
    return '&#x03D1;'
        if ( $c eq '&thetasym;' );    # greek small letter theta symbol
    return '&#x03D2;' if ( $c eq '&upsih;' ); # greek upsilon with hook symbol
    return '&#x03D6;' if ( $c eq '&piv;' );   # greek pi symbol
    return '&#x2002;' if ( $c eq '&ensp;' );  # en space
    return '&#x2003;' if ( $c eq '&emsp;' );  # em space
    return '&#x2009;' if ( $c eq '&thinsp;' );   # thin space
    return '&#x200C;' if ( $c eq '&zwnj;' );     # zero width non-joiner
    return '&#x200D;' if ( $c eq '&zwj;' );      # zero width joiner
    return '&#x200E;' if ( $c eq '&lrm;' );      # left-to-right mark
    return '&#x200F;' if ( $c eq '&rlm;' );      # right-to-left mark
    return '&#x2013;' if ( $c eq '&ndash;' );    # en dash
    return '&#x2014;' if ( $c eq '&mdash;' );    # em dash
    return '&#x2018;' if ( $c eq '&lsquo;' );    # left single quotation mark
    return '&#x2019;' if ( $c eq '&rsquo;' );    # right single quotation mark
    return '&#x201A;' if ( $c eq '&sbquo;' );    # single low-9 quotation mark
    return '&#x201C;' if ( $c eq '&ldquo;' );    # left double quotation mark
    return '&#x201D;' if ( $c eq '&rdquo;' );    # right double quotation mark
    return '&#x201E;' if ( $c eq '&bdquo;' );    # double low-9 quotation mark
    return '&#x2020;' if ( $c eq '&dagger;' );   # dagger
    return '&#x2021;' if ( $c eq '&Dagger;' );   # double dagger
    return '&#x2022;' if ( $c eq '&bull;' );     # bullet = black small circle
    return '&#x2026;'
        if ( $c eq '&hellip;' );    # horizontal ellipsis = three dot leader
    return '&#x2030;' if ( $c eq '&permil;' );    # per mille sign
    return '&#x2032;' if ( $c eq '&prime;' );     # prime = minutes = feet
    return '&#x2033;'
        if ( $c eq '&Prime;' );    # double prime = seconds = inches
    return '&#x2039;'
        if ( $c eq '&lsaquo;' );   # single left-pointing angle quotation mark
    return '&#x203A;'
        if ( $c eq '&rsaquo;' );  # single right-pointing angle quotation mark
    return '&#x203E;' if ( $c eq '&oline;' );   # overline = spacing overscore
    return '&#x2044;' if ( $c eq '&frasl;' );   # fraction slash
    return '&#x20AC;' if ( $c eq '&euro;' );    # euro sign
    return '&#x2111;'
        if ( $c eq '&image;' );    # blackletter capital I = imaginary part
    return '&#x2118;'
        if ( $c eq '&weierp;' )
        ;    # script capital P = power set = Weierstrass p
    return '&#x211C;'
        if ( $c eq '&real;' );    # blackletter capital R = real part symbol
    return '&#x2122;' if ( $c eq '&trade;' );    # trade mark sign
    return '&#x2135;'
        if ( $c eq '&alefsym;' );   # alef symbol = first transfinite cardinal
    return '&#x2190;' if ( $c eq '&larr;' );    # leftwards arrow
    return '&#x2191;' if ( $c eq '&uarr;' );    # upwards arrow
    return '&#x2192;' if ( $c eq '&rarr;' );    # rightwards arrow
    return '&#x2193;' if ( $c eq '&darr;' );    # downwards arrow
    return '&#x2194;' if ( $c eq '&harr;' );    # left right arrow
    return '&#x21B5;'
        if ( $c eq '&crarr;' )
        ;    # downwards arrow with corner leftwards = carriage return
    return '&#x21D0;' if ( $c eq '&lArr;' );      # leftwards double arrow
    return '&#x21D1;' if ( $c eq '&uArr;' );      # upwards double arrow
    return '&#x21D2;' if ( $c eq '&rArr;' );      # rightwards double arrow
    return '&#x21D3;' if ( $c eq '&dArr;' );      # downwards double arrow
    return '&#x21D4;' if ( $c eq '&hArr;' );      # left right double arrow
    return '&#x2200;' if ( $c eq '&forall;' );    # for all
    return '&#x2202;' if ( $c eq '&part;' );      # partial differential
    return '&#x2203;' if ( $c eq '&exist;' );     # there exists
    return '&#x2205;'
        if ( $c eq '&empty;' );    # empty set = null set = diameter
    return '&#x2207;' if ( $c eq '&nabla;' );   # nabla = backward difference
    return '&#x2208;' if ( $c eq '&isin;' );    # element of
    return '&#x2209;' if ( $c eq '&notin;' );   # not an element of
    return '&#x220B;' if ( $c eq '&ni;' );      # contains as member
    return '&#x220F;' if ( $c eq '&prod;' );    # n-ary product = product sign
    return '&#x2211;' if ( $c eq '&sum;' );     # n-ary sumation
    return '&#x2212;' if ( $c eq '&minus;' );   # minus sign
    return '&#x2217;' if ( $c eq '&lowast;' );  # asterisk operator
    return '&#x221A;' if ( $c eq '&radic;' );   # square root = radical sign
    return '&#x221D;' if ( $c eq '&prop;' );    # proportional to
    return '&#x221E;' if ( $c eq '&infin;' );   # infinity
    return '&#x2220;' if ( $c eq '&ang;' );     # angle
    return '&#x2227;' if ( $c eq '&and;' );     # logical and = wedge
    return '&#x2228;' if ( $c eq '&or;' );      # logical or = vee
    return '&#x2229;' if ( $c eq '&cap;' );     # intersection = cap
    return '&#x222A;' if ( $c eq '&cup;' );     # union = cup
    return '&#x222B;' if ( $c eq '&int;' );     # integral
    return '&#x2234;' if ( $c eq '&there4;' );  # therefore
    return '&#x223C;'
        if ( $c eq '&sim;' );    # tilde operator = varies with = similar to
    return '&#x2245;' if ( $c eq '&cong;' );    # approximately equal to
    return '&#x2248;'
        if ( $c eq '&asymp;' );    # almost equal to = asymptotic to
    return '&#x2260;' if ( $c eq '&ne;' );       # not equal to
    return '&#x2261;' if ( $c eq '&equiv;' );    # identical to
    return '&#x2264;' if ( $c eq '&le;' );       # less-than or equal to
    return '&#x2265;' if ( $c eq '&ge;' );       # greater-than or equal to
    return '&#x2282;' if ( $c eq '&sub;' );      # subset of
    return '&#x2283;' if ( $c eq '&sup;' );      # superset of
    return '&#x2284;' if ( $c eq '&nsub;' );     # not a subset of
    return '&#x2286;' if ( $c eq '&sube;' );     # subset of or equal to
    return '&#x2287;' if ( $c eq '&supe;' );     # superset of or equal to
    return '&#x2295;' if ( $c eq '&oplus;' );    # circled plus = direct sum
    return '&#x2297;'
        if ( $c eq '&otimes;' );    # circled times = vector product
    return '&#x22A5;'
        if ( $c eq '&perp;' );      # up tack = orthogonal to = perpendicular
    return '&#x22C5;' if ( $c eq '&sdot;' );      # dot operator
    return '&#x2308;' if ( $c eq '&lceil;' );     # left ceiling = apl upstile
    return '&#x2309;' if ( $c eq '&rceil;' );     # right ceiling
    return '&#x230A;' if ( $c eq '&lfloor;' );    # left floor = apl downstile
    return '&#x230B;' if ( $c eq '&rfloor;' );    # right floor
    return '&#x2329;'
        if ( $c eq '&lang;' );    # left-pointing angle bracket = bra
    return '&#x232A;'
        if ( $c eq '&rang;' );    # right-pointing angle bracket = ket
    return '&#x25CA;' if ( $c eq '&loz;' );     # lozenge
    return '&#x2660;' if ( $c eq '&spades;' );  # black spade suit
    return '&#x2663;' if ( $c eq '&clubs;' );   # black club suit = shamrock
    return '&#x2665;' if ( $c eq '&hearts;' );  # black heart suit = valentine
    return '&#x2666;' if ( $c eq '&diams;' );   # black diamond suit
    return $c;
}

##############################################################################
#
# Function: isoents2numeric
#
# Receives an XML entity. If the entity is a named ISO characters entity,
# returns the hexadecimal numeric character entity. Best used like this:
# $content =~ s{(&[A-Za-z0-9.]+;)}{isoents2numeric($1)}xmsge;
#
##############################################################################

sub isoents2numeric {
    my $c = shift;
    return "&amp;"    if ( $c eq "&amp;" );       # AMPERSAND
    return "&apos;"   if ( $c eq "&apos;" );      # APOSTROPHE
    return "&gt;"     if ( $c eq "&gt;" );        # GREATER-THAN SIGN
    return "&quot;"   if ( $c eq "&quot;" );      # QUOTATION MARK
    return "&lt;"     if ( $c eq "&lt;" );        # LESS-THAN SIGN
    return '&#x0021;' if ( $c eq '&excl;' );      # EXCLAMATION MARK
    return '&#x0023;' if ( $c eq '&num;' );       # NUMBER SIGN
    return '&#x0024;' if ( $c eq '&dollar;' );    # DOLLAR SIGN
    return '&#x0025;' if ( $c eq '&percnt;' );    # PERCENT SIGN
    return '&#x0028;' if ( $c eq '&lpar;' );      # LEFT PARENTHESIS
    return '&#x0029;' if ( $c eq '&rpar;' );      # RIGHT PARENTHESIS
    return '&#x002A;' if ( $c eq '&ast;' );       # ASTERISK
    return '&#x002B;' if ( $c eq '&plus;' );      # PLUS SIGN
    return '&#x002C;' if ( $c eq '&comma;' );     # COMMA
    return '&#x002E;' if ( $c eq '&period;' );    # FULL STOP
    return '&#x002F;' if ( $c eq '&sol;' );       # SOLIDUS
    return '&#x003A;' if ( $c eq '&colon;' );     # COLON
    return '&#x003B;' if ( $c eq '&semi;' );      # SEMICOLON
    return '&#x003D;' if ( $c eq '&equals;' );    # EQUALS SIGN
    return '&#x003F;' if ( $c eq '&quest;' );     # QUESTION MARK
    return '&#x0040;' if ( $c eq '&commat;' );    # COMMERCIAL AT
    return '&#x005B;' if ( $c eq '&lsqb;' );      # LEFT SQUARE BRACKET
    return '&#x005C;' if ( $c eq '&bsol;' );      # REVERSE SOLIDUS
    return '&#x005D;' if ( $c eq '&rsqb;' );      # RIGHT SQUARE BRACKET
    return '&#x005F;' if ( $c eq '&lowbar;' );    # LOW LINE
    return '&#x0060;' if ( $c eq '&grave;' );     # GRAVE ACCENT
    return '&#x006A;' if ( $c eq '&jnodot;' );    # LATIN SMALL LETTER J
    return '&#x007B;' if ( $c eq '&lcub;' );      # LEFT CURLY BRACKET
    return '&#x007C;' if ( $c eq '&verbar;' );    # VERTICAL LINE
    return '&#x007D;' if ( $c eq '&rcub;' );      # RIGHT CURLY BRACKET
    return '&#x00A0;' if ( $c eq '&nbsp;' );      # NO-BREAK SPACE
    return '&#x00A1;' if ( $c eq '&iexcl;' );     # INVERTED EXCLAMATION MARK
    return '&#x00A2;' if ( $c eq '&cent;' );      # CENT SIGN
    return '&#x00A3;' if ( $c eq '&pound;' );     # POUND SIGN
    return '&#x00A4;' if ( $c eq '&curren;' );    # CURRENCY SIGN
    return '&#x00A5;' if ( $c eq '&yen;' );       # YEN SIGN
    return '&#x00A6;' if ( $c eq '&brvbar;' );    # BROKEN BAR
    return '&#x00A7;' if ( $c eq '&sect;' );      # SECTION SIGN
    return '&#x00A8;' if ( $c eq '&Dot;' );       # DIAERESIS
    return '&#x00A8;' if ( $c eq '&die;' );       # DIAERESIS
    return '&#x00A8;' if ( $c eq '&uml;' );       # DIAERESIS
    return '&#x00A9;' if ( $c eq '&copy;' );      # COPYRIGHT SIGN
    return '&#x00AA;' if ( $c eq '&ordf;' );      # FEMININE ORDINAL INDICATOR
    return '&#x00AB;'
        if ( $c eq '&laquo;' );    # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    return '&#x00AC;' if ( $c eq '&not;' );      # NOT SIGN
    return '&#x00AD;' if ( $c eq '&shy;' );      # SOFT HYPHEN
    return '&#x00AE;' if ( $c eq '&reg;' );      # REGISTERED SIGN
    return '&#x00AF;' if ( $c eq '&macr;' );     # MACRON
    return '&#x00B0;' if ( $c eq '&deg;' );      # DEGREE SIGN
    return '&#x00B1;' if ( $c eq '&plusmn;' );   # PLUS-MINUS SIGN
    return '&#x00B2;' if ( $c eq '&sup2;' );     # SUPERSCRIPT TWO
    return '&#x00B3;' if ( $c eq '&sup3;' );     # SUPERSCRIPT THREE
    return '&#x00B4;' if ( $c eq '&acute;' );    # ACUTE ACCENT
    return '&#x00B5;' if ( $c eq '&micro;' );    # MICRO SIGN
    return '&#x00B6;' if ( $c eq '&para;' );     # PILCROW SIGN
    return '&#x00B7;' if ( $c eq '&middot;' );   # MIDDLE DOT
    return '&#x00B8;' if ( $c eq '&cedil;' );    # CEDILLA
    return '&#x00B9;' if ( $c eq '&sup1;' );     # SUPERSCRIPT ONE
    return '&#x00BA;' if ( $c eq '&ordm;' );     # MASCULINE ORDINAL INDICATOR
    return '&#x00BB;'
        if ( $c eq '&raquo;' );   # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    return '&#x00BC;' if ( $c eq '&frac14;' );   # VULGAR FRACTION ONE QUARTER
    return '&#x00BD;' if ( $c eq '&frac12;' );   # VULGAR FRACTION ONE HALF
    return '&#x00BD;' if ( $c eq '&half;' );     # VULGAR FRACTION ONE HALF
    return '&#x00BE;'
        if ( $c eq '&frac34;' );    # VULGAR FRACTION THREE QUARTERS
    return '&#x00BF;' if ( $c eq '&iquest;' );    # INVERTED QUESTION MARK
    return '&#x00C0;'
        if ( $c eq '&Agrave;' );    # LATIN CAPITAL LETTER A WITH GRAVE
    return '&#x00C1;'
        if ( $c eq '&Aacute;' );    # LATIN CAPITAL LETTER A WITH ACUTE
    return '&#x00C2;'
        if ( $c eq '&Acirc;' );     # LATIN CAPITAL LETTER A WITH CIRCUMFLEX
    return '&#x00C3;'
        if ( $c eq '&Atilde;' );    # LATIN CAPITAL LETTER A WITH TILDE
    return '&#x00C4;'
        if ( $c eq '&Auml;' );      # LATIN CAPITAL LETTER A WITH DIAERESIS
    return '&#x00C5;'
        if ( $c eq '&Aring;' );     # LATIN CAPITAL LETTER A WITH RING ABOVE
    return '&#x00C6;' if ( $c eq '&AElig;' );    # LATIN CAPITAL LETTER AE
    return '&#x00C7;'
        if ( $c eq '&Ccedil;' );    # LATIN CAPITAL LETTER C WITH CEDILLA
    return '&#x00C8;'
        if ( $c eq '&Egrave;' );    # LATIN CAPITAL LETTER E WITH GRAVE
    return '&#x00C9;'
        if ( $c eq '&Eacute;' );    # LATIN CAPITAL LETTER E WITH ACUTE
    return '&#x00CA;'
        if ( $c eq '&Ecirc;' );     # LATIN CAPITAL LETTER E WITH CIRCUMFLEX
    return '&#x00CB;'
        if ( $c eq '&Euml;' );      # LATIN CAPITAL LETTER E WITH DIAERESIS
    return '&#x00CC;'
        if ( $c eq '&Igrave;' );    # LATIN CAPITAL LETTER I WITH GRAVE
    return '&#x00CD;'
        if ( $c eq '&Iacute;' );    # LATIN CAPITAL LETTER I WITH ACUTE
    return '&#x00CE;'
        if ( $c eq '&Icirc;' );     # LATIN CAPITAL LETTER I WITH CIRCUMFLEX
    return '&#x00CF;'
        if ( $c eq '&Iuml;' );      # LATIN CAPITAL LETTER I WITH DIAERESIS
    return '&#x00D0;' if ( $c eq '&ETH;' );    # LATIN CAPITAL LETTER ETH
    return '&#x00D1;'
        if ( $c eq '&Ntilde;' );    # LATIN CAPITAL LETTER N WITH TILDE
    return '&#x00D2;'
        if ( $c eq '&Ograve;' );    # LATIN CAPITAL LETTER O WITH GRAVE
    return '&#x00D3;'
        if ( $c eq '&Oacute;' );    # LATIN CAPITAL LETTER O WITH ACUTE
    return '&#x00D4;'
        if ( $c eq '&Ocirc;' );     # LATIN CAPITAL LETTER O WITH CIRCUMFLEX
    return '&#x00D5;'
        if ( $c eq '&Otilde;' );    # LATIN CAPITAL LETTER O WITH TILDE
    return '&#x00D6;'
        if ( $c eq '&Ouml;' );      # LATIN CAPITAL LETTER O WITH DIAERESIS
    return '&#x00D7;' if ( $c eq '&times;' );    # MULTIPLICATION SIGN
    return '&#x00D8;'
        if ( $c eq '&Oslash;' );    # LATIN CAPITAL LETTER O WITH STROKE
    return '&#x00D9;'
        if ( $c eq '&Ugrave;' );    # LATIN CAPITAL LETTER U WITH GRAVE
    return '&#x00DA;'
        if ( $c eq '&Uacute;' );    # LATIN CAPITAL LETTER U WITH ACUTE
    return '&#x00DB;'
        if ( $c eq '&Ucirc;' );     # LATIN CAPITAL LETTER U WITH CIRCUMFLEX
    return '&#x00DC;'
        if ( $c eq '&Uuml;' );      # LATIN CAPITAL LETTER U WITH DIAERESIS
    return '&#x00DD;'
        if ( $c eq '&Yacute;' );    # LATIN CAPITAL LETTER Y WITH ACUTE
    return '&#x00DE;' if ( $c eq '&THORN;' );    # LATIN CAPITAL LETTER THORN
    return '&#x00DF;' if ( $c eq '&szlig;' );    # LATIN SMALL LETTER SHARP S
    return '&#x00E0;'
        if ( $c eq '&agrave;' );    # LATIN SMALL LETTER A WITH GRAVE
    return '&#x00E1;'
        if ( $c eq '&aacute;' );    # LATIN SMALL LETTER A WITH ACUTE
    return '&#x00E2;'
        if ( $c eq '&acirc;' );     # LATIN SMALL LETTER A WITH CIRCUMFLEX
    return '&#x00E3;'
        if ( $c eq '&atilde;' );    # LATIN SMALL LETTER A WITH TILDE
    return '&#x00E4;'
        if ( $c eq '&auml;' );      # LATIN SMALL LETTER A WITH DIAERESIS
    return '&#x00E5;'
        if ( $c eq '&aring;' );     # LATIN SMALL LETTER A WITH RING ABOVE
    return '&#x00E6;' if ( $c eq '&aelig;' );    # LATIN SMALL LETTER AE
    return '&#x00E7;'
        if ( $c eq '&ccedil;' );    # LATIN SMALL LETTER C WITH CEDILLA
    return '&#x00E8;'
        if ( $c eq '&egrave;' );    # LATIN SMALL LETTER E WITH GRAVE
    return '&#x00E9;'
        if ( $c eq '&eacute;' );    # LATIN SMALL LETTER E WITH ACUTE
    return '&#x00EA;'
        if ( $c eq '&ecirc;' );     # LATIN SMALL LETTER E WITH CIRCUMFLEX
    return '&#x00EB;'
        if ( $c eq '&euml;' );      # LATIN SMALL LETTER E WITH DIAERESIS
    return '&#x00EC;'
        if ( $c eq '&igrave;' );    # LATIN SMALL LETTER I WITH GRAVE
    return '&#x00ED;'
        if ( $c eq '&iacute;' );    # LATIN SMALL LETTER I WITH ACUTE
    return '&#x00EE;'
        if ( $c eq '&icirc;' );     # LATIN SMALL LETTER I WITH CIRCUMFLEX
    return '&#x00EF;'
        if ( $c eq '&iuml;' );      # LATIN SMALL LETTER I WITH DIAERESIS
    return '&#x00F0;' if ( $c eq '&eth;' );    # LATIN SMALL LETTER ETH
    return '&#x00F1;'
        if ( $c eq '&ntilde;' );    # LATIN SMALL LETTER N WITH TILDE
    return '&#x00F2;'
        if ( $c eq '&ograve;' );    # LATIN SMALL LETTER O WITH GRAVE
    return '&#x00F3;'
        if ( $c eq '&oacute;' );    # LATIN SMALL LETTER O WITH ACUTE
    return '&#x00F4;'
        if ( $c eq '&ocirc;' );     # LATIN SMALL LETTER O WITH CIRCUMFLEX
    return '&#x00F5;'
        if ( $c eq '&otilde;' );    # LATIN SMALL LETTER O WITH TILDE
    return '&#x00F6;'
        if ( $c eq '&ouml;' );      # LATIN SMALL LETTER O WITH DIAERESIS
    return '&#x00F7;' if ( $c eq '&divide;' );    # DIVISION SIGN
    return '&#x00F8;'
        if ( $c eq '&oslash;' );    # LATIN SMALL LETTER O WITH STROKE
    return '&#x00F9;'
        if ( $c eq '&ugrave;' );    # LATIN SMALL LETTER U WITH GRAVE
    return '&#x00FA;'
        if ( $c eq '&uacute;' );    # LATIN SMALL LETTER U WITH ACUTE
    return '&#x00FB;'
        if ( $c eq '&ucirc;' );     # LATIN SMALL LETTER U WITH CIRCUMFLEX
    return '&#x00FC;'
        if ( $c eq '&uuml;' );      # LATIN SMALL LETTER U WITH DIAERESIS
    return '&#x00FD;'
        if ( $c eq '&yacute;' );    # LATIN SMALL LETTER Y WITH ACUTE
    return '&#x00FE;' if ( $c eq '&thorn;' );    # LATIN SMALL LETTER THORN
    return '&#x00FF;'
        if ( $c eq '&yuml;' );    # LATIN SMALL LETTER Y WITH DIAERESIS
    return '&#x0100;'
        if ( $c eq '&Amacr;' );    # LATIN CAPITAL LETTER A WITH MACRON
    return '&#x0101;'
        if ( $c eq '&amacr;' );    # LATIN SMALL LETTER A WITH MACRON
    return '&#x0102;'
        if ( $c eq '&Abreve;' );    # LATIN CAPITAL LETTER A WITH BREVE
    return '&#x0103;'
        if ( $c eq '&abreve;' );    # LATIN SMALL LETTER A WITH BREVE
    return '&#x0104;'
        if ( $c eq '&Aogon;' );     # LATIN CAPITAL LETTER A WITH OGONEK
    return '&#x0105;'
        if ( $c eq '&aogon;' );     # LATIN SMALL LETTER A WITH OGONEK
    return '&#x0106;'
        if ( $c eq '&Cacute;' );    # LATIN CAPITAL LETTER C WITH ACUTE
    return '&#x0107;'
        if ( $c eq '&cacute;' );    # LATIN SMALL LETTER C WITH ACUTE
    return '&#x0108;'
        if ( $c eq '&Ccirc;' );     # LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    return '&#x0109;'
        if ( $c eq '&ccirc;' );     # LATIN SMALL LETTER C WITH CIRCUMFLEX
    return '&#x010A;'
        if ( $c eq '&Cdot;' );      # LATIN CAPITAL LETTER C WITH DOT ABOVE
    return '&#x010B;'
        if ( $c eq '&cdot;' );      # LATIN SMALL LETTER C WITH DOT ABOVE
    return '&#x010C;'
        if ( $c eq '&Ccaron;' );    # LATIN CAPITAL LETTER C WITH CARON
    return '&#x010D;'
        if ( $c eq '&ccaron;' );    # LATIN SMALL LETTER C WITH CARON
    return '&#x010E;'
        if ( $c eq '&Dcaron;' );    # LATIN CAPITAL LETTER D WITH CARON
    return '&#x010F;'
        if ( $c eq '&dcaron;' );    # LATIN SMALL LETTER D WITH CARON
    return '&#x0110;'
        if ( $c eq '&Dstrok;' );    # LATIN CAPITAL LETTER D WITH STROKE
    return '&#x0111;'
        if ( $c eq '&dstrok;' );    # LATIN SMALL LETTER D WITH STROKE
    return '&#x0112;'
        if ( $c eq '&Emacr;' );     # LATIN CAPITAL LETTER E WITH MACRON
    return '&#x0113;'
        if ( $c eq '&emacr;' );     # LATIN SMALL LETTER E WITH MACRON
    return '&#x0116;'
        if ( $c eq '&Edot;' );      # LATIN CAPITAL LETTER E WITH DOT ABOVE
    return '&#x0117;'
        if ( $c eq '&edot;' );      # LATIN SMALL LETTER E WITH DOT ABOVE
    return '&#x0118;'
        if ( $c eq '&Eogon;' );     # LATIN CAPITAL LETTER E WITH OGONEK
    return '&#x0119;'
        if ( $c eq '&eogon;' );     # LATIN SMALL LETTER E WITH OGONEK
    return '&#x011A;'
        if ( $c eq '&Ecaron;' );    # LATIN CAPITAL LETTER E WITH CARON
    return '&#x011B;'
        if ( $c eq '&ecaron;' );    # LATIN SMALL LETTER E WITH CARON
    return '&#x011C;'
        if ( $c eq '&Gcirc;' );     # LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    return '&#x011D;'
        if ( $c eq '&gcirc;' );     # LATIN SMALL LETTER G WITH CIRCUMFLEX
    return '&#x011E;'
        if ( $c eq '&Gbreve;' );    # LATIN CAPITAL LETTER G WITH BREVE
    return '&#x011F;'
        if ( $c eq '&gbreve;' );    # LATIN SMALL LETTER G WITH BREVE
    return '&#x0120;'
        if ( $c eq '&Gdot;' );      # LATIN CAPITAL LETTER G WITH DOT ABOVE
    return '&#x0121;'
        if ( $c eq '&gdot;' );      # LATIN SMALL LETTER G WITH DOT ABOVE
    return '&#x0122;'
        if ( $c eq '&Gcedil;' );    # LATIN CAPITAL LETTER G WITH CEDILLA
    return '&#x0124;'
        if ( $c eq '&Hcirc;' );     # LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    return '&#x0125;'
        if ( $c eq '&hcirc;' );     # LATIN SMALL LETTER H WITH CIRCUMFLEX
    return '&#x0126;'
        if ( $c eq '&Hstrok;' );    # LATIN CAPITAL LETTER H WITH STROKE
    return '&#x0127;'
        if ( $c eq '&hstrok;' );    # LATIN SMALL LETTER H WITH STROKE
    return '&#x0128;'
        if ( $c eq '&Itilde;' );    # LATIN CAPITAL LETTER I WITH TILDE
    return '&#x0129;'
        if ( $c eq '&itilde;' );    # LATIN SMALL LETTER I WITH TILDE
    return '&#x012A;'
        if ( $c eq '&Imacr;' );     # LATIN CAPITAL LETTER I WITH MACRON
    return '&#x012B;'
        if ( $c eq '&imacr;' );     # LATIN SMALL LETTER I WITH MACRON
    return '&#x012E;'
        if ( $c eq '&Iogon;' );     # LATIN CAPITAL LETTER I WITH OGONEK
    return '&#x012F;'
        if ( $c eq '&iogon;' );     # LATIN SMALL LETTER I WITH OGONEK
    return '&#x0130;'
        if ( $c eq '&Idot;' );      # LATIN CAPITAL LETTER I WITH DOT ABOVE
    return '&#x0131;' if ( $c eq '&inodot;' );  # LATIN SMALL LETTER DOTLESS I
    return '&#x0132;' if ( $c eq '&IJlig;' );   # LATIN CAPITAL LIGATURE IJ
    return '&#x0133;' if ( $c eq '&ijlig;' );   # LATIN SMALL LIGATURE IJ
    return '&#x0134;'
        if ( $c eq '&Jcirc;' );    # LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    return '&#x0135;'
        if ( $c eq '&jcirc;' );    # LATIN SMALL LETTER J WITH CIRCUMFLEX
    return '&#x0136;'
        if ( $c eq '&Kcedil;' );    # LATIN CAPITAL LETTER K WITH CEDILLA
    return '&#x0137;'
        if ( $c eq '&kcedil;' );    # LATIN SMALL LETTER K WITH CEDILLA
    return '&#x0138;' if ( $c eq '&kgreen;' );    # LATIN SMALL LETTER KRA
    return '&#x0139;'
        if ( $c eq '&Lacute;' );    # LATIN CAPITAL LETTER L WITH ACUTE
    return '&#x013A;'
        if ( $c eq '&lacute;' );    # LATIN SMALL LETTER L WITH ACUTE
    return '&#x013B;'
        if ( $c eq '&Lcedil;' );    # LATIN CAPITAL LETTER L WITH CEDILLA
    return '&#x013C;'
        if ( $c eq '&lcedil;' );    # LATIN SMALL LETTER L WITH CEDILLA
    return '&#x013D;'
        if ( $c eq '&Lcaron;' );    # LATIN CAPITAL LETTER L WITH CARON
    return '&#x013E;'
        if ( $c eq '&lcaron;' );    # LATIN SMALL LETTER L WITH CARON
    return '&#x013F;'
        if ( $c eq '&Lmidot;' );    # LATIN CAPITAL LETTER L WITH MIDDLE DOT
    return '&#x0140;'
        if ( $c eq '&lmidot;' );    # LATIN SMALL LETTER L WITH MIDDLE DOT
    return '&#x0141;'
        if ( $c eq '&Lstrok;' );    # LATIN CAPITAL LETTER L WITH STROKE
    return '&#x0142;'
        if ( $c eq '&lstrok;' );    # LATIN SMALL LETTER L WITH STROKE
    return '&#x0143;'
        if ( $c eq '&Nacute;' );    # LATIN CAPITAL LETTER N WITH ACUTE
    return '&#x0144;'
        if ( $c eq '&nacute;' );    # LATIN SMALL LETTER N WITH ACUTE
    return '&#x0145;'
        if ( $c eq '&Ncedil;' );    # LATIN CAPITAL LETTER N WITH CEDILLA
    return '&#x0146;'
        if ( $c eq '&ncedil;' );    # LATIN SMALL LETTER N WITH CEDILLA
    return '&#x0147;'
        if ( $c eq '&Ncaron;' );    # LATIN CAPITAL LETTER N WITH CARON
    return '&#x0148;'
        if ( $c eq '&ncaron;' );    # LATIN SMALL LETTER N WITH CARON
    return '&#x0149;'
        if ( $c eq '&napos;' );  # LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    return '&#x014A;' if ( $c eq '&ENG;' );    # LATIN CAPITAL LETTER ENG
    return '&#x014B;' if ( $c eq '&eng;' );    # LATIN SMALL LETTER ENG
    return '&#x014C;'
        if ( $c eq '&Omacr;' );    # LATIN CAPITAL LETTER O WITH MACRON
    return '&#x014D;'
        if ( $c eq '&omacr;' );    # LATIN SMALL LETTER O WITH MACRON
    return '&#x0150;'
        if ( $c eq '&Odblac;' );    # LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    return '&#x0151;'
        if ( $c eq '&odblac;' );    # LATIN SMALL LETTER O WITH DOUBLE ACUTE
    return '&#x0152;' if ( $c eq '&OElig;' );    # LATIN CAPITAL LIGATURE OE
    return '&#x0153;' if ( $c eq '&oelig;' );    # LATIN SMALL LIGATURE OE
    return '&#x0154;'
        if ( $c eq '&Racute;' );    # LATIN CAPITAL LETTER R WITH ACUTE
    return '&#x0155;'
        if ( $c eq '&racute;' );    # LATIN SMALL LETTER R WITH ACUTE
    return '&#x0156;'
        if ( $c eq '&Rcedil;' );    # LATIN CAPITAL LETTER R WITH CEDILLA
    return '&#x0157;'
        if ( $c eq '&rcedil;' );    # LATIN SMALL LETTER R WITH CEDILLA
    return '&#x0158;'
        if ( $c eq '&Rcaron;' );    # LATIN CAPITAL LETTER R WITH CARON
    return '&#x0159;'
        if ( $c eq '&rcaron;' );    # LATIN SMALL LETTER R WITH CARON
    return '&#x015A;'
        if ( $c eq '&Sacute;' );    # LATIN CAPITAL LETTER S WITH ACUTE
    return '&#x015B;'
        if ( $c eq '&sacute;' );    # LATIN SMALL LETTER S WITH ACUTE
    return '&#x015C;'
        if ( $c eq '&Scirc;' );     # LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    return '&#x015D;'
        if ( $c eq '&scirc;' );     # LATIN SMALL LETTER S WITH CIRCUMFLEX
    return '&#x015E;'
        if ( $c eq '&Scedil;' );    # LATIN CAPITAL LETTER S WITH CEDILLA
    return '&#x015F;'
        if ( $c eq '&scedil;' );    # LATIN SMALL LETTER S WITH CEDILLA
    return '&#x0160;'
        if ( $c eq '&Scaron;' );    # LATIN CAPITAL LETTER S WITH CARON
    return '&#x0161;'
        if ( $c eq '&scaron;' );    # LATIN SMALL LETTER S WITH CARON
    return '&#x0162;'
        if ( $c eq '&Tcedil;' );    # LATIN CAPITAL LETTER T WITH CEDILLA
    return '&#x0163;'
        if ( $c eq '&tcedil;' );    # LATIN SMALL LETTER T WITH CEDILLA
    return '&#x0164;'
        if ( $c eq '&Tcaron;' );    # LATIN CAPITAL LETTER T WITH CARON
    return '&#x0165;'
        if ( $c eq '&tcaron;' );    # LATIN SMALL LETTER T WITH CARON
    return '&#x0166;'
        if ( $c eq '&Tstrok;' );    # LATIN CAPITAL LETTER T WITH STROKE
    return '&#x0167;'
        if ( $c eq '&tstrok;' );    # LATIN SMALL LETTER T WITH STROKE
    return '&#x0168;'
        if ( $c eq '&Utilde;' );    # LATIN CAPITAL LETTER U WITH TILDE
    return '&#x0169;'
        if ( $c eq '&utilde;' );    # LATIN SMALL LETTER U WITH TILDE
    return '&#x016A;'
        if ( $c eq '&Umacr;' );     # LATIN CAPITAL LETTER U WITH MACRON
    return '&#x016B;'
        if ( $c eq '&umacr;' );     # LATIN SMALL LETTER U WITH MACRON
    return '&#x016C;'
        if ( $c eq '&Ubreve;' );    # LATIN CAPITAL LETTER U WITH BREVE
    return '&#x016D;'
        if ( $c eq '&ubreve;' );    # LATIN SMALL LETTER U WITH BREVE
    return '&#x016E;'
        if ( $c eq '&Uring;' );     # LATIN CAPITAL LETTER U WITH RING ABOVE
    return '&#x016F;'
        if ( $c eq '&uring;' );     # LATIN SMALL LETTER U WITH RING ABOVE
    return '&#x0170;'
        if ( $c eq '&Udblac;' );    # LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    return '&#x0171;'
        if ( $c eq '&udblac;' );    # LATIN SMALL LETTER U WITH DOUBLE ACUTE
    return '&#x0172;'
        if ( $c eq '&Uogon;' );     # LATIN CAPITAL LETTER U WITH OGONEK
    return '&#x0173;'
        if ( $c eq '&uogon;' );     # LATIN SMALL LETTER U WITH OGONEK
    return '&#x0174;'
        if ( $c eq '&Wcirc;' );     # LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    return '&#x0175;'
        if ( $c eq '&wcirc;' );     # LATIN SMALL LETTER W WITH CIRCUMFLEX
    return '&#x0176;'
        if ( $c eq '&Ycirc;' );     # LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    return '&#x0177;'
        if ( $c eq '&ycirc;' );     # LATIN SMALL LETTER Y WITH CIRCUMFLEX
    return '&#x0178;'
        if ( $c eq '&Yuml;' );      # LATIN CAPITAL LETTER Y WITH DIAERESIS
    return '&#x0179;'
        if ( $c eq '&Zacute;' );    # LATIN CAPITAL LETTER Z WITH ACUTE
    return '&#x017A;'
        if ( $c eq '&zacute;' );    # LATIN SMALL LETTER Z WITH ACUTE
    return '&#x017B;'
        if ( $c eq '&Zdot;' );      # LATIN CAPITAL LETTER Z WITH DOT ABOVE
    return '&#x017C;'
        if ( $c eq '&zdot;' );      # LATIN SMALL LETTER Z WITH DOT ABOVE
    return '&#x017D;'
        if ( $c eq '&Zcaron;' );    # LATIN CAPITAL LETTER Z WITH CARON
    return '&#x017E;'
        if ( $c eq '&zcaron;' );    # LATIN SMALL LETTER Z WITH CARON
    return '&#x0192;' if ( $c eq '&fnof;' );  # LATIN SMALL LETTER F WITH HOOK
    return '&#x01F5;'
        if ( $c eq '&gacute;' );    # LATIN SMALL LETTER G WITH ACUTE
    return '&#x02C6;'
        if ( $c eq '&circ;' );      # MODIFIER LETTER CIRCUMFLEX ACCENT
    return '&#x02C7;' if ( $c eq '&caron;' );    # CARON
    return '&#x02D8;' if ( $c eq '&breve;' );    # BREVE
    return '&#x02D9;' if ( $c eq '&dot;' );      # DOT ABOVE
    return '&#x02DA;' if ( $c eq '&ring;' );     # RING ABOVE
    return '&#x02DB;' if ( $c eq '&ogon;' );     # OGONEK
    return '&#x02DC;' if ( $c eq '&tilde;' );    # SMALL TILDE
    return '&#x02DD;' if ( $c eq '&dblac;' );    # DOUBLE ACUTE ACCENT
    return '&#x0386;'
        if ( $c eq '&Aacgr;' );    # GREEK CAPITAL LETTER ALPHA WITH TONOS
    return '&#x0388;'
        if ( $c eq '&Eacgr;' );    # GREEK CAPITAL LETTER EPSILON WITH TONOS
    return '&#x0389;'
        if ( $c eq '&EEacgr;' );    # GREEK CAPITAL LETTER ETA WITH TONOS
    return '&#x038A;'
        if ( $c eq '&Iacgr;' );     # GREEK CAPITAL LETTER IOTA WITH TONOS
    return '&#x038C;'
        if ( $c eq '&Oacgr;' );     # GREEK CAPITAL LETTER OMICRON WITH TONOS
    return '&#x038E;'
        if ( $c eq '&Uacgr;' );     # GREEK CAPITAL LETTER UPSILON WITH TONOS
    return '&#x038F;'
        if ( $c eq '&OHacgr;' );    # GREEK CAPITAL LETTER OMEGA WITH TONOS
    return '&#x0390;'
        if ( $c eq '&idiagr;' )
        ;    # GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    return '&#x0391;' if ( $c eq '&Agr;' );     # GREEK CAPITAL LETTER ALPHA
    return '&#x0392;' if ( $c eq '&Bgr;' );     # GREEK CAPITAL LETTER BETA
    return '&#x0393;' if ( $c eq '&Gamma;' );   # GREEK CAPITAL LETTER GAMMA
    return '&#x0393;' if ( $c eq '&Ggr;' );     # GREEK CAPITAL LETTER GAMMA
    return '&#x0394;' if ( $c eq '&Delta;' );   # GREEK CAPITAL LETTER DELTA
    return '&#x0394;' if ( $c eq '&Dgr;' );     # GREEK CAPITAL LETTER DELTA
    return '&#x0395;' if ( $c eq '&Egr;' );     # GREEK CAPITAL LETTER EPSILON
    return '&#x0396;' if ( $c eq '&Zgr;' );     # GREEK CAPITAL LETTER ZETA
    return '&#x0397;' if ( $c eq '&EEgr;' );    # GREEK CAPITAL LETTER ETA
    return '&#x0398;' if ( $c eq '&THgr;' );    # GREEK CAPITAL LETTER THETA
    return '&#x0398;' if ( $c eq '&Theta;' );   # GREEK CAPITAL LETTER THETA
    return '&#x0399;' if ( $c eq '&Igr;' );     # GREEK CAPITAL LETTER IOTA
    return '&#x039A;' if ( $c eq '&Kgr;' );     # GREEK CAPITAL LETTER KAPPA
    return '&#x039B;' if ( $c eq '&Lambda;' );  # GREEK CAPITAL LETTER LAMDA
    return '&#x039B;' if ( $c eq '&Lgr;' );     # GREEK CAPITAL LETTER LAMDA
    return '&#x039C;' if ( $c eq '&Mgr;' );     # GREEK CAPITAL LETTER MU
    return '&#x039D;' if ( $c eq '&Ngr;' );     # GREEK CAPITAL LETTER NU
    return '&#x039E;' if ( $c eq '&Xgr;' );     # GREEK CAPITAL LETTER XI
    return '&#x039E;' if ( $c eq '&Xi;' );      # GREEK CAPITAL LETTER XI
    return '&#x039F;' if ( $c eq '&Ogr;' );     # GREEK CAPITAL LETTER OMICRON
    return '&#x03A0;' if ( $c eq '&Pgr;' );     # GREEK CAPITAL LETTER PI
    return '&#x03A0;' if ( $c eq '&Pi;' );      # GREEK CAPITAL LETTER PI
    return '&#x03A1;' if ( $c eq '&Rgr;' );     # GREEK CAPITAL LETTER RHO
    return '&#x03A3;' if ( $c eq '&Sgr;' );     # GREEK CAPITAL LETTER SIGMA
    return '&#x03A3;' if ( $c eq '&Sigma;' );   # GREEK CAPITAL LETTER SIGMA
    return '&#x03A4;' if ( $c eq '&Tgr;' );     # GREEK CAPITAL LETTER TAU
    return '&#x03A5;' if ( $c eq '&Ugr;' );     # GREEK CAPITAL LETTER UPSILON
    return '&#x03A6;' if ( $c eq '&PHgr;' );    # GREEK CAPITAL LETTER PHI
    return '&#x03A6;' if ( $c eq '&Phi;' );     # GREEK CAPITAL LETTER PHI
    return '&#x03A7;' if ( $c eq '&KHgr;' );    # GREEK CAPITAL LETTER CHI
    return '&#x03A8;' if ( $c eq '&PSgr;' );    # GREEK CAPITAL LETTER PSI
    return '&#x03A8;' if ( $c eq '&Psi;' );     # GREEK CAPITAL LETTER PSI
    return '&#x03A9;' if ( $c eq '&OHgr;' );    # GREEK CAPITAL LETTER OMEGA
    return '&#x03A9;' if ( $c eq '&Omega;' );   # GREEK CAPITAL LETTER OMEGA
    return '&#x03AA;'
        if ( $c eq '&Idigr;' );    # GREEK CAPITAL LETTER IOTA WITH DIALYTIKA
    return '&#x03AB;'
        if ( $c eq '&Udigr;' );  # GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    return '&#x03AC;'
        if ( $c eq '&aacgr;' );    # GREEK SMALL LETTER ALPHA WITH TONOS
    return '&#x03AD;'
        if ( $c eq '&eacgr;' );    # GREEK SMALL LETTER EPSILON WITH TONOS
    return '&#x03AE;'
        if ( $c eq '&eeacgr;' );    # GREEK SMALL LETTER ETA WITH TONOS
    return '&#x03AF;'
        if ( $c eq '&iacgr;' );     # GREEK SMALL LETTER IOTA WITH TONOS
    return '&#x03B0;'
        if ( $c eq '&udiagr;' )
        ;    # GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS
    return '&#x03B1;' if ( $c eq '&agr;' );       # GREEK SMALL LETTER ALPHA
    return '&#x03B1;' if ( $c eq '&alpha;' );     # GREEK SMALL LETTER ALPHA
    return '&#x03B2;' if ( $c eq '&beta;' );      # GREEK SMALL LETTER BETA
    return '&#x03B2;' if ( $c eq '&bgr;' );       # GREEK SMALL LETTER BETA
    return '&#x03B3;' if ( $c eq '&gamma;' );     # GREEK SMALL LETTER GAMMA
    return '&#x03B3;' if ( $c eq '&ggr;' );       # GREEK SMALL LETTER GAMMA
    return '&#x03B4;' if ( $c eq '&delta;' );     # GREEK SMALL LETTER DELTA
    return '&#x03B4;' if ( $c eq '&dgr;' );       # GREEK SMALL LETTER DELTA
    return '&#x03B5;' if ( $c eq '&egr;' );       # GREEK SMALL LETTER EPSILON
    return '&#x03B5;' if ( $c eq '&epsiv;' );     # GREEK SMALL LETTER EPSILON
    return '&#x03B6;' if ( $c eq '&zeta;' );      # GREEK SMALL LETTER ZETA
    return '&#x03B6;' if ( $c eq '&zgr;' );       # GREEK SMALL LETTER ZETA
    return '&#x03B7;' if ( $c eq '&eegr;' );      # GREEK SMALL LETTER ETA
    return '&#x03B7;' if ( $c eq '&eta;' );       # GREEK SMALL LETTER ETA
    return '&#x03B8;' if ( $c eq '&thetas;' );    # GREEK SMALL LETTER THETA
    return '&#x03B8;' if ( $c eq '&thgr;' );      # GREEK SMALL LETTER THETA
    return '&#x03B9;' if ( $c eq '&igr;' );       # GREEK SMALL LETTER IOTA
    return '&#x03B9;' if ( $c eq '&iota;' );      # GREEK SMALL LETTER IOTA
    return '&#x03BA;' if ( $c eq '&kappa;' );     # GREEK SMALL LETTER KAPPA
    return '&#x03BA;' if ( $c eq '&kgr;' );       # GREEK SMALL LETTER KAPPA
    return '&#x03BB;' if ( $c eq '&lambda;' );    # GREEK SMALL LETTER LAMDA
    return '&#x03BB;' if ( $c eq '&lgr;' );       # GREEK SMALL LETTER LAMDA
    return '&#x03BC;' if ( $c eq '&mgr;' );       # GREEK SMALL LETTER MU
    return '&#x03BC;' if ( $c eq '&mu;' );        # GREEK SMALL LETTER MU
    return '&#x03BD;' if ( $c eq '&ngr;' );       # GREEK SMALL LETTER NU
    return '&#x03BD;' if ( $c eq '&nu;' );        # GREEK SMALL LETTER NU
    return '&#x03BE;' if ( $c eq '&xgr;' );       # GREEK SMALL LETTER XI
    return '&#x03BE;' if ( $c eq '&xi;' );        # GREEK SMALL LETTER XI
    return '&#x03BF;' if ( $c eq '&ogr;' );       # GREEK SMALL LETTER OMICRON
    return '&#x03C0;' if ( $c eq '&pgr;' );       # GREEK SMALL LETTER PI
    return '&#x03C0;' if ( $c eq '&pi;' );        # GREEK SMALL LETTER PI
    return '&#x03C1;' if ( $c eq '&rgr;' );       # GREEK SMALL LETTER RHO
    return '&#x03C1;' if ( $c eq '&rho;' );       # GREEK SMALL LETTER RHO
    return '&#x03C2;' if ( $c eq '&sfgr;' );  # GREEK SMALL LETTER FINAL SIGMA
    return '&#x03C2;'
        if ( $c eq '&sigmav;' );              # GREEK SMALL LETTER FINAL SIGMA
    return '&#x03C3;' if ( $c eq '&sgr;' );   # GREEK SMALL LETTER SIGMA
    return '&#x03C3;' if ( $c eq '&sigma;' ); # GREEK SMALL LETTER SIGMA
    return '&#x03C4;' if ( $c eq '&tau;' );   # GREEK SMALL LETTER TAU
    return '&#x03C4;' if ( $c eq '&tgr;' );   # GREEK SMALL LETTER TAU
    return '&#x03C5;' if ( $c eq '&ugr;' );   # GREEK SMALL LETTER UPSILON
    return '&#x03C5;' if ( $c eq '&upsi;' );  # GREEK SMALL LETTER UPSILON
    return '&#x03C6;' if ( $c eq '&phgr;' );  # GREEK SMALL LETTER PHI
    return '&#x03C6;' if ( $c eq '&phiv;' );  # GREEK SMALL LETTER PHI
    return '&#x03C7;' if ( $c eq '&chi;' );   # GREEK SMALL LETTER CHI
    return '&#x03C7;' if ( $c eq '&khgr;' );  # GREEK SMALL LETTER CHI
    return '&#x03C8;' if ( $c eq '&psgr;' );  # GREEK SMALL LETTER PSI
    return '&#x03C8;' if ( $c eq '&psi;' );   # GREEK SMALL LETTER PSI
    return '&#x03C9;' if ( $c eq '&ohgr;' );  # GREEK SMALL LETTER OMEGA
    return '&#x03C9;' if ( $c eq '&omega;' ); # GREEK SMALL LETTER OMEGA
    return '&#x03CA;'
        if ( $c eq '&idigr;' );    # GREEK SMALL LETTER IOTA WITH DIALYTIKA
    return '&#x03CB;'
        if ( $c eq '&udigr;' );    # GREEK SMALL LETTER UPSILON WITH DIALYTIKA
    return '&#x03CC;'
        if ( $c eq '&oacgr;' );    # GREEK SMALL LETTER OMICRON WITH TONOS
    return '&#x03CD;'
        if ( $c eq '&uacgr;' );    # GREEK SMALL LETTER UPSILON WITH TONOS
    return '&#x03CE;'
        if ( $c eq '&ohacgr;' );    # GREEK SMALL LETTER OMEGA WITH TONOS
    return '&#x03D1;' if ( $c eq '&thetav;' );    # GREEK THETA SYMBOL
    return '&#x03D2;' if ( $c eq '&Upsi;' );  # GREEK UPSILON WITH HOOK SYMBOL
    return '&#x03D5;' if ( $c eq '&phis;' );  # GREEK PHI SYMBOL
    return '&#x03D6;' if ( $c eq '&piv;' );   # GREEK PI SYMBOL
    return '&#x03DC;' if ( $c eq '&b.Gammad;' ); # GREEK LETTER DIGAMMA
    return '&#x03DD;' if ( $c eq '&b.gammad;' ); # GREEK SMALL LETTER DIGAMMA
    return '&#x03DD;' if ( $c eq '&gammad;' );   # GREEK SMALL LETTER DIGAMMA
    return '&#x03F0;' if ( $c eq '&kappav;' );   # GREEK KAPPA SYMBOL
    return '&#x03F1;' if ( $c eq '&rhov;' );     # GREEK RHO SYMBOL
    return '&#x03F5;' if ( $c eq '&epsi;' );     # GREEK LUNATE EPSILON SYMBOL
    return '&#x03F5;' if ( $c eq '&epsis;' );    # GREEK LUNATE EPSILON SYMBOL
    return '&#x03F6;'
        if ( $c eq '&bepsi;' );    # GREEK REVERSED LUNATE EPSILON SYMBOL
    return '&#x0401;' if ( $c eq '&IOcy;' );    # CYRILLIC CAPITAL LETTER IO
    return '&#x0402;' if ( $c eq '&DJcy;' );    # CYRILLIC CAPITAL LETTER DJE
    return '&#x0403;' if ( $c eq '&GJcy;' );    # CYRILLIC CAPITAL LETTER GJE
    return '&#x0404;'
        if ( $c eq '&Jukcy;' );    # CYRILLIC CAPITAL LETTER UKRAINIAN IE
    return '&#x0405;' if ( $c eq '&DScy;' );    # CYRILLIC CAPITAL LETTER DZE
    return '&#x0406;'
        if ( $c eq '&Iukcy;' )
        ;    # CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I
    return '&#x0407;' if ( $c eq '&YIcy;' );    # CYRILLIC CAPITAL LETTER YI
    return '&#x0408;' if ( $c eq '&Jsercy;' );  # CYRILLIC CAPITAL LETTER JE
    return '&#x0409;' if ( $c eq '&LJcy;' );    # CYRILLIC CAPITAL LETTER LJE
    return '&#x040A;' if ( $c eq '&NJcy;' );    # CYRILLIC CAPITAL LETTER NJE
    return '&#x040B;' if ( $c eq '&TSHcy;' );   # CYRILLIC CAPITAL LETTER TSHE
    return '&#x040C;' if ( $c eq '&KJcy;' );    # CYRILLIC CAPITAL LETTER KJE
    return '&#x040E;'
        if ( $c eq '&Ubrcy;' );    # CYRILLIC CAPITAL LETTER SHORT U
    return '&#x040F;' if ( $c eq '&DZcy;' ); # CYRILLIC CAPITAL LETTER DZHE
    return '&#x0410;' if ( $c eq '&Acy;' );  # CYRILLIC CAPITAL LETTER A
    return '&#x0411;' if ( $c eq '&Bcy;' );  # CYRILLIC CAPITAL LETTER BE
    return '&#x0412;' if ( $c eq '&Vcy;' );  # CYRILLIC CAPITAL LETTER VE
    return '&#x0413;' if ( $c eq '&Gcy;' );  # CYRILLIC CAPITAL LETTER GHE
    return '&#x0414;' if ( $c eq '&Dcy;' );  # CYRILLIC CAPITAL LETTER DE
    return '&#x0415;' if ( $c eq '&IEcy;' ); # CYRILLIC CAPITAL LETTER IE
    return '&#x0416;' if ( $c eq '&ZHcy;' ); # CYRILLIC CAPITAL LETTER ZHE
    return '&#x0417;' if ( $c eq '&Zcy;' );  # CYRILLIC CAPITAL LETTER ZE
    return '&#x0418;' if ( $c eq '&Icy;' );  # CYRILLIC CAPITAL LETTER I
    return '&#x0419;' if ( $c eq '&Jcy;' );  # CYRILLIC CAPITAL LETTER SHORT I
    return '&#x041A;' if ( $c eq '&Kcy;' );  # CYRILLIC CAPITAL LETTER KA
    return '&#x041B;' if ( $c eq '&Lcy;' );  # CYRILLIC CAPITAL LETTER EL
    return '&#x041C;' if ( $c eq '&Mcy;' );  # CYRILLIC CAPITAL LETTER EM
    return '&#x041D;' if ( $c eq '&Ncy;' );  # CYRILLIC CAPITAL LETTER EN
    return '&#x041E;' if ( $c eq '&Ocy;' );  # CYRILLIC CAPITAL LETTER O
    return '&#x041F;' if ( $c eq '&Pcy;' );  # CYRILLIC CAPITAL LETTER PE
    return '&#x0420;' if ( $c eq '&Rcy;' );  # CYRILLIC CAPITAL LETTER ER
    return '&#x0421;' if ( $c eq '&Scy;' );  # CYRILLIC CAPITAL LETTER ES
    return '&#x0422;' if ( $c eq '&Tcy;' );  # CYRILLIC CAPITAL LETTER TE
    return '&#x0423;' if ( $c eq '&Ucy;' );  # CYRILLIC CAPITAL LETTER U
    return '&#x0424;' if ( $c eq '&Fcy;' );  # CYRILLIC CAPITAL LETTER EF
    return '&#x0425;' if ( $c eq '&KHcy;' ); # CYRILLIC CAPITAL LETTER HA
    return '&#x0426;' if ( $c eq '&TScy;' ); # CYRILLIC CAPITAL LETTER TSE
    return '&#x0427;' if ( $c eq '&CHcy;' ); # CYRILLIC CAPITAL LETTER CHE
    return '&#x0428;' if ( $c eq '&SHcy;' ); # CYRILLIC CAPITAL LETTER SHA
    return '&#x0429;' if ( $c eq '&SHCHcy;' ); # CYRILLIC CAPITAL LETTER SHCHA
    return '&#x042A;'
        if ( $c eq '&HARDcy;' );    # CYRILLIC CAPITAL LETTER HARD SIGN
    return '&#x042B;' if ( $c eq '&Ycy;' );    # CYRILLIC CAPITAL LETTER YERU
    return '&#x042C;'
        if ( $c eq '&SOFTcy;' );    # CYRILLIC CAPITAL LETTER SOFT SIGN
    return '&#x042D;' if ( $c eq '&Ecy;' );    # CYRILLIC CAPITAL LETTER E
    return '&#x042E;' if ( $c eq '&YUcy;' );   # CYRILLIC CAPITAL LETTER YU
    return '&#x042F;' if ( $c eq '&YAcy;' );   # CYRILLIC CAPITAL LETTER YA
    return '&#x0430;' if ( $c eq '&acy;' );    # CYRILLIC SMALL LETTER A
    return '&#x0431;' if ( $c eq '&bcy;' );    # CYRILLIC SMALL LETTER BE
    return '&#x0432;' if ( $c eq '&vcy;' );    # CYRILLIC SMALL LETTER VE
    return '&#x0433;' if ( $c eq '&gcy;' );    # CYRILLIC SMALL LETTER GHE
    return '&#x0434;' if ( $c eq '&dcy;' );    # CYRILLIC SMALL LETTER DE
    return '&#x0435;' if ( $c eq '&iecy;' );   # CYRILLIC SMALL LETTER IE
    return '&#x0436;' if ( $c eq '&zhcy;' );   # CYRILLIC SMALL LETTER ZHE
    return '&#x0437;' if ( $c eq '&zcy;' );    # CYRILLIC SMALL LETTER ZE
    return '&#x0438;' if ( $c eq '&icy;' );    # CYRILLIC SMALL LETTER I
    return '&#x0439;' if ( $c eq '&jcy;' );    # CYRILLIC SMALL LETTER SHORT I
    return '&#x043A;' if ( $c eq '&kcy;' );    # CYRILLIC SMALL LETTER KA
    return '&#x043B;' if ( $c eq '&lcy;' );    # CYRILLIC SMALL LETTER EL
    return '&#x043C;' if ( $c eq '&mcy;' );    # CYRILLIC SMALL LETTER EM
    return '&#x043D;' if ( $c eq '&ncy;' );    # CYRILLIC SMALL LETTER EN
    return '&#x043E;' if ( $c eq '&ocy;' );    # CYRILLIC SMALL LETTER O
    return '&#x043F;' if ( $c eq '&pcy;' );    # CYRILLIC SMALL LETTER PE
    return '&#x0440;' if ( $c eq '&rcy;' );    # CYRILLIC SMALL LETTER ER
    return '&#x0441;' if ( $c eq '&scy;' );    # CYRILLIC SMALL LETTER ES
    return '&#x0442;' if ( $c eq '&tcy;' );    # CYRILLIC SMALL LETTER TE
    return '&#x0443;' if ( $c eq '&ucy;' );    # CYRILLIC SMALL LETTER U
    return '&#x0444;' if ( $c eq '&fcy;' );    # CYRILLIC SMALL LETTER EF
    return '&#x0445;' if ( $c eq '&khcy;' );   # CYRILLIC SMALL LETTER HA
    return '&#x0446;' if ( $c eq '&tscy;' );   # CYRILLIC SMALL LETTER TSE
    return '&#x0447;' if ( $c eq '&chcy;' );   # CYRILLIC SMALL LETTER CHE
    return '&#x0448;' if ( $c eq '&shcy;' );   # CYRILLIC SMALL LETTER SHA
    return '&#x0449;' if ( $c eq '&shchcy;' ); # CYRILLIC SMALL LETTER SHCHA
    return '&#x044A;'
        if ( $c eq '&hardcy;' );    # CYRILLIC SMALL LETTER HARD SIGN
    return '&#x044B;' if ( $c eq '&ycy;' );    # CYRILLIC SMALL LETTER YERU
    return '&#x044C;'
        if ( $c eq '&softcy;' );    # CYRILLIC SMALL LETTER SOFT SIGN
    return '&#x044D;' if ( $c eq '&ecy;' );     # CYRILLIC SMALL LETTER E
    return '&#x044E;' if ( $c eq '&yucy;' );    # CYRILLIC SMALL LETTER YU
    return '&#x044F;' if ( $c eq '&yacy;' );    # CYRILLIC SMALL LETTER YA
    return '&#x0451;' if ( $c eq '&iocy;' );    # CYRILLIC SMALL LETTER IO
    return '&#x0452;' if ( $c eq '&djcy;' );    # CYRILLIC SMALL LETTER DJE
    return '&#x0453;' if ( $c eq '&gjcy;' );    # CYRILLIC SMALL LETTER GJE
    return '&#x0454;'
        if ( $c eq '&jukcy;' );    # CYRILLIC SMALL LETTER UKRAINIAN IE
    return '&#x0455;' if ( $c eq '&dscy;' );    # CYRILLIC SMALL LETTER DZE
    return '&#x0456;'
        if ( $c eq '&iukcy;' )
        ;    # CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    return '&#x0457;' if ( $c eq '&yicy;' );   # CYRILLIC SMALL LETTER YI
    return '&#x0458;' if ( $c eq '&jsercy;' ); # CYRILLIC SMALL LETTER JE
    return '&#x0459;' if ( $c eq '&ljcy;' );   # CYRILLIC SMALL LETTER LJE
    return '&#x045A;' if ( $c eq '&njcy;' );   # CYRILLIC SMALL LETTER NJE
    return '&#x045B;' if ( $c eq '&tshcy;' );  # CYRILLIC SMALL LETTER TSHE
    return '&#x045C;' if ( $c eq '&kjcy;' );   # CYRILLIC SMALL LETTER KJE
    return '&#x045E;' if ( $c eq '&ubrcy;' );  # CYRILLIC SMALL LETTER SHORT U
    return '&#x045F;' if ( $c eq '&dzcy;' );   # CYRILLIC SMALL LETTER DZHE
    return '&#x2002;' if ( $c eq '&ensp;' );   # EN SPACE
    return '&#x2003;' if ( $c eq '&emsp;' );   # EM SPACE
    return '&#x2004;' if ( $c eq '&emsp13;' ); # THREE-PER-EM SPACE
    return '&#x2005;' if ( $c eq '&emsp14;' ); # FOUR-PER-EM SPACE
    return '&#x2007;' if ( $c eq '&numsp;' );  # FIGURE SPACE
    return '&#x2008;' if ( $c eq '&puncsp;' ); # PUNCTUATION SPACE
    return '&#x2009;' if ( $c eq '&thinsp;' ); # THIN SPACE
    return '&#x200A;' if ( $c eq '&hairsp;' ); # HAIR SPACE
    return '&#x2010;' if ( $c eq '&dash;' );   # HYPHEN
    return '&#x2010;' if ( $c eq '&hyphen;' ); # HYPHEN
    return '&#x2013;' if ( $c eq '&ndash;' );  # EN DASH
    return '&#x2014;' if ( $c eq '&mdash;' );  # EM DASH
    return '&#x2015;' if ( $c eq '&horbar;' ); # HORIZONTAL BAR
    return '&#x2016;' if ( $c eq '&Verbar;' ); # DOUBLE VERTICAL LINE
    return '&#x2018;' if ( $c eq '&lsquo;' );  # LEFT SINGLE QUOTATION MARK
    return '&#x2019;' if ( $c eq '&rsquo;' );  # RIGHT SINGLE QUOTATION MARK
    return '&#x2019;' if ( $c eq '&rsquor;' ); # RIGHT SINGLE QUOTATION MARK
    return '&#x201A;' if ( $c eq '&lsquor;' ); # SINGLE LOW-9 QUOTATION MARK
    return '&#x201C;' if ( $c eq '&ldquo;' );  # LEFT DOUBLE QUOTATION MARK
    return '&#x201D;' if ( $c eq '&rdquo;' );  # RIGHT DOUBLE QUOTATION MARK
    return '&#x201D;' if ( $c eq '&rdquor;' ); # RIGHT DOUBLE QUOTATION MARK
    return '&#x201E;' if ( $c eq '&ldquor;' ); # DOUBLE LOW-9 QUOTATION MARK
    return '&#x2020;' if ( $c eq '&dagger;' ); # DAGGER
    return '&#x2021;' if ( $c eq '&Dagger;' ); # DOUBLE DAGGER
    return '&#x2022;' if ( $c eq '&bull;' );   # BULLET
    return '&#x2025;' if ( $c eq '&nldr;' );   # TWO DOT LEADER
    return '&#x2026;' if ( $c eq '&hellip;' ); # HORIZONTAL ELLIPSIS
    return '&#x2026;' if ( $c eq '&mldr;' );   # HORIZONTAL ELLIPSIS
    return '&#x2030;' if ( $c eq '&permil;' ); # PER MILLE SIGN
    return '&#x2032;' if ( $c eq '&prime;' );  # PRIME
    return '&#x2032;' if ( $c eq '&vprime;' ); # PRIME
    return '&#x2033;' if ( $c eq '&Prime;' );  # DOUBLE PRIME
    return '&#x2034;' if ( $c eq '&tprime;' ); # TRIPLE PRIME
    return '&#x2035;' if ( $c eq '&bprime;' ); # REVERSED PRIME
    return '&#x2041;' if ( $c eq '&caret;' );  # CARET INSERTION POINT
    return '&#x2043;' if ( $c eq '&hybull;' ); # HYPHEN BULLET
    return '&#x20DB;' if ( $c eq '&tdot;' );   # COMBINING THREE DOTS ABOVE
    return '&#x20DC;' if ( $c eq '&DotDot;' ); # COMBINING FOUR DOTS ABOVE
    return '&#x2105;' if ( $c eq '&incare;' ); # CARE OF
    return '&#x210B;' if ( $c eq '&hamilt;' ); # SCRIPT CAPITAL H
    return '&#x210F;' if ( $c eq '&planck;' ); # PLANCK CONSTANT OVER TWO PI
    return '&#x2112;' if ( $c eq '&lagran;' ); # SCRIPT CAPITAL L
    return '&#x2113;' if ( $c eq '&ell;' );    # SCRIPT SMALL L
    return '&#x2116;' if ( $c eq '&numero;' ); # NUMERO SIGN
    return '&#x2117;' if ( $c eq '&copysr;' ); # SOUND RECORDING COPYRIGHT
    return '&#x2118;' if ( $c eq '&weierp;' ); # SCRIPT CAPITAL P
    return '&#x211C;' if ( $c eq '&real;' );   # BLACK-LETTER CAPITAL R
    return '&#x211E;' if ( $c eq '&rx;' );     # PRESCRIPTION TAKE
    return '&#x2122;' if ( $c eq '&trade;' );  # TRADE MARK SIGN
    return '&#x2126;' if ( $c eq '&ohm;' );    # OHM SIGN
    return '&#x212B;' if ( $c eq '&angst;' );  # ANGSTROM SIGN
    return '&#x212C;' if ( $c eq '&bernou;' ); # SCRIPT CAPITAL B
    return '&#x2133;' if ( $c eq '&phmmat;' ); # SCRIPT CAPITAL M
    return '&#x2134;' if ( $c eq '&order;' );  # SCRIPT SMALL O
    return '&#x2135;' if ( $c eq '&aleph;' );  # ALEF SYMBOL
    return '&#x2136;' if ( $c eq '&beth;' );   # BET SYMBOL
    return '&#x2137;' if ( $c eq '&gimel;' );  # GIMEL SYMBOL
    return '&#x2138;' if ( $c eq '&daleth;' ); # DALET SYMBOL
    return '&#x2153;' if ( $c eq '&frac13;' ); # VULGAR FRACTION ONE THIRD
    return '&#x2154;' if ( $c eq '&frac23;' ); # VULGAR FRACTION TWO THIRDS
    return '&#x2155;' if ( $c eq '&frac15;' ); # VULGAR FRACTION ONE FIFTH
    return '&#x2156;' if ( $c eq '&frac25;' ); # VULGAR FRACTION TWO FIFTHS
    return '&#x2157;' if ( $c eq '&frac35;' ); # VULGAR FRACTION THREE FIFTHS
    return '&#x2158;' if ( $c eq '&frac45;' ); # VULGAR FRACTION FOUR FIFTHS
    return '&#x2159;' if ( $c eq '&frac16;' ); # VULGAR FRACTION ONE SIXTH
    return '&#x215A;' if ( $c eq '&frac56;' ); # VULGAR FRACTION FIVE SIXTHS
    return '&#x215B;' if ( $c eq '&frac18;' ); # VULGAR FRACTION ONE EIGHTH
    return '&#x215C;' if ( $c eq '&frac38;' ); # VULGAR FRACTION THREE EIGHTHS
    return '&#x215D;' if ( $c eq '&frac58;' ); # VULGAR FRACTION FIVE EIGHTHS
    return '&#x215E;' if ( $c eq '&frac78;' ); # VULGAR FRACTION SEVEN EIGHTHS
    return '&#x2190;' if ( $c eq '&larr;' );   # LEFTWARDS ARROW
    return '&#x2191;' if ( $c eq '&uarr;' );   # UPWARDS ARROW
    return '&#x2192;' if ( $c eq '&rarr;' );   # RIGHTWARDS ARROW
    return '&#x2193;' if ( $c eq '&darr;' );   # DOWNWARDS ARROW
    return '&#x2194;' if ( $c eq '&harr;' );   # LEFT RIGHT ARROW
    return '&#x2195;' if ( $c eq '&varr;' );   # UP DOWN ARROW
    return '&#x2196;' if ( $c eq '&nwarr;' );  # NORTH WEST ARROW
    return '&#x2197;' if ( $c eq '&nearr;' );  # NORTH EAST ARROW
    return '&#x2198;' if ( $c eq '&drarr;' );  # SOUTH EAST ARROW
    return '&#x2199;' if ( $c eq '&dlarr;' );  # SOUTH WEST ARROW
    return '&#x219A;' if ( $c eq '&nlarr;' );  # LEFTWARDS ARROW WITH STROKE
    return '&#x219B;' if ( $c eq '&nrarr;' );  # RIGHTWARDS ARROW WITH STROKE
    return '&#x219D;' if ( $c eq '&rarrw;' );  # RIGHTWARDS WAVE ARROW
    return '&#x219E;' if ( $c eq '&Larr;' );   # LEFTWARDS TWO HEADED ARROW
    return '&#x21A0;' if ( $c eq '&Rarr;' );   # RIGHTWARDS TWO HEADED ARROW
    return '&#x21A2;' if ( $c eq '&larrtl;' ); # LEFTWARDS ARROW WITH TAIL
    return '&#x21A3;' if ( $c eq '&rarrtl;' ); # RIGHTWARDS ARROW WITH TAIL
    return '&#x21A6;' if ( $c eq '&map;' );    # RIGHTWARDS ARROW FROM BAR
    return '&#x21A9;' if ( $c eq '&larrhk;' ); # LEFTWARDS ARROW WITH HOOK
    return '&#x21AA;' if ( $c eq '&rarrhk;' ); # RIGHTWARDS ARROW WITH HOOK
    return '&#x21AB;' if ( $c eq '&larrlp;' ); # LEFTWARDS ARROW WITH LOOP
    return '&#x21AC;' if ( $c eq '&rarrlp;' ); # RIGHTWARDS ARROW WITH LOOP
    return '&#x21AD;' if ( $c eq '&harrw;' );  # LEFT RIGHT WAVE ARROW
    return '&#x21AE;' if ( $c eq '&nharr;' );  # LEFT RIGHT ARROW WITH STROKE
    return '&#x21B0;' if ( $c eq '&lsh;' ); # UPWARDS ARROW WITH TIP LEFTWARDS
    return '&#x21B1;'
        if ( $c eq '&rsh;' );    # UPWARDS ARROW WITH TIP RIGHTWARDS
    return '&#x21B6;'
        if ( $c eq '&cularr;' );    # ANTICLOCKWISE TOP SEMICIRCLE ARROW
    return '&#x21B7;'
        if ( $c eq '&curarr;' );    # CLOCKWISE TOP SEMICIRCLE ARROW
    return '&#x21BA;'
        if ( $c eq '&olarr;' );     # ANTICLOCKWISE OPEN CIRCLE ARROW
    return '&#x21BB;' if ( $c eq '&orarr;' );    # CLOCKWISE OPEN CIRCLE ARROW
    return '&#x21BC;'
        if ( $c eq '&lharu;' );    # LEFTWARDS HARPOON WITH BARB UPWARDS
    return '&#x21BD;'
        if ( $c eq '&lhard;' );    # LEFTWARDS HARPOON WITH BARB DOWNWARDS
    return '&#x21BE;'
        if ( $c eq '&uharr;' );    # UPWARDS HARPOON WITH BARB RIGHTWARDS
    return '&#x21BF;'
        if ( $c eq '&uharl;' );    # UPWARDS HARPOON WITH BARB LEFTWARDS
    return '&#x21C0;'
        if ( $c eq '&rharu;' );    # RIGHTWARDS HARPOON WITH BARB UPWARDS
    return '&#x21C1;'
        if ( $c eq '&rhard;' );    # RIGHTWARDS HARPOON WITH BARB DOWNWARDS
    return '&#x21C2;'
        if ( $c eq '&dharr;' );    # DOWNWARDS HARPOON WITH BARB RIGHTWARDS
    return '&#x21C3;'
        if ( $c eq '&dharl;' );    # DOWNWARDS HARPOON WITH BARB LEFTWARDS
    return '&#x21C4;'
        if ( $c eq '&rlarr2;' );    # RIGHTWARDS ARROW OVER LEFTWARDS ARROW
    return '&#x21C6;'
        if ( $c eq '&lrarr2;' );    # LEFTWARDS ARROW OVER RIGHTWARDS ARROW
    return '&#x21C7;' if ( $c eq '&larr2;' );    # LEFTWARDS PAIRED ARROWS
    return '&#x21C8;' if ( $c eq '&uarr2;' );    # UPWARDS PAIRED ARROWS
    return '&#x21C9;' if ( $c eq '&rarr2;' );    # RIGHTWARDS PAIRED ARROWS
    return '&#x21CA;' if ( $c eq '&darr2;' );    # DOWNWARDS PAIRED ARROWS
    return '&#x21CB;'
        if ( $c eq '&lrhar2;' );   # LEFTWARDS HARPOON OVER RIGHTWARDS HARPOON
    return '&#x21CC;'
        if ( $c eq '&rlhar2;' );   # RIGHTWARDS HARPOON OVER LEFTWARDS HARPOON
    return '&#x21CD;'
        if ( $c eq '&nlArr;' );    # LEFTWARDS DOUBLE ARROW WITH STROKE
    return '&#x21CE;'
        if ( $c eq '&nhArr;' );    # LEFT RIGHT DOUBLE ARROW WITH STROKE
    return '&#x21CF;'
        if ( $c eq '&nrArr;' );    # RIGHTWARDS DOUBLE ARROW WITH STROKE
    return '&#x21D0;' if ( $c eq '&lArr;' );     # LEFTWARDS DOUBLE ARROW
    return '&#x21D1;' if ( $c eq '&uArr;' );     # UPWARDS DOUBLE ARROW
    return '&#x21D2;' if ( $c eq '&rArr;' );     # RIGHTWARDS DOUBLE ARROW
    return '&#x21D3;' if ( $c eq '&dArr;' );     # DOWNWARDS DOUBLE ARROW
    return '&#x21D4;' if ( $c eq '&hArr;' );     # LEFT RIGHT DOUBLE ARROW
    return '&#x21D4;' if ( $c eq '&iff;' );      # LEFT RIGHT DOUBLE ARROW
    return '&#x21D5;' if ( $c eq '&vArr;' );     # UP DOWN DOUBLE ARROW
    return '&#x21DA;' if ( $c eq '&lAarr;' );    # LEFTWARDS TRIPLE ARROW
    return '&#x21DB;' if ( $c eq '&rAarr;' );    # RIGHTWARDS TRIPLE ARROW
    return '&#x2200;' if ( $c eq '&forall;' );   # FOR ALL
    return '&#x2201;' if ( $c eq '&comp;' );     # COMPLEMENT
    return '&#x2202;' if ( $c eq '&part;' );     # PARTIAL DIFFERENTIAL
    return '&#x2203;' if ( $c eq '&exist;' );    # THERE EXISTS
    return '&#x2204;' if ( $c eq '&nexist;' );   # THERE DOES NOT EXIST
    return '&#x2205;' if ( $c eq '&empty;' );    # EMPTY SET
    return '&#x2207;' if ( $c eq '&nabla;' );    # NABLA
    return '&#x2208;' if ( $c eq '&isin;' );     # ELEMENT OF
    return '&#x2209;' if ( $c eq '&notin;' );    # NOT AN ELEMENT OF
    return '&#x220B;' if ( $c eq '&ni;' );       # CONTAINS AS MEMBER
    return '&#x220F;' if ( $c eq '&prod;' );     # N-ARY PRODUCT
    return '&#x2210;' if ( $c eq '&coprod;' );   # N-ARY COPRODUCT
    return '&#x2210;' if ( $c eq '&samalg;' );   # N-ARY COPRODUCT
    return '&#x2211;' if ( $c eq '&sum;' );      # N-ARY SUMMATION
    return '&#x2212;' if ( $c eq '&minus;' );    # MINUS SIGN
    return '&#x2213;' if ( $c eq '&mnplus;' );   # MINUS-OR-PLUS SIGN
    return '&#x2214;' if ( $c eq '&plusdo;' );   # DOT PLUS
    return '&#x2216;' if ( $c eq '&setmn;' );    # SET MINUS
    return '&#x2216;' if ( $c eq '&ssetmn;' );   # SET MINUS
    return '&#x2217;' if ( $c eq '&lowast;' );   # ASTERISK OPERATOR
    return '&#x2218;' if ( $c eq '&compfn;' );   # RING OPERATOR
    return '&#x221A;' if ( $c eq '&radic;' );    # SQUARE ROOT
    return '&#x221D;' if ( $c eq '&prop;' );     # PROPORTIONAL TO
    return '&#x221D;' if ( $c eq '&vprop;' );    # PROPORTIONAL TO
    return '&#x221E;' if ( $c eq '&infin;' );    # INFINITY
    return '&#x221F;' if ( $c eq '&ang90;' );    # RIGHT ANGLE
    return '&#x2220;' if ( $c eq '&ang;' );      # ANGLE
    return '&#x2221;' if ( $c eq '&angmsd;' );   # MEASURED ANGLE
    return '&#x2222;' if ( $c eq '&angsph;' );   # SPHERICAL ANGLE
    return '&#x2223;' if ( $c eq '&mid;' );      # DIVIDES
    return '&#x2223;' if ( $c eq '&smid;' );     # DIVIDES
    return '&#x2224;' if ( $c eq '&nmid;' );     # DOES NOT DIVIDE
    return '&#x2224;' if ( $c eq '&nsmid;' );    # DOES NOT DIVIDE
    return '&#x2225;' if ( $c eq '&par;' );      # PARALLEL TO
    return '&#x2225;' if ( $c eq '&spar;' );     # PARALLEL TO
    return '&#x2226;' if ( $c eq '&npar;' );     # NOT PARALLEL TO
    return '&#x2226;' if ( $c eq '&nspar;' );    # NOT PARALLEL TO
    return '&#x2227;' if ( $c eq '&and;' );      # LOGICAL AND
    return '&#x2228;' if ( $c eq '&or;' );       # LOGICAL OR
    return '&#x2229;' if ( $c eq '&cap;' );      # INTERSECTION
    return '&#x222A;' if ( $c eq '&cup;' );      # UNION
    return '&#x222B;' if ( $c eq '&int;' );      # INTEGRAL
    return '&#x222E;' if ( $c eq '&conint;' );   # CONTOUR INTEGRAL
    return '&#x2234;' if ( $c eq '&there4;' );   # THEREFORE
    return '&#x2235;' if ( $c eq '&becaus;' );   # BECAUSE
    return '&#x223C;' if ( $c eq '&sim;' );      # TILDE OPERATOR
    return '&#x223C;' if ( $c eq '&thksim;' );   # TILDE OPERATOR
    return '&#x223D;' if ( $c eq '&bsim;' );     # REVERSED TILDE
    return '&#x2240;' if ( $c eq '&wreath;' );   # WREATH PRODUCT
    return '&#x2241;' if ( $c eq '&nsim;' );     # NOT TILDE
    return '&#x2243;' if ( $c eq '&sime;' );     # ASYMPTOTICALLY EQUAL TO
    return '&#x2244;' if ( $c eq '&nsime;' );    # NOT ASYMPTOTICALLY EQUAL TO
    return '&#x2245;' if ( $c eq '&cong;' );     # APPROXIMATELY EQUAL TO
    return '&#x2247;'
        if ( $c eq '&ncong;' );  # NEITHER APPROXIMATELY NOR ACTUALLY EQUAL TO
    return '&#x2248;' if ( $c eq '&ap;' );       # ALMOST EQUAL TO
    return '&#x2248;' if ( $c eq '&asymp;' );    # ALMOST EQUAL TO
    return '&#x2248;' if ( $c eq '&thkap;' );    # ALMOST EQUAL TO
    return '&#x2249;' if ( $c eq '&nap;' );      # NOT ALMOST EQUAL TO
    return '&#x224A;' if ( $c eq '&ape;' );      # ALMOST EQUAL OR EQUAL TO
    return '&#x224C;' if ( $c eq '&bcong;' );    # ALL EQUAL TO
    return '&#x224E;' if ( $c eq '&bump;' );     # GEOMETRICALLY EQUIVALENT TO
    return '&#x224F;' if ( $c eq '&bumpe;' );    # DIFFERENCE BETWEEN
    return '&#x2250;' if ( $c eq '&esdot;' );    # APPROACHES THE LIMIT
    return '&#x2251;' if ( $c eq '&eDot;' );     # GEOMETRICALLY EQUAL TO
    return '&#x2252;'
        if ( $c eq '&efDot;' );    # APPROXIMATELY EQUAL TO OR THE IMAGE OF
    return '&#x2253;'
        if ( $c eq '&erDot;' );    # IMAGE OF OR APPROXIMATELY EQUAL TO
    return '&#x2254;' if ( $c eq '&colone;' );    # COLON EQUALS
    return '&#x2255;' if ( $c eq '&ecolon;' );    # EQUALS COLON
    return '&#x2256;' if ( $c eq '&ecir;' );      # RING IN EQUAL TO
    return '&#x2257;' if ( $c eq '&cire;' );      # RING EQUAL TO
    return '&#x2259;' if ( $c eq '&wedgeq;' );    # ESTIMATES
    return '&#x225C;' if ( $c eq '&trie;' );      # DELTA EQUAL TO
    return '&#x2260;' if ( $c eq '&ne;' );        # NOT EQUAL TO
    return '&#x2261;' if ( $c eq '&equiv;' );     # IDENTICAL TO
    return '&#x2262;' if ( $c eq '&nequiv;' );    # NOT IDENTICAL TO
    return '&#x2264;' if ( $c eq '&le;' );        # LESS-THAN OR EQUAL TO
    return '&#x2265;' if ( $c eq '&ge;' );        # GREATER-THAN OR EQUAL TO
    return '&#x2266;' if ( $c eq '&lE;' );        # LESS-THAN OVER EQUAL TO
    return "&#x2266;&#x0338;"
        if ( $c eq "&nlE;" );    # LESS-THAN OVER EQUAL TO with slash
    return '&#x2267;' if ( $c eq '&gE;' );    # GREATER-THAN OVER EQUAL TO
    return "&#x2267;&#x0338;"
        if ( $c eq "&ngE;" );    # GREATER-THAN OVER EQUAL TO with slash
    return '&#x2268;' if ( $c eq '&lnE;' );    # LESS-THAN BUT NOT EQUAL TO
    return "&#x2268;&#xFE00;"
        if ( $c eq "&lvnE;" )
        ;    # LESS-THAN BUT NOT EQUAL TO - with vertical stroke
    return '&#x2269;' if ( $c eq '&gnE;' );    # GREATER-THAN BUT NOT EQUAL TO
    return "&#x2269;&#xFE00;"
        if ( $c eq "&gvnE;" )
        ;    # GREATER-THAN BUT NOT EQUAL TO - with vertical stroke
    return '&#x226A;' if ( $c eq '&Lt;' );    # MUCH LESS-THAN
    return '&#x226B;' if ( $c eq '&Gt;' );    # MUCH GREATER-THAN
    return '&#x226C;' if ( $c eq '&twixt;' ); # BETWEEN
    return '&#x226E;' if ( $c eq '&nlt;' );   # NOT LESS-THAN
    return '&#x226F;' if ( $c eq '&ngt;' );   # NOT GREATER-THAN
    return '&#x2270;' if ( $c eq '&nle;' );   # NEITHER LESS-THAN NOR EQUAL TO
    return '&#x2271;'
        if ( $c eq '&nge;' );    # NEITHER GREATER-THAN NOR EQUAL TO
    return '&#x2272;' if ( $c eq '&lsim;' );   # LESS-THAN OR EQUIVALENT TO
    return '&#x2273;' if ( $c eq '&gsim;' );   # GREATER-THAN OR EQUIVALENT TO
    return '&#x2276;' if ( $c eq '&lg;' );     # LESS-THAN OR GREATER-THAN
    return '&#x2277;' if ( $c eq '&gl;' );     # GREATER-THAN OR LESS-THAN
    return '&#x227A;' if ( $c eq '&pr;' );     # PRECEDES
    return '&#x227B;' if ( $c eq '&sc;' );     # SUCCEEDS
    return '&#x227C;' if ( $c eq '&cupre;' );  # PRECEDES OR EQUAL TO
    return '&#x227D;' if ( $c eq '&sccue;' );  # SUCCEEDS OR EQUAL TO
    return '&#x227E;' if ( $c eq '&prsim;' );  # PRECEDES OR EQUIVALENT TO
    return '&#x227F;' if ( $c eq '&scsim;' );  # SUCCEEDS OR EQUIVALENT TO
    return '&#x2280;' if ( $c eq '&npr;' );    # DOES NOT PRECEDE
    return '&#x2281;' if ( $c eq '&nsc;' );    # DOES NOT SUCCEED
    return '&#x2282;' if ( $c eq '&sub;' );    # SUBSET OF
    return '&#x2283;' if ( $c eq '&sup;' );    # SUPERSET OF
    return '&#x2284;' if ( $c eq '&nsub;' );   # NOT A SUBSET OF
    return '&#x2285;' if ( $c eq '&nsup;' );   # NOT A SUPERSET OF
    return '&#x2286;' if ( $c eq '&sube;' );   # SUBSET OF OR EQUAL TO
    return '&#x2287;' if ( $c eq '&supe;' );   # SUPERSET OF OR EQUAL TO
    return '&#x2288;'
        if ( $c eq '&nsube;' );    # NEITHER A SUBSET OF NOR EQUAL TO
    return '&#x2289;'
        if ( $c eq '&nsupe;' );    # NEITHER A SUPERSET OF NOR EQUAL TO
    return '&#x228A;' if ( $c eq '&subne;' );    # SUBSET OF WITH NOT EQUAL TO
    return "&#x228A;&#xFE00;"
        if ( $c eq "&vsubne;" )
        ; # SUBSET OF WITH NOT EQUAL TO - variant with stroke through bottom members
    return '&#x228B;' if ( $c eq '&supne;' );  # SUPERSET OF WITH NOT EQUAL TO
    return "&#x228B;&#xFE00;"
        if ( $c eq "&vsupne;" )
        ; # SUPERSET OF WITH NOT EQUAL TO - variant with stroke through bottom members
    return '&#x228E;' if ( $c eq '&uplus;' );    # MULTISET UNION
    return '&#x228F;' if ( $c eq '&sqsub;' );    # SQUARE IMAGE OF
    return '&#x2290;' if ( $c eq '&sqsup;' );    # SQUARE ORIGINAL OF
    return '&#x2291;' if ( $c eq '&sqsube;' );   # SQUARE IMAGE OF OR EQUAL TO
    return '&#x2292;'
        if ( $c eq '&sqsupe;' );    # SQUARE ORIGINAL OF OR EQUAL TO
    return '&#x2293;' if ( $c eq '&sqcap;' );     # SQUARE CAP
    return '&#x2294;' if ( $c eq '&sqcup;' );     # SQUARE CUP
    return '&#x2295;' if ( $c eq '&oplus;' );     # CIRCLED PLUS
    return '&#x2296;' if ( $c eq '&ominus;' );    # CIRCLED MINUS
    return '&#x2297;' if ( $c eq '&otimes;' );    # CIRCLED TIMES
    return '&#x2298;' if ( $c eq '&osol;' );      # CIRCLED DIVISION SLASH
    return '&#x2299;' if ( $c eq '&odot;' );      # CIRCLED DOT OPERATOR
    return '&#x229A;' if ( $c eq '&ocir;' );      # CIRCLED RING OPERATOR
    return '&#x229B;' if ( $c eq '&oast;' );      # CIRCLED ASTERISK OPERATOR
    return '&#x229D;' if ( $c eq '&odash;' );     # CIRCLED DASH
    return '&#x229E;' if ( $c eq '&plusb;' );     # SQUARED PLUS
    return '&#x229F;' if ( $c eq '&minusb;' );    # SQUARED MINUS
    return '&#x22A0;' if ( $c eq '&timesb;' );    # SQUARED TIMES
    return '&#x22A1;' if ( $c eq '&sdotb;' );     # SQUARED DOT OPERATOR
    return '&#x22A2;' if ( $c eq '&vdash;' );     # RIGHT TACK
    return '&#x22A3;' if ( $c eq '&dashv;' );     # LEFT TACK
    return '&#x22A4;' if ( $c eq '&top;' );       # DOWN TACK
    return '&#x22A5;' if ( $c eq '&bottom;' );    # UP TACK
    return '&#x22A5;' if ( $c eq '&perp;' );      # UP TACK
    return '&#x22A7;' if ( $c eq '&models;' );    # MODELS
    return '&#x22A8;' if ( $c eq '&vDash;' );     # TRUE
    return '&#x22A9;' if ( $c eq '&Vdash;' );     # FORCES
    return '&#x22AA;'
        if ( $c eq '&Vvdash;' );    # TRIPLE VERTICAL BAR RIGHT TURNSTILE
    return '&#x22AC;' if ( $c eq '&nvdash;' );    # DOES NOT PROVE
    return '&#x22AD;' if ( $c eq '&nvDash;' );    # NOT TRUE
    return '&#x22AE;' if ( $c eq '&nVdash;' );    # DOES NOT FORCE
    return '&#x22AF;'
        if ( $c eq '&nVDash;' )
        ;    # NEGATED DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE
    return '&#x22B2;' if ( $c eq '&vltri;' ); # NORMAL SUBGROUP OF
    return '&#x22B3;' if ( $c eq '&vrtri;' ); # CONTAINS AS NORMAL SUBGROUP
    return '&#x22B4;' if ( $c eq '&ltrie;' ); # NORMAL SUBGROUP OF OR EQUAL TO
    return '&#x22B5;'
        if ( $c eq '&rtrie;' );    # CONTAINS AS NORMAL SUBGROUP OR EQUAL TO
    return '&#x22B8;' if ( $c eq '&mumap;' );     # MULTIMAP
    return '&#x22BA;' if ( $c eq '&intcal;' );    # INTERCALATE
    return '&#x22BB;' if ( $c eq '&veebar;' );    # XOR
    return '&#x22C4;' if ( $c eq '&diam;' );      # DIAMOND OPERATOR
    return '&#x22C5;' if ( $c eq '&sdot;' );      # DOT OPERATOR
    return '&#x22C6;' if ( $c eq '&sstarf;' );    # STAR OPERATOR
    return '&#x22C7;' if ( $c eq '&divonx;' );    # DIVISION TIMES
    return '&#x22C8;' if ( $c eq '&bowtie;' );    # BOWTIE
    return '&#x22C9;'
        if ( $c eq '&ltimes;' );    # LEFT NORMAL FACTOR SEMIDIRECT PRODUCT
    return '&#x22CA;'
        if ( $c eq '&rtimes;' );    # RIGHT NORMAL FACTOR SEMIDIRECT PRODUCT
    return '&#x22CB;' if ( $c eq '&lthree;' );    # LEFT SEMIDIRECT PRODUCT
    return '&#x22CC;' if ( $c eq '&rthree;' );    # RIGHT SEMIDIRECT PRODUCT
    return '&#x22CD;' if ( $c eq '&bsime;' );     # REVERSED TILDE EQUALS
    return '&#x22CE;' if ( $c eq '&cuvee;' );     # CURLY LOGICAL OR
    return '&#x22CF;' if ( $c eq '&cuwed;' );     # CURLY LOGICAL AND
    return '&#x22D0;' if ( $c eq '&Sub;' );       # DOUBLE SUBSET
    return '&#x22D1;' if ( $c eq '&Sup;' );       # DOUBLE SUPERSET
    return '&#x22D2;' if ( $c eq '&Cap;' );       # DOUBLE INTERSECTION
    return '&#x22D3;' if ( $c eq '&Cup;' );       # DOUBLE UNION
    return '&#x22D4;' if ( $c eq '&fork;' );      # PITCHFORK
    return '&#x22D6;' if ( $c eq '&ldot;' );      # LESS-THAN WITH DOT
    return '&#x22D7;' if ( $c eq '&gsdot;' );     # GREATER-THAN WITH DOT
    return '&#x22D8;' if ( $c eq '&Ll;' );        # VERY MUCH LESS-THAN
    return '&#x22D9;' if ( $c eq '&Gg;' );        # VERY MUCH GREATER-THAN
    return '&#x22DA;'
        if ( $c eq '&leg;' );    # LESS-THAN EQUAL TO OR GREATER-THAN
    return '&#x22DB;'
        if ( $c eq '&gel;' );    # GREATER-THAN EQUAL TO OR LESS-THAN
    return '&#x22DE;' if ( $c eq '&cuepr;' );    # EQUAL TO OR PRECEDES
    return '&#x22DF;' if ( $c eq '&cuesc;' );    # EQUAL TO OR SUCCEEDS
    return '&#x22E6;'
        if ( $c eq '&lnsim;' );    # LESS-THAN BUT NOT EQUIVALENT TO
    return '&#x22E7;'
        if ( $c eq '&gnsim;' );    # GREATER-THAN BUT NOT EQUIVALENT TO
    return '&#x22E8;'
        if ( $c eq '&prnsim;' );    # PRECEDES BUT NOT EQUIVALENT TO
    return '&#x22E9;'
        if ( $c eq '&scnsim;' );    # SUCCEEDS BUT NOT EQUIVALENT TO
    return '&#x22EA;' if ( $c eq '&nltri;' );    # NOT NORMAL SUBGROUP OF
    return '&#x22EB;'
        if ( $c eq '&nrtri;' );    # DOES NOT CONTAIN AS NORMAL SUBGROUP
    return '&#x22EC;'
        if ( $c eq '&nltrie;' );    # NOT NORMAL SUBGROUP OF OR EQUAL TO
    return '&#x22ED;'
        if ( $c eq '&nrtrie;' )
        ;    # DOES NOT CONTAIN AS NORMAL SUBGROUP OR EQUAL
    return '&#x22EE;' if ( $c eq '&vellip;' );  # VERTICAL ELLIPSIS
    return '&#x2305;' if ( $c eq '&barwed;' );  # PROJECTIVE
    return '&#x2306;' if ( $c eq '&Barwed;' );  # PERSPECTIVE
    return '&#x2308;' if ( $c eq '&lceil;' );   # LEFT CEILING
    return '&#x2309;' if ( $c eq '&rceil;' );   # RIGHT CEILING
    return '&#x230A;' if ( $c eq '&lfloor;' );  # LEFT FLOOR
    return '&#x230B;' if ( $c eq '&rfloor;' );  # RIGHT FLOOR
    return '&#x230C;' if ( $c eq '&drcrop;' );  # BOTTOM RIGHT CROP
    return '&#x230D;' if ( $c eq '&dlcrop;' );  # BOTTOM LEFT CROP
    return '&#x230E;' if ( $c eq '&urcrop;' );  # TOP RIGHT CROP
    return '&#x230F;' if ( $c eq '&ulcrop;' );  # TOP LEFT CROP
    return '&#x2315;' if ( $c eq '&telrec;' );  # TELEPHONE RECORDER
    return '&#x2316;' if ( $c eq '&target;' );  # POSITION INDICATOR
    return '&#x231C;' if ( $c eq '&ulcorn;' );  # TOP LEFT CORNER
    return '&#x231D;' if ( $c eq '&urcorn;' );  # TOP RIGHT CORNER
    return '&#x231E;' if ( $c eq '&dlcorn;' );  # BOTTOM LEFT CORNER
    return '&#x231F;' if ( $c eq '&drcorn;' );  # BOTTOM RIGHT CORNER
    return '&#x2322;' if ( $c eq '&frown;' );   # FROWN
    return '&#x2322;' if ( $c eq '&sfrown;' );  # FROWN
    return '&#x2323;' if ( $c eq '&smile;' );   # SMILE
    return '&#x2323;' if ( $c eq '&ssmile;' );  # SMILE
    return '&#x2329;' if ( $c eq '&lang;' );    # LEFT-POINTING ANGLE BRACKET
    return '&#x232A;' if ( $c eq '&rang;' );    # RIGHT-POINTING ANGLE BRACKET
    return '&#x2423;' if ( $c eq '&blank;' );   # OPEN BOX
    return '&#x24C8;' if ( $c eq '&oS;' );    # CIRCLED LATIN CAPITAL LETTER S
    return '&#x2500;' if ( $c eq '&boxh;' );  # BOX DRAWINGS LIGHT HORIZONTAL
    return '&#x2502;' if ( $c eq '&boxv;' );  # BOX DRAWINGS LIGHT VERTICAL
    return '&#x250C;'
        if ( $c eq '&boxdr;' );    # BOX DRAWINGS LIGHT DOWN AND RIGHT
    return '&#x2510;'
        if ( $c eq '&boxdl;' );    # BOX DRAWINGS LIGHT DOWN AND LEFT
    return '&#x2514;'
        if ( $c eq '&boxur;' );    # BOX DRAWINGS LIGHT UP AND RIGHT
    return '&#x2518;' if ( $c eq '&boxul;' ); # BOX DRAWINGS LIGHT UP AND LEFT
    return '&#x251C;'
        if ( $c eq '&boxvr;' );    # BOX DRAWINGS LIGHT VERTICAL AND RIGHT
    return '&#x2524;'
        if ( $c eq '&boxvl;' );    # BOX DRAWINGS LIGHT VERTICAL AND LEFT
    return '&#x252C;'
        if ( $c eq '&boxhd;' );    # BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
    return '&#x2534;'
        if ( $c eq '&boxhu;' );    # BOX DRAWINGS LIGHT UP AND HORIZONTAL
    return '&#x253C;'
        if ( $c eq '&boxvh;' );   # BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
    return '&#x2550;' if ( $c eq '&boxH;' );  # BOX DRAWINGS DOUBLE HORIZONTAL
    return '&#x2551;' if ( $c eq '&boxV;' );  # BOX DRAWINGS DOUBLE VERTICAL
    return '&#x2552;'
        if ( $c eq '&boxdR;' );    # BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
    return '&#x2553;'
        if ( $c eq '&boxDr;' );    # BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
    return '&#x2554;'
        if ( $c eq '&boxDR;' );    # BOX DRAWINGS DOUBLE DOWN AND RIGHT
    return '&#x2555;'
        if ( $c eq '&boxdL;' );    # BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
    return '&#x2556;'
        if ( $c eq '&boxDl;' );    # BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
    return '&#x2557;'
        if ( $c eq '&boxDL;' );    # BOX DRAWINGS DOUBLE DOWN AND LEFT
    return '&#x2558;'
        if ( $c eq '&boxuR;' );    # BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
    return '&#x2559;'
        if ( $c eq '&boxUr;' );    # BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
    return '&#x255A;'
        if ( $c eq '&boxUR;' );    # BOX DRAWINGS DOUBLE UP AND RIGHT
    return '&#x255B;'
        if ( $c eq '&boxuL;' );    # BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
    return '&#x255C;'
        if ( $c eq '&boxUl;' );    # BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
    return '&#x255D;'
        if ( $c eq '&boxUL;' );    # BOX DRAWINGS DOUBLE UP AND LEFT
    return '&#x255E;'
        if ( $c eq '&boxvR;' )
        ;    # BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
    return '&#x255F;'
        if ( $c eq '&boxVr;' )
        ;    # BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
    return '&#x2560;'
        if ( $c eq '&boxVR;' );    # BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
    return '&#x2561;'
        if ( $c eq '&boxvL;' ); # BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
    return '&#x2562;'
        if ( $c eq '&boxVl;' ); # BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
    return '&#x2563;'
        if ( $c eq '&boxVL;' );    # BOX DRAWINGS DOUBLE VERTICAL AND LEFT
    return '&#x2564;'
        if ( $c eq '&boxHd;' )
        ;    # BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
    return '&#x2565;'
        if ( $c eq '&boxhD;' )
        ;    # BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
    return '&#x2566;'
        if ( $c eq '&boxHD;' );    # BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
    return '&#x2567;'
        if ( $c eq '&boxHu;' ); # BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
    return '&#x2568;'
        if ( $c eq '&boxhU;' ); # BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
    return '&#x2569;'
        if ( $c eq '&boxHU;' );    # BOX DRAWINGS DOUBLE UP AND HORIZONTAL
    return '&#x256A;'
        if ( $c eq '&boxvH;' )
        ;    # BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
    return '&#x256B;'
        if ( $c eq '&boxVh;' )
        ;    # BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
    return '&#x256C;'
        if ( $c eq '&boxVH;' );  # BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
    return '&#x2580;' if ( $c eq '&uhblk;' );     # UPPER HALF BLOCK
    return '&#x2584;' if ( $c eq '&lhblk;' );     # LOWER HALF BLOCK
    return '&#x2588;' if ( $c eq '&block;' );     # FULL BLOCK
    return '&#x2591;' if ( $c eq '&blk14;' );     # LIGHT SHADE
    return '&#x2592;' if ( $c eq '&blk12;' );     # MEDIUM SHADE
    return '&#x2593;' if ( $c eq '&blk34;' );     # DARK SHADE
    return '&#x25A1;' if ( $c eq '&squ;' );       # WHITE SQUARE
    return '&#x25A1;' if ( $c eq '&square;' );    # WHITE SQUARE
    return '&#x25AA;' if ( $c eq '&squf;' );      # BLACK SMALL SQUARE
    return '&#x25AD;' if ( $c eq '&rect;' );      # WHITE RECTANGLE
    return '&#x25AE;' if ( $c eq '&marker;' );    # BLACK VERTICAL RECTANGLE
    return '&#x25B3;' if ( $c eq '&xutri;' );     # WHITE UP-POINTING TRIANGLE
    return '&#x25B4;'
        if ( $c eq '&utrif;' );    # BLACK UP-POINTING SMALL TRIANGLE
    return '&#x25B5;'
        if ( $c eq '&utri;' );     # WHITE UP-POINTING SMALL TRIANGLE
    return '&#x25B8;'
        if ( $c eq '&rtrif;' );    # BLACK RIGHT-POINTING SMALL TRIANGLE
    return '&#x25B9;'
        if ( $c eq '&rtri;' );     # WHITE RIGHT-POINTING SMALL TRIANGLE
    return '&#x25BD;' if ( $c eq '&xdtri;' );   # WHITE DOWN-POINTING TRIANGLE
    return '&#x25BE;'
        if ( $c eq '&dtrif;' );    # BLACK DOWN-POINTING SMALL TRIANGLE
    return '&#x25BF;'
        if ( $c eq '&dtri;' );     # WHITE DOWN-POINTING SMALL TRIANGLE
    return '&#x25C2;'
        if ( $c eq '&ltrif;' );    # BLACK LEFT-POINTING SMALL TRIANGLE
    return '&#x25C3;'
        if ( $c eq '&ltri;' );     # WHITE LEFT-POINTING SMALL TRIANGLE
    return '&#x25CA;' if ( $c eq '&loz;' );     # LOZENGE
    return '&#x25CB;' if ( $c eq '&cir;' );     # WHITE CIRCLE
    return '&#x25EF;' if ( $c eq '&xcirc;' );   # LARGE CIRCLE
    return '&#x2605;' if ( $c eq '&starf;' );   # BLACK STAR
    return '&#x2606;' if ( $c eq '&star;' );    # WHITE STAR
    return '&#x260E;' if ( $c eq '&phone;' );   # BLACK TELEPHONE
    return '&#x2640;' if ( $c eq '&female;' );  # FEMALE SIGN
    return '&#x2642;' if ( $c eq '&male;' );    # MALE SIGN
    return '&#x2660;' if ( $c eq '&spades;' );  # BLACK SPADE SUIT
    return '&#x2663;' if ( $c eq '&clubs;' );   # BLACK CLUB SUIT
    return '&#x2665;' if ( $c eq '&hearts;' );  # BLACK HEART SUIT
    return '&#x2666;' if ( $c eq '&diams;' );   # BLACK DIAMOND SUIT
    return '&#x266A;' if ( $c eq '&sung;' );    # EIGHTH NOTE
    return '&#x266D;' if ( $c eq '&flat;' );    # MUSIC FLAT SIGN
    return '&#x266E;' if ( $c eq '&natur;' );   # MUSIC NATURAL SIGN
    return '&#x266F;' if ( $c eq '&sharp;' );   # MUSIC SHARP SIGN
    return '&#x2713;' if ( $c eq '&check;' );   # CHECK MARK
    return '&#x2717;' if ( $c eq '&cross;' );   # BALLOT X
    return '&#x2720;' if ( $c eq '&malt;' );    # MALTESE CROSS
    return '&#x2736;' if ( $c eq '&sext;' );    # SIX POINTED BLACK STAR
    return '&#x27F7;' if ( $c eq '&xharr;' );   # LONG LEFT RIGHT ARROW
    return '&#x27F8;' if ( $c eq '&xlArr;' );   # LONG LEFTWARDS DOUBLE ARROW
    return '&#x27F9;' if ( $c eq '&xrArr;' );   # LONG RIGHTWARDS DOUBLE ARROW
    return '&#x27FA;' if ( $c eq '&xhArr;' );   # LONG LEFT RIGHT DOUBLE ARROW
    return '&#x2994;'
        if ( $c eq '&rpargt;' );    # RIGHT ARC GREATER-THAN BRACKET
    return '&#x29A0;' if ( $c eq '&lpargt;' ); # SPHERICAL ANGLE OPENING LEFT
    return '&#x29EB;' if ( $c eq '&lozf;' );   # BLACK LOZENGE
    return '&#x2A3F;' if ( $c eq '&amalg;' );  # AMALGAMATION OR COPRODUCT
    return '&#x2A7D;' if ( $c eq '&les;' );    # LESS-THAN OR SLANTED EQUAL TO
    return "&#x2A7D;&#x0338;"
        if ( $c eq "&nles;" );    # LESS-THAN OR SLANTED EQUAL TO with slash
    return '&#x2A7E;' if ( $c eq '&ges;' ); # GREATER-THAN OR SLANTED EQUAL TO
    return "&#x2A7E;&#x0338;"
        if ( $c eq "&nges;" );   # GREATER-THAN OR SLANTED EQUAL TO with slash
    return '&#x2A85;' if ( $c eq '&lap;' );    # LESS-THAN OR APPROXIMATE
    return '&#x2A86;' if ( $c eq '&gap;' );    # GREATER-THAN OR APPROXIMATE
    return '&#x2A87;'
        if ( $c eq '&lne;' );    # LESS-THAN AND SINGLE-LINE NOT EQUAL TO
    return '&#x2A88;'
        if ( $c eq '&gne;' );    # GREATER-THAN AND SINGLE-LINE NOT EQUAL TO
    return '&#x2A89;' if ( $c eq '&lnap;' );   # LESS-THAN AND NOT APPROXIMATE
    return '&#x2A8A;'
        if ( $c eq '&gnap;' );    # GREATER-THAN AND NOT APPROXIMATE
    return '&#x2A8B;'
        if ( $c eq '&lEg;' )
        ;    # LESS-THAN ABOVE DOUBLE-LINE EQUAL ABOVE GREATER-THAN
    return '&#x2A8C;'
        if ( $c eq '&gEl;' )
        ;    # GREATER-THAN ABOVE DOUBLE-LINE EQUAL ABOVE LESS-THAN
    return '&#x2A95;' if ( $c eq '&els;' ); # SLANTED EQUAL TO OR LESS-THAN
    return '&#x2A96;' if ( $c eq '&egs;' ); # SLANTED EQUAL TO OR GREATER-THAN
    return '&#x2AAF;'
        if ( $c eq '&pre;' );    # PRECEDES ABOVE SINGLE-LINE EQUALS SIGN
    return "&#x2AAF;&#x0338;"
        if ( $c eq "&npre;" )
        ;    # PRECEDES ABOVE SINGLE-LINE EQUALS SIGN with slash
    return '&#x2AB0;'
        if ( $c eq '&sce;' );    # SUCCEEDS ABOVE SINGLE-LINE EQUALS SIGN
    return "&#x2AB0;&#x0338;"
        if ( $c eq "&nsce;" )
        ;    # SUCCEEDS ABOVE SINGLE-LINE EQUALS SIGN with slash
    return '&#x2AB5;' if ( $c eq '&prnE;' );  # PRECEDES ABOVE NOT EQUAL TO
    return '&#x2AB6;' if ( $c eq '&scnE;' );  # SUCCEEDS ABOVE NOT EQUAL TO
    return '&#x2AB7;' if ( $c eq '&prap;' );  # PRECEDES ABOVE ALMOST EQUAL TO
    return '&#x2AB8;' if ( $c eq '&scap;' );  # SUCCEEDS ABOVE ALMOST EQUAL TO
    return '&#x2AB9;'
        if ( $c eq '&prnap;' );    # PRECEDES ABOVE NOT ALMOST EQUAL TO
    return '&#x2ABA;'
        if ( $c eq '&scnap;' );    # SUCCEEDS ABOVE NOT ALMOST EQUAL TO
    return '&#x2AC5;' if ( $c eq '&subE;' );    # SUBSET OF ABOVE EQUALS SIGN
    return "&#x2AC5;&#x0338;"
        if ( $c eq "&nsubE;" );    # SUBSET OF ABOVE EQUALS SIGN with slash
    return '&#x2AC6;' if ( $c eq '&supE;' );   # SUPERSET OF ABOVE EQUALS SIGN
    return "&#x2AC6;&#x0338;"
        if ( $c eq "&nsupE;" );    # SUPERSET OF ABOVE EQUALS SIGN with slash
    return '&#x2ACB;' if ( $c eq '&subnE;' );   # SUBSET OF ABOVE NOT EQUAL TO
    return "&#x2ACB;&#xFE00;"
        if ( $c eq "&vsubnE;" )
        ; # SUBSET OF ABOVE NOT EQUAL TO - variant with stroke through bottom members
    return '&#x2ACC;' if ( $c eq '&supnE;' ); # SUPERSET OF ABOVE NOT EQUAL TO
    return "&#x2ACC;&#xFE00;"
        if ( $c eq "&vsupnE;" )
        ; # SUPERSET OF ABOVE NOT EQUAL TO - variant with stroke through bottom members
    return '&#xFB00;' if ( $c eq '&fflig;' );     # LATIN SMALL LIGATURE FF
    return '&#xFB01;' if ( $c eq '&filig;' );     # LATIN SMALL LIGATURE FI
    return '&#xFB02;' if ( $c eq '&fllig;' );     # LATIN SMALL LIGATURE FL
    return '&#xFB03;' if ( $c eq '&ffilig;' );    # LATIN SMALL LIGATURE FFI
    return '&#xFB04;' if ( $c eq '&ffllig;' );    # LATIN SMALL LIGATURE FFL
    return '&#xFE68;' if ( $c eq '&sbsol;' );     # SMALL REVERSE SOLIDUS
    return '&#x1D6AA;'
        if ( $c eq '&b.Gamma;' );    # MATHEMATICAL BOLD CAPITAL GAMMA
    return '&#x1D6AB;'
        if ( $c eq '&b.Delta;' );    # MATHEMATICAL BOLD CAPITAL DELTA
    return '&#x1D6AF;'
        if ( $c eq '&b.Theta;' );    # MATHEMATICAL BOLD CAPITAL THETA
    return '&#x1D6B2;'
        if ( $c eq '&b.Lambda;' );    # MATHEMATICAL BOLD CAPITAL LAMDA
    return '&#x1D6B5;' if ( $c eq '&b.Xi;' );   # MATHEMATICAL BOLD CAPITAL XI
    return '&#x1D6B7;' if ( $c eq '&b.Pi;' );   # MATHEMATICAL BOLD CAPITAL PI
    return '&#x1D6BA;'
        if ( $c eq '&b.Sigma;' );    # MATHEMATICAL BOLD CAPITAL SIGMA
    return '&#x1D6BC;'
        if ( $c eq '&b.Upsi;' );     # MATHEMATICAL BOLD CAPITAL UPSILON
    return '&#x1D6BD;' if ( $c eq '&b.Phi;' ); # MATHEMATICAL BOLD CAPITAL PHI
    return '&#x1D6BF;' if ( $c eq '&b.Psi;' ); # MATHEMATICAL BOLD CAPITAL PSI
    return '&#x1D6C0;'
        if ( $c eq '&b.Omega;' );    # MATHEMATICAL BOLD CAPITAL OMEGA
    return '&#x1D6C2;'
        if ( $c eq '&b.alpha;' );    # MATHEMATICAL BOLD SMALL ALPHA
    return '&#x1D6C3;' if ( $c eq '&b.beta;' ); # MATHEMATICAL BOLD SMALL BETA
    return '&#x1D6C4;'
        if ( $c eq '&b.gamma;' );    # MATHEMATICAL BOLD SMALL GAMMA
    return '&#x1D6C5;'
        if ( $c eq '&b.delta;' );    # MATHEMATICAL BOLD SMALL DELTA
    return '&#x1D6C6;'
        if ( $c eq '&b.epsi;' );     # MATHEMATICAL BOLD SMALL EPSILON
    return '&#x1D6C7;' if ( $c eq '&b.zeta;' ); # MATHEMATICAL BOLD SMALL ZETA
    return '&#x1D6C8;' if ( $c eq '&b.eta;' );  # MATHEMATICAL BOLD SMALL ETA
    return '&#x1D6C9;'
        if ( $c eq '&b.thetas;' );    # MATHEMATICAL BOLD SMALL THETA
    return '&#x1D6CA;' if ( $c eq '&b.iota;' ); # MATHEMATICAL BOLD SMALL IOTA
    return '&#x1D6CB;'
        if ( $c eq '&b.kappa;' );    # MATHEMATICAL BOLD SMALL KAPPA
    return '&#x1D6CC;'
        if ( $c eq '&b.lambda;' );    # MATHEMATICAL BOLD SMALL LAMDA
    return '&#x1D6CD;' if ( $c eq '&b.mu;' );    # MATHEMATICAL BOLD SMALL MU
    return '&#x1D6CE;' if ( $c eq '&b.nu;' );    # MATHEMATICAL BOLD SMALL NU
    return '&#x1D6CF;' if ( $c eq '&b.xi;' );    # MATHEMATICAL BOLD SMALL XI
    return '&#x1D6D1;' if ( $c eq '&b.pi;' );    # MATHEMATICAL BOLD SMALL PI
    return '&#x1D6D2;' if ( $c eq '&b.rho;' );   # MATHEMATICAL BOLD SMALL RHO
    return '&#x1D6D3;'
        if ( $c eq '&b.sigmav;' );    # MATHEMATICAL BOLD SMALL FINAL SIGMA
    return '&#x1D6D4;'
        if ( $c eq '&b.sigma;' );     # MATHEMATICAL BOLD SMALL SIGMA
    return '&#x1D6D5;' if ( $c eq '&b.tau;' );   # MATHEMATICAL BOLD SMALL TAU
    return '&#x1D6D6;'
        if ( $c eq '&b.upsi;' );    # MATHEMATICAL BOLD SMALL UPSILON
    return '&#x1D6D7;' if ( $c eq '&b.phi;' );   # MATHEMATICAL BOLD SMALL PHI
    return '&#x1D6D8;' if ( $c eq '&b.chi;' );   # MATHEMATICAL BOLD SMALL CHI
    return '&#x1D6D9;' if ( $c eq '&b.psi;' );   # MATHEMATICAL BOLD SMALL PSI
    return '&#x1D6DA;'
        if ( $c eq '&b.omega;' );    # MATHEMATICAL BOLD SMALL OMEGA
    return '&#x1D6DC;'
        if ( $c eq '&b.epsiv;' );    # MATHEMATICAL BOLD EPSILON SYMBOL
    return '&#x1D6DD;'
        if ( $c eq '&b.thetav;' );    # MATHEMATICAL BOLD THETA SYMBOL
    return '&#x1D6DE;'
        if ( $c eq '&b.kappav;' );    # MATHEMATICAL BOLD KAPPA SYMBOL
    return '&#x1D6DF;' if ( $c eq '&b.phiv;' ); # MATHEMATICAL BOLD PHI SYMBOL
    return '&#x1D6E0;' if ( $c eq '&b.rhov;' ); # MATHEMATICAL BOLD RHO SYMBOL
    return '&#x1D6E1;' if ( $c eq '&b.piv;' );  # MATHEMATICAL BOLD PI SYMBOL
    return $c;
}

##############################################################################
#
# Function: numeric2ascii
#
# Receives an XML entity. If the entity is a numeric entity representing
# a safe XML value, return the ASCII value. Best used like this:
# $content =~ s{(&\#x?[A-fa-f\d]+;)}{numeric2ascii($1)}xmsge;
#
##############################################################################

sub numeric2ascii {
    my $c = shift;
    my $v = $c;
    $v =~ s{&#([A-fa-fx\d]+);}{$1};
    $v = hex $v if ( $v =~ s{^x}{} );
    return $c     if ( $v < $SHORTCUT_CHAR_LENGTH_NINE );
    return chr $v if ( $v == $SHORTCUT_CHAR_LENGTH_NINE );  # HORIZONTAL TAB
    return chr $v if ( $v == $SHORTCUT_CHAR_LENGTH_TEN );   # LINE FEED
    return $c     if ( $v == $ELEVEN );
    return $c     if ( $v == $TWELVE );
    return chr $v if ( $v == $THIRTEEN );                   # CARRIAGE RETURN
    return $c     if ( $v > $THIRTEEN && $v < $THREETWO );
    return $c     if ( $v == $THREEFOUR );                  # QUOTATION MARK
    return $c     if ( $v == $THREEEIGHT );                 # AMPERSAND
    return $c     if ( $v == $THREENINE );                  # APOSTROPHE-QUOTE
    return $c     if ( $v == $SHORTCUT_CHAR_LENGTH_SIXTY ); # LESS-THAN SIGN
    return $c if ( $v == $SIXTWO );     # GREATER-THAN SIGN
    return $c if ( $v > $ONETWOSIX );
    return chr $v;
}

##############################################################################
# Function: correctrepresentationhtml
# Parameters:
#
# Description: Represent HTML tags correctly
##############################################################################

sub correctrepresentationhtml {
    my $datacontent = shift;

    $datacontent = encode_entities( decode_entities($datacontent) );
    $datacontent =~ s/&lt;([^&]+)&gt;/<$1>/gis;
    $datacontent =~ s/&quot;/"/gis;
    return $datacontent;
}

##############################################################################
# Function: verify_chunk
# Parameters:
#
# Description: Verify chunk and remove the duplicate starting part.
##############################################################################

sub verify_chunk {
    my $hd = shift;
    my $bd = shift;
    $hd = removehtmlmarkupbodyall($hd);
    $bd = removehtmlmarkupbodyall($bd);
    if ( $hd =~ m/^\Q$bd\E$/ ) {
        return 0;
    }
    $bd =~ s/^\Q$hd\E//gis;
    $bd =~ s/^(\s)+//gis;
    $bd =~ s/(\s)+$//gis;
    if ( $hd =~ m/^\Q$bd\E$/sxm ) {
        return 0;
    }
    return 1;
}

##############################################################################
# Function: get_descriptor
# Parameters:
#     Part Number
# Description:
#     Takes the Library Part Number and generates and fetches the XML
##############################################################################

sub get_descriptor {
    my $librarypn = shift;
    $librarypn = format_pn($librarypn);
    my $xml_data       = q();
    my $pdb_url_to_use = q();

    my $ua = LWP::UserAgent->new();

    if ( $pdb_url_to_use =~ m{ stdoc }xms ) {
        $ua->credentials( 'stdoc.us.oracle.com', q(), q(), q() );
    }
    else {
        $ua->credentials( 'pdb.us.oracle.com', q(), q(), q() );
    }

    $ua->proxy( [ 'http', 'ftp' ], 'http://www-proxy.us.oracle.com:80' );
    $ua->no_proxy(
        'oracle.com',       'oracleleads.com',
        'us.oracle.com',    'uk.oracle.com',
        'ca.oracle.com',    'oraclecorp.com',
        'oracleportal.com', 'ie.oracle.com'
    );

    print 'Posting: ' . $pdb_url . 'ip.jsp?pn=' . $librarypn . "\n";

    my $response = $ua->post( $pdb_url . 'ip.jsp?pn=' . $librarypn.'&c=y' );

    if ( $response->is_success() ) {

        # So the XML is now generated on the server. Get it
        print 'Reading XML: '
            . $pdb_url . 'xml/'
            . format_pn_docarch($librarypn) . '.xml';

        $response = $ua->get(
            $pdb_url . 'xml/' . format_pn_docarch($librarypn) . '.xml' );

        if ( $response->is_success() ) {
            $xml_data = $response->content();
        }
        else {
            $errors{'Error while reading the descriptor'}
                = $response->status_line;
        }

    }
    else {
        $errors{'Error while fetching the descriptor'}
            = $response->status_line;
    }

    return $xml_data;
}

##############################################################################
# Function: format_pn
# Parameters:
#     Part Number
# Description:
#     Takes the part number and uppercases and converts the underscore to hyphen
##############################################################################

sub format_pn {
    my $raw_pn = shift;

    $raw_pn = uc $raw_pn;
    $raw_pn =~ s{_}{-}xms;

    return $raw_pn;
}

##############################################################################
# Function: format_pn_docarch
# Parameters:
#     Part Number
# Description:
#     Takes the part number and uppercases and converts the hyphen to underscore
##############################################################################

sub format_pn_docarch {
    my $raw_pn = shift;

    $raw_pn = uc $raw_pn;
    $raw_pn =~ s{ - }{_}xms;

    return $raw_pn;
}

##############################################################################
# Function: format_pn_lower
# Parameters:
#     Part Number
# Description:
#     Takes the part number and lowercases and removes hyphen/underscore
##############################################################################

sub format_pn_lower {
    my $raw_pn = shift;

    $raw_pn = lc $raw_pn;
    $raw_pn =~ s{ [-_](.*) }{}xms;
    if ( $raw_pn =~ m/mglos/isxm ) {
        $raw_pn = 'm-glos';
    }
    if ( $raw_pn =~ m/mindx/isxm ) {
        $raw_pn = 'm-indx';
    }
    return $raw_pn;
}

##############################################################################
# Function: print_errors
# Parameters:
#
# Description: Print the errors
##############################################################################

sub print_errors {

    for my $keys ( keys %errors ) {
        print 'Error: ' . $keys . "\n";
        print 'Description: ' . $errors{$keys} . "\n";
    }

    exit;
}

##############################################################################
# Function: build_data
# Parameters:
#     part number
# Description:
#     Takes a part number and builds up the library data hash
##############################################################################

sub build_data {
    my $library_pn  = shift;
    my $scan_folder = shift;
    my $xml_data    = get_descriptor($library_pn);

    if ( ( scalar keys %errors ) > 0 ) {
        #print_errors();
    }
    else {
        parse_xml( $xml_data, $scan_folder );
        if ( ( scalar keys %errors ) > 0 ) {
            print_errors();
        }
    }

    # At this point, there would be the xml file in the string. Parse it!

    # At this point, the data hash is built!
    return;
}

##############################################################################
# Function: parse_xml
# Parameters:
#     XML Data
# Description:
#     Takes the XML String and validates, parses, and builds the data hash
##############################################################################

sub parse_xml {
    my $xml_data    = shift;
    my $scan_folder = shift;
    my $parser      = XML::LibXML->new();
    my $doc;
    my $counter  = 0;
    my $alltotal = 0;
    $doc = $parser->parse_string($xml_data);

    my $library = $doc->getDocumentElement();
    $library->findnodes('library');
    my $prodattr   = 'product';
    my $relnumattr = 'releasenumber';
    my $relvalattr = 'releasevalue';
    my $partnoattr = 'partno';
    my $titleattr  = 'title';
    my $prodtitle  = $library->findvalue( q(@) . $prodattr );
    for ( $library->findnodes('category') ) {

        for ( $_->findnodes('document') ) {
            my $partno_add = $_->findvalue('@partno');
            my $complete_pn = $partno_add;
            $partno_add = lc $partno_add;
            $partno_add =~ s/-(.*)//sxm;
            $partno_add =~ s/_(.*)//sxm;
            my $docid_add = $_->findvalue('@docid');
            $docid_add = uc $docid_add;

            my $title_add = $_->findvalue('@title');
            my $title_prod = $_->findvalue('@product');
            $title_add =~ s/^$title_prod//i;
            $title_add =~ s/^( )+//;
            $title_add =~ s/( )+$//;
            $title_add = encode_entities_html($title_add);
            my $product_name_title_add = $_->findvalue('@product');
            my $book_release           = $_->findvalue('@releasevalue');
            $product_name_title_add =~ s/\+/\\+/gis;
            $product_name_title_add =~ s/\*/\\*/gis;
            $product_name_title_add
                = encode_entities_html($product_name_title_add);
            $additional_data{$partno_add} = [
                $complete_pn, $docid_add,
                $title_add,  $product_name_title_add,
                $book_release
            ];
            $additional_data{$docid_add} = [
                $complete_pn, $docid_add,
                $title_add,  $product_name_title_add,
                $book_release
            ];

        }
    }

    $prodtitle = encode_entities_html($prodtitle);
    my $rn = $library->findvalue( q(@) . $relnumattr );
    $rn =~ s{ [.] }{}xmsg;
    my $librel          = $library->findvalue( q(@) . $relvalattr );
    my $initial_release = q();
    my $final_release   = q();
    if ($librel) {
        if ( $librel =~ m/(([\d]+)([[:lower:]]+))(\s)+/isxm ) {
            my $rel_first = $1;
            $initial_release = $rel_first;
            my $digit_rel   = $2;
            my $charitalics = $3;
            $final_release = $digit_rel . q(<i>) . $charitalics . q(</i>);
            $librel =~ s/$1/$digit_rel<i>$charitalics<\/i>/gis;
        }
    }
    $library_data{ $library->findvalue( q(@) . $partnoattr ) } = [
        $prodtitle . q( ) . $library->findvalue( q(@) . $titleattr ),  # Title
        $prodtitle,    # Product Name
        q(),           # Platform
        q(),           # feature
        q(),           # category
        ##$library->findvalue( q(@) . $relvalattr ),    # Release Value
        $librel,       # Release Value
    ];

    ## build data from PDB as well so as to fill the missing values

    # now scan the html files and build the info.
    find( { wanted => \&wanted_all_files, no_chdir => 1 }, $scan_folder );

    my $partno_scanned   = q();
    my $doctitle_scanned = q();
    my $prod_scanned     = q();
    my $platform_scanned = q();
    my $feature_scanned  = q();
    my $category_scanned = q();
    my $release_scanned  = q();
    my $docid            = q();
    my $firsth1          = q();

    for my $htm_file_of_lib (@html_files_from_library) {
        ##print "Scanning: $htm_file_of_lib\n";
        open my $SCANNER, '<', $htm_file_of_lib
            or carp "Can't read $htm_file_of_lib: $OS_ERROR";
        my @file_scanned = <$SCANNER>;
        close $SCANNER
            or carp "Can't close $htm_file_of_lib: $OS_ERROR";

        my $scanned_file_joined = join q(), @file_scanned;
        if (   $scanned_file_joined =~ m{ <[?]xml [^>]+ encoding [[:space:]]* = [[:space:]]* ['"] utf -? 8 }xmsi
            || $scanned_file_joined =~ m{ <meta [^>]+ charset [[:space:]]* = [[:space:]]* utf -? 8 }xmsi )
        {
            $scanned_file_joined = Encode::decode( 'utf8', $scanned_file_joined );
        }


        my $idstream = HTML::TokeParser->new( \$scanned_file_joined )
            or print "Couldn't parse: $OS_ERROR";
        my $tocleftvalign = 0;
        my $maintitlefrompage = q();
        while ( my $token = $idstream->get_token() ) {

            if (   length $partno_scanned > 0
                && length $doctitle_scanned > 0
                && length $prod_scanned > 0
                && length $docid > 0
                && length $release_scanned > 0 )
            {
                last;
            }
            if ( !$token->[0] || !$token->[1] || !$token->[2] ) {
                next;
            }
            if ($token->[0] eq 'S'    # doctitle
                && $token->[1] eq 'title'
                )
            {
                my $pagetitle = $idstream->get_trimmed_text('/title');
                if ( $pagetitle =~ m/(.*)(master(\s)*glossary)(.*)/isxm ) {
                    $doctitle_scanned = $2;
                    $release_scanned  = $4;
                    $prod_scanned     = $1;
                    $docid            = 'DOCID';
                    $partno_scanned   = 'MGLOS';
                    last;
                }
                elsif ( $pagetitle =~ m/(.*)(master(\s)*index)(.*)/isxm ) {
                    $doctitle_scanned = $2;
                    $release_scanned  = $4;
                    $prod_scanned     = $1;
                    $docid            = 'DOCID';
                    $partno_scanned   = 'MINDX';
                    last;
                }
                elsif ( lc $pagetitle ne 'contents' ) {
                    $doctitle_scanned = $pagetitle;
                }
                ##print "Initial title: $doctitle_scanned\n";
            }
            if ($token->[0] eq 'S'    # doctitle
                && $token->[1] eq 'meta'
                && exists $token->[2]{'name'}
                && ($token->[2]{'name'} eq 'docid' || $token->[2]{'name'} =~ m/isVersionOf/)
                )
            {
                $docid = $token->[2]{'content'};
            }
            if ($token->[0] eq 'S'    # doctitle
                && $token->[1] eq 'meta'
                && exists $token->[2]{'name'}
                && ($token->[2]{'name'} =~ m/title/)
                )
            {
                $maintitlefrompage = $token->[2]{'content'};
                ## add logic to extract the release number from here
            }
            if ($token->[0] eq 'S'    # doctitle
                && $token->[1] eq 'meta'
                && exists $token->[2]{'name'}
                && ($token->[2]{'name'} eq 'doctitle' || $token->[2]{'name'} =~ m/dcterms.title/)
                )
            {
                $doctitle_scanned = $token->[2]{'content'};
                ## add logic to extract the release number from here
            }

            if ($token->[0] eq 'S'    # partno
                && $token->[1] eq 'meta'
                && exists $token->[2]{'name'}
                && ($token->[2]{'name'} eq 'partno' || $token->[2]{'name'} =~ m/partno/ || $token->[2]{'name'} =~ m/dcterms.identifier/)
                )
            {
                $partno_scanned = $token->[2]{'content'};
            }
            if ($token->[0] eq 'S'    # docinfo
                && $token->[1] eq 'h1'
                && exists $token->[2]{'class'}
                && $token->[2]{'class'} eq 'docinfo'
                )
            {
                $prod_scanned = $idstream->get_trimmed_text('/h1');
            }
            if ($token->[0] eq 'S'    # docinfo
                && $token->[1] eq 'h1'
                )
            {
                $firsth1 = $idstream->get_trimmed_text('/h1');
            }
            if ($token->[0] eq 'S'    # prodname
                && ( $token->[1] eq 'span' || $token->[1] eq 'p' || $token->[1] eq 'div')
                && exists $token->[2]{'class'}
                && lc $token->[2]{'class'} eq 'booktitle' ##&& length ($doctitle_scanned) < 1
                )
            {
                $doctitle_scanned
                    = $idstream->get_trimmed_text( q(/) . $token->[1] );
                if ( length $firsth1 > 0 ) {
                    $doctitle_scanned = $firsth1 . q( ) . $doctitle_scanned;
                    $prod_scanned     = $firsth1;
                }
                ##print "BOOKTITLE: $doctitle_scanned\n";
            }

            if ($token->[0] eq 'S'    # prodname
                && $token->[1] eq 'td'
                && exists $token->[2]{'align'}
                && lc $token->[2]{'align'} eq 'left'
                && exists $token->[2]{'valign'}
                && lc $token->[2]{'valign'} eq 'top'
                && $tocleftvalign < 1
                )
            {
                $tocleftvalign++;
                $doctitle_scanned = $idstream->get_trimmed_text('br');
                $idstream->get_token();
                $release_scanned = $idstream->get_trimmed_text('/b');
                if ( $release_scanned =~ m/(.*?)(for(.*))/isxm ) {
                    $release_scanned = $1;
                    my $platformhere = $2;
                    $release_scanned =~ s{ / \z}{}xms;
                    $doctitle_scanned
                        = $doctitle_scanned . q( ) . $platformhere;
                    $doctitle_scanned =~ s{ / \z}{}xms;
                }
                $idstream->get_token();
                $partno_scanned = $idstream->get_trimmed_text('/td');
                if ( length $docid < 1 ) {
                    $docid = 'DOCID';
                }
                if ( length $partno_scanned < 1 ) {
                    $partno_scanned = 'PARTNO-' . $docid;
                }
                $prod_scanned = q(Oracle&reg;);
                ##print "Entered 1 here\n";
            }

            if ($token->[0] eq 'S'    # prodname
                && ( $token->[1] eq 'span' || $token->[1] eq 'p' )
                && exists $token->[2]{'class'}
                && lc $token->[2]{'class'} eq 'invpartnumber'
                )
            {
                $partno_scanned
                    = $idstream->get_trimmed_text( q(/) . $token->[1] );
            }
            if ($token->[0] eq 'S'    # prodname
                && ( $token->[1] eq 'span' || $token->[1] eq 'p' )
                && exists $token->[2]{'class'}
                && lc $token->[2]{'class'} eq 'revnumber'
                )
            {
                $release_scanned = $idstream->get_trimmed_text('/p');
                $release_scanned =~ s/for(.*)//gis;
                $release_scanned =~ s/^(\s)*//gis;
            }

            if ($token->[0] eq 'S'    # prodname
                && ( $token->[1] eq 'td' )
                && exists $token->[2]{'class'}
                && $token->[2]{'class'} eq 'zz-nav-header-cell'
                )
            {
                my $completetext = $idstream->get_trimmed_text('/td');
                if ( $completetext
                    =~ m/(.*?)(release[ ]+[\d]+(.*?))((.*?)($PARTNO))/gis )
                {
                    $release_scanned  = $3;
                    $partno_scanned   = $6;
                    $doctitle_scanned = $1;
                    $release_scanned =~ s/[\s]+$//gis;
                    $release_scanned =~ s/^[\s]+//gis;
                    $release_scanned =~ s/[\s]+/ /gis;
                    $partno_scanned =~ s/[\s]+$//gis;
                    $partno_scanned =~ s/^[\s]+//gis;
                    $partno_scanned =~ s/[\s]+/ /gis;
                    $doctitle_scanned =~ s/[\s]+$//gis;
                    $doctitle_scanned =~ s/^[\s]+//gis;
                    $doctitle_scanned =~ s/[\s]+/ /gis;
                    $doctitle_scanned =~ s/^Oracle([^(\s)])*(\s)*//xmsgi;
                    $doctitle_scanned =~ s{\A Oracle }{}xmsgi;
                    $prod_scanned = $doctitle_scanned;
                    $docid        = q(DOCID);
                }
            }

            if ( $htm_file_of_lib =~ m/constant-values/isxm )    #javadoc
            {
                my $titleval = q();
                if (   $token->[0] eq 'S'
                    && $token->[1] eq 'td'
                    && exists $token->[2]{'align'}
                    && $token->[2]{'align'} eq 'right'
                    && exists $token->[2]{'valign'}
                    && $token->[2]{'valign'} eq 'top'
                    && exists $token->[2]{'rowspan'}
                    && $token->[2]{'rowspan'} eq '3' )
                {
       # prodname would have got extracted. remove the "Constant Field Values"
                    ##$titleval = $idstream->get_trimmed_text('/td');
                    $doctitle_scanned = $idstream->get_trimmed_text('br');
                    $idstream->get_token();
                    $release_scanned = $idstream->get_trimmed_text('br');
                    $idstream->get_token();
                    $partno_scanned = $idstream->get_trimmed_text('/td');
                    $docid          = 'DOCID';
                }
                elsif ($token->[0] eq 'S'
                    && $token->[1] eq 'div'
                    && exists $token->[2]{'class'}
                    && $token->[2]{'class'} eq 'aboutLanguage' )
                {
                    $doctitle_scanned = $idstream->get_trimmed_text('br');
                    $idstream->get_token();
                    $release_scanned = $idstream->get_trimmed_text('br');
                    $idstream->get_token();
                    $partno_scanned = $idstream->get_trimmed_text('/div');
                    $docid          = 'DOCID';
                }
                ##elsif ($token->[0] eq 'S'
                ##    && $token->[1] eq 'div'
                ##    && exists $token->[2]{'class'}
                ##    && $token->[2]{'class'} eq 'aboutLanguage' )
                ##{
                ##    $titleval = $idstream->get_trimmed_text('/div');
                ##}

                if (length $titleval > 0
                    && (   length $partno_scanned < 1
                        || length $release_scanned < 1
                        || length $doctitle_scanned < 1 )
                    )
                {
                    $prod_scanned =~ s/\QConstant Field Values\E//gis;
                    $prod_scanned =~ s/^(\W)+/ /gis;
                    $prod_scanned =~ s/(\W)+$/ /gis;
                    $prod_scanned =~ s/^\s+//gis;
                    $prod_scanned =~ s/\s+$//gis;
                    $prod_scanned
                        =~ s/(([[:alpha:]])(.*)([[:alpha:]]))/$1/gis;
                    $prod_scanned =~ s/^Oracle([^(\s)])*(\s)*//xmsgi;
                    $prod_scanned =~ s{\A Oracle }{}xmsgi;

                    # Remove Oracle (and/or reg symbol)
                    my $titlevaltemp = $titleval;
                    $titlevaltemp =~ s/^Oracle([^(\s)])*(\s)*//xmsgi;
                    $titlevaltemp =~ s{\A Oracle }{}xmsgi;
                    $titlevaltemp =~ s/\Q$prod_scanned\E//xmsgi;
                    $titlevaltemp =~ s/^(\s)*//gis;
                    $titlevaltemp =~ s/(\s)*$//gis;
                    if ( $titlevaltemp =~ m/($PARTNO)/xms ) {
                        $partno_scanned = $1;
                        $titlevaltemp =~ s/\Q$partno_scanned\E//gis;
                        $titlevaltemp =~ s/^(\s)*//gis;
                        $titlevaltemp =~ s/(\s)*$//gis;
                        $release_scanned = $titlevaltemp;
                        $titleval =~ s/\Q$partno_scanned\E//gis;
                        $titleval =~ s/\Q$release_scanned\E//gis;
                        $titleval =~ s/^(\s)*//gis;
                        $titleval =~ s/(\s)*$//gis;
                        $doctitle_scanned = $titleval;
                        $docid            = 'DOCID';
                    }
                }
            }
        }
        if (length $doctitle_scanned < 1 && length $maintitlefrompage > 0) {
            $doctitle_scanned = $maintitlefrompage;
        }

        # enter the values to library_data
        if (   length $partno_scanned < 1
            || length $doctitle_scanned < 1 )
        {
            print
                "RETRYING: $partno_scanned : $doctitle_scanned : $prod_scanned : $docid : $release_scanned\n";
            $idstream = HTML::TokeParser->new( \$scanned_file_joined )
                or print "Couldn't parse: $OS_ERROR";
            while ( my $token = $idstream->get_token() ) {
                my $alltext = $idstream->get_trimmed_text('/body');
                if ( $alltext =~ m/($PARTNO)/gis ) {
                    $partno_scanned = $1;
                    print q(Part No.: ) . $partno_scanned . "\n";
                    last;
                }
            }
            if ( length $doctitle_scanned > 0 && length $prod_scanned < 1 ) {
                $prod_scanned = $doctitle_scanned;
                $docid        = q(DOCID);
            }
            if ( length $doctitle_scanned < 1 && length $prod_scanned > 0 ) {
                $doctitle_scanned = $prod_scanned;
                $docid            = q(DOCID);
            }
        }
        if (   length $partno_scanned > 0
            && length $doctitle_scanned > 0 )
        {
            if ( $partno_scanned =~ m/($PARTNO)/xms ) {
                $partno_scanned = $1;
            }
            $prod_scanned = encode_entities_html($prod_scanned);
            $doctitle_scanned
                = encode_entities_html($doctitle_scanned);
            if ( $doctitle_scanned
                =~ m/Oracle[&\#\d;[:alpha:]]+(\s)*Database/gis )
            {
                $doctitle_scanned
                    =~ s{Oracle[&\#\d;[:alpha:]]+(\s)*Database}{}xmsg;
            }
            $doctitle_scanned =~ s{\A Oracle® }{}xmsgi;
            $doctitle_scanned =~ s{\A Oracle[&\#\d;[:alpha:]]+ }{}xmsgi;
            $doctitle_scanned =~ s{\A Oracle }{}xmsgi;
            $doctitle_scanned =~ s{ \Q$release_scanned\E}{}xmsgi;
            $doctitle_scanned =~ s{ \(\s*\)}{}xmsgi;

            $doctitle_scanned =~ s{\A \s+ }{}xmsgi;
            $doctitle_scanned =~ s{ \s+ \z}{}xmsgi;
            my $complete_pn = $partno_scanned;
            $partno_scanned =~ s{ _ .* }{}xmsgi;
            $partno_scanned =~ s{ - .* }{}xmsgi;
            $partno_scanned = lc $partno_scanned;
            $counter++;

            if ( length $initial_release > 0 && length $final_release > 0 ) {
                $doctitle_scanned =~ s/$initial_release/$final_release/gis;
            }

            $library_data{$partno_scanned} = [
                appropriate_titles($doctitle_scanned),
                $prod_scanned,     $platform_scanned, $feature_scanned,
                $category_scanned, $release_scanned,  $docid,
                $complete_pn
            ];
            $library_data_docid{$docid} = [
                appropriate_titles($doctitle_scanned),
                $prod_scanned,     $platform_scanned, $feature_scanned,
                $category_scanned, $release_scanned,  $docid,
                $complete_pn
            ];
            my $whichfolder = $htm_file_of_lib;
            $whichfolder =~ s/^[.]\///gis;
            $whichfolder =~ s/^\///gis;
            $whichfolder =~ s/(.*)\/(.*)/$1/gis;
            ##print "Adding $whichfolder for $partno_scanned and $doctitle_scanned\n";
            $folderpnmapping{$whichfolder} = $partno_scanned;
        }
        else {
            print
                "FAILED: $partno_scanned : $doctitle_scanned : $prod_scanned : $docid : $release_scanned\n";
        }
        $partno_scanned   = q();
        $doctitle_scanned = q();
        $prod_scanned     = q();
        $platform_scanned = q();
        $feature_scanned  = q();
        $category_scanned = q();
        $release_scanned  = q();
        $docid            = q();
    }
    return;
}

##############################################################################
# Function: appropriate_titles
# Parameters:
#
# Description: Convert the titles to appropriate representation.
##############################################################################

sub appropriate_titles {
    my $raw_title = shift;
    $raw_title =~ s/&\#x00AE;//gis;
    $raw_title =~ s/&reg;//gis;
    if (   $raw_title =~ m/^Client(\s)+Installation/isxm
        || $raw_title =~ m/^Client(\s)+Quick(\s)+Installation/isxm
        || $raw_title =~ m/^Client(\s)+Quick(\s)+Installation/isxm
        || $raw_title =~ m/^Clusterware/isxm
        || $raw_title =~ m/^Examples/isxm
        || $raw_title =~ m/^Gateway/isxm
        || $raw_title =~ m/^Installation(\s)+Guide/isxm
        || $raw_title =~ m/^Quick(\s)+Installation(\s)+Guide/isxm
        || $raw_title =~ m/^Vault(\s)+/isxm )
    {
        $raw_title = 'Database ' . $raw_title;
    }

    if (   $raw_title =~ m/^C[+][+](\s)+Call(\s)+Interface/isxm
        || $raw_title =~ m/^Call(\s)+Interface(\s)+Programmer/isxm )
    {
        $raw_title = 'Oracle ' . $raw_title;
    }

    if ( $raw_title =~ m/Java(\s)+API(\s)+Reference$/isxm ) {
        $raw_title = $raw_title . ' (Javadoc)';
    }
    $raw_title =~ s/,(\s)*/ /isxm;
    $raw_title =~ s/^(\s)*//isxm;
    $raw_title =~ s/(\s)*$//isxm;
    return $raw_title;
}

##############################################################################
#
# Function: wanted_all_files
#
# Parameter: filename
# builds toc and title array
#
##############################################################################

sub wanted_all_files {
    my $temp = $_;

    if (   $temp =~ m{  title[.]html? }xms
        || $temp =~ m{ toc[.]html? }xms )
    {
        if ( $temp =~ m/(.*)toc[.]ht[m|ml]/gis ) {
            ##print "Passed: $temp\n";
            my $checkpath = $1;
            my $keeppath  = $checkpath;
            $checkpath = $checkpath . 'title.htm';
            $keeppath  = $keeppath . 'constant-values.html';
            if ( -f $checkpath ) {
                if ( !any { $_ eq $checkpath } @html_files_from_library ) {
                    ##print "Entering 1: $temp\n";
                    push @html_files_from_library, $checkpath;
                }
            }
            elsif ( -f $keeppath ) {
                if ( !any { $_ eq $keeppath } @html_files_from_library ) {
                    ##print "Entering 2: $temp\n";
                    push @html_files_from_library, $keeppath;
                }
            }
            elsif ( !any { $_ eq $checkpath } @html_files_from_library ) {
                    ##print "Entering 3: $temp\n";
                push @html_files_from_library, $temp;
            }
        }
        else {
            ##print "Entering 4: $temp\n";
            push @html_files_from_library, $temp;
        }
    }

    return;
}



##############################################################################
#
# Function: get_topic_type
#
# Parameter: sec_title, book_name, fname_type, topic_class
# determines the topic type for the search result.
#
##############################################################################

sub get_topic_type {
   my $sec_title     = shift;
   my $book_title    = shift;
   my $fname_type    = shift;
   my $topic_class   = shift;
   my $type_of_topic = q();

   my $is_concept_type = is_concepts($sec_title, $book_title, $fname_type);
   my $is_task_type    = is_tasks($sec_title);
   my $is_ref_type     = is_references($book_title, $sec_title);
   my $is_gloss_type   = is_glossary($topic_class);
   my $is_chapter_type = is_chapter($topic_class);
   my $is_example_type = is_example($sec_title);
   my $is_faq_type     = is_faq($sec_title);
   my $is_error_type   = is_error_message($sec_title, $book_title);

   if ($is_concept_type eq q(1)) {$topic_types_found{q(concepts)}++; $type_of_topic .= q(,concepts);}
   if ($is_task_type eq q(1)) {$topic_types_found{q(tasks)}++; $type_of_topic .= q(,tasks);}
   if ($is_ref_type eq q(1)) {$topic_types_found{q(references)}++; $type_of_topic .= q(,references);}
   if ($is_gloss_type eq q(1)) {$topic_types_found{q(glossary)}++; $type_of_topic .= q(,glossary);}
   if ($is_chapter_type eq q(1)) {$topic_types_found{q(chapter)}++; $type_of_topic .= q(,chapter);}
   if ($is_example_type eq q(1)) {$topic_types_found{q(example)}++; $type_of_topic .= q(,example);}
   if ($is_faq_type eq q(1)) {$topic_types_found{q(faq)}++; $type_of_topic .= q(,faq);}
   if ($is_error_type eq q(1)) {$topic_types_found{q(error)}++; $type_of_topic .= q(,error_message);}

   $type_of_topic =~ s/^,//sxm;
   $type_of_topic =~ s/,$//sxm;
   return $type_of_topic;
}

##############################################################################
#
# Function: is_concepts
#
# Parameter: sec_title, book_name, fname_type
# Determines if the topic is of type concepts
##############################################################################

sub is_concepts {
   my $sec_title = shift;
   my $book_title = shift;
   my $fname_type = shift;
   my $is_intro = q(0);
   if ($fname_type =~ m/toc[.]htm/sxm) {return q(0);}
   my @concepts_strings = ('^Why _se(\s)*.*', '^Features.*', '^Advantages.*', '^Benefits.*',
                           '^Introduc.*', '^What _s(\s)*.*', '^What _re(\s)*.*', '^What(\s)*.*?',
                           '^Who(\s)*.*?', '^When(\s)*.*?', '^When _o(\s)*.*', '^Where(\s)*.*?',
                           '^Why(\s)*.*?', '^Learning.*', '^Choosing.*', '^Understanding.*',
                           '.*Deciding.*','^About.*','.*Planning.*','.*Overview.*',
                           '.*Typical.*','.*Concept.*','.*Guidelines.*','.*Highlights.*',
                           '.*Consideration.*','.*Restriction.*','.*Limits.*',
                           '.*Limitations.*','.*Architectur.*','.*Primer',
                           '.*Getting(\s)Started.*','.*at(\s)*a(\s)*Glance','.*Nutshell',
                           '.*FAQ.*','.*Roadmap.*','.*Cookbook.*','.*VERSUS.*','.*(\s)VS(\s)*.*',
                           '.*Comparison.*','.*Frequently(\s)*Asked(\s)*Questions.*');

  if (any { $sec_title =~ m/$_/sxm } @concepts_strings ) {
    $is_intro = q(1);
  }
  undef @concepts_strings;
  if ($is_intro ne q(1) &&
        (
        $sec_title =~ /How(\s)+.*/sxm
        && uc ($sec_title) !~ /HOW(\s)*DO(\s)*.*/sxm
        && uc ($sec_title) !~ /HOW(\s)*TO(\s)*.*/sxm
        )
  ){
    $is_intro = q(1);
  }
  if ($is_intro ne q(1)
      && $book_title =~ m/.*Concepts.*/sxm
      && $book_title !~ m/.*Concepts(\s)and(\s)Administration.*/sxm
      ) {
    $is_intro = q(1);
  }
  if ($is_intro ne q(1)
      && (
        $book_title =~ m/.*Overview.*/sxm
        || $book_title =~ m/.*New(\s)Features.*/sxm
         )
  ){
    $is_intro = q(1);
  }
  if ($is_intro eq q(0)) {return q(0);}

  return q(1);

}

##############################################################################
#
# Function: is_example
#
# Parameter: sec_title, book_name, fname_type
# Determines if the topic is of type example
# Rules:                   section_title like '%Example%'
#			or section_title like '%Sample%'
#			or section_title like '%Demo%'
#			or section_title like '%Scenario%'
#
#
##############################################################################

sub is_example {
   my $sec_title = shift;
   my $is_example = q(0);
   my @example_strings = ('.*Examples.*', '.*Sample.*', '.*Demo.*', '.*Scenario.*');

  if (any { $sec_title =~ m/$_/sxm } @example_strings ) {
    $is_example = q(1);
  }
  undef @example_strings;
  return $is_example;

}

##############################################################################
#
# Function: is_error_msg
#
# Parameter: sec_title
# Determines if the topic is of type error message
# Rules:
#                       section_title like '%Debug%'
#			or section_title like '%Diagnos%'
#			or section_title like '%Trouble%'
#			or section_title like '%Conflict%'
#			or section_title like '%Deadlock%'
#			or section_title like '%Problem%'
#			or section_title like '%Error%'
#			or section_title like '%Avoiding%'
#			or section_title like '%Failure%'
#			or section_title like '%Restriction%'
#			or (section_title like '%Limit%' and section_title not like '%Delimit%')
#			or section_title like '%Dump%'
#			or section_title like '%Emergency%'
#			or section_title like '%Test%'
#			or section_title like '%Exception%'
#			or booktitle like '%Performance%'
#			or booktitle like '%Tuning%'
#			or booktitle like '%Error%'
#			or booktitle like '%Messages%'
#
#
##############################################################################

sub is_error_message {
   my $sec_title  = shift;
   my $book_title = shift;
   my $is_error = q(0);
   my @error_strings = ('.*Debug.*', '.*Diagnos.*', '.*Trouble.*', '.*Conflict.*', '.*Deadlock.*',
                        '.*Problem.*', '.*Error.*', '.*Avoiding.*', '.*Failure.*', '.*Restriction.*',
                        '.*Dump.*', '.*Emergency.*', '.*Test.*', '.*Exception.*', '.*Performance.*',
                        '.*Tuning.*', '.*Error.*', '.*Messages.*');
   my @error_book_strings = ('.*Performance.*', '.*Tuning.*', '.*Error.*', '.*Messages.*');

  if (any { $sec_title =~ m/$_/sxm } @error_strings ) {
    $is_error = q(1);
  }
  if ($sec_title =~ m/Limit/sxm && $sec_title !~ m/Delimit/sxm) {
    $is_error = q(1);
  }
  if ((any { $book_title =~ m/$_/sxm } @error_book_strings) ) {
    $is_error = q(1);
  }
  undef @error_book_strings;
  undef @error_strings;
  return $is_error;
}

##############################################################################
#
# Function: is_faq
#
# Parameter: sec_title
# Determines if the topic is of type faq
#
##############################################################################

sub is_faq {
   my $sec_title = shift;
   my $is_faq = q(0);
   my @faq_strings = ('.*Frequently(\s)*Asked.*', '.*FAQ.*');

  if (any { $sec_title =~ m/$_/sxm } @faq_strings ) {
    $is_faq = q(1);
  }
  undef @faq_strings;
  return $is_faq;

}


##############################################################################
#
# Function: is_tasks
#
# Parameter: sec_title
# Determines if the topic is of type task
##############################################################################

sub is_tasks {

    my $sec_title = shift;
    my $is_tasks  = q(0);
    my @not_tasks_word = ('^STRING(\s).*', '^Listing _f(\s).*', '^Incoming(\s).*', '^Outgoing(\s).*', '^Operating(\s)*System.*', '^Networking.*');
    my @tasks_word = ('^HOW(\s)DO(\s).*', '^HOW(\s)TO(\s).*' ,'.*TUTORIAL.*');

    # first word with ing
    if ($sec_title =~ m/^([a-zA-Z]+)ing(\s)+(.*)/sxm) {
      $is_tasks  = q(1);
    }

    # and substr(section_title,1,1) between 'A' and 'Z'
    if ($sec_title !~ m/^[A-Z]+(.*)/sxm) {
      $is_tasks  = q(0);
    }

    if (any { $sec_title =~ m/$_/sxm } @not_tasks_word ) {
      $is_tasks = q(0);
    }
    if (any { $sec_title =~ m/$_/isxm } @tasks_word ) {
      $is_tasks = q(1);
    }
    undef @not_tasks_word;
    undef @tasks_word;
    return $is_tasks;
}

##############################################################################
#
# Function: is_references
#
# Parameter: book_title
# Determines if the topic is of type references
##############################################################################

sub is_references {
    my $booktitle = shift;
    my $sec_title = shift;
    
    my $is_reference = q(0);
    my @list_of_ref_titles = ('Concepts','Administrator'."'".'s Guide','SQL Reference',
                              'SQL Language Reference','Reference',
                              'PL/SQL User'."'".'s Guide and Reference','Performance Tuning Guide and Reference',
                              'Performance Tuning Guide','Application Developer'."'".'s Guide - Fundamentals',
                              '2 Day DBA');
    if (any { $booktitle =~ m/$_/ } @list_of_ref_titles ) {
      $is_reference = q(1);
    }
    if ($sec_title eq uc $sec_title) {
      $is_reference = q(1);
    }
    undef @list_of_ref_titles;
    return $is_reference;
}

##############################################################################
#
# Function: is_glossary
#
# Parameter: topic_class
# Determines if the topic is of type glossary
##############################################################################

sub is_glossary {
  my $topic_class = shift;
  if ($topic_class eq q(glossterm) || $topic_class eq q(glossary) || $topic_class eq q(glossentry)) {
     return q(1);
  }
  return q(0);
}

##############################################################################
#
# Function: is_chapter
#
# Parameter: topic_class
# Determines if the topic is of type chapter
##############################################################################

sub is_chapter {
  my $topic_class = shift;
  if (lc $topic_class eq q(chapter) ) {
     return q(1);
  }
  return q(0);
}

##############################################################################
#
# Function: roles
#
# Parameter: path, book_title
# Determines if the topic is of type glossary
##############################################################################

sub roles {
  my $pathv      = shift;
  my $book_title = shift;
  my $type_pn    = shift;
  my $type_path  = shift;
  my $roles = q();

  if ($book_title =~ m/user/sxmi) {
     $roles .= ",user";
  }
  elsif (any { $type_pn =~ m/$_/sxmi } @user_pns ) {
     $roles .= ",user";
  }

  if (   $pathv =~ m/.*appdev[.].*/sxm
      || $pathv =~ m/.*dev[.].*/sxm
      || $pathv =~ m/.*java[.].*/sxm
      || $book_title =~ m/.*Developer's.*/sxm
      || $book_title =~ m/.*(\s)+API(\s)+.*/sxm
      || $book_title =~ m/.*Javadoc.*/sxm

  ) {
     $roles .= ",developer";
  }
  elsif (any { $type_pn =~ m/$_/sxmi } @developer_pns ) {
     $roles .= ",developer";
  }


  if (   $pathv =~ m/.*server[.].*/sxm
      || $pathv =~ m/.*em[.].*/sxm
      || $pathv =~ m/.*admin[.].*/sxm
      || $book_title =~ m/.*Administrat.*/sxm
      || $book_title =~ m/.*2(\s)+Day(\s)+DBA.*/sxm

  ) {
     $roles .= ",administrator";
  }
  elsif (any { $type_pn =~ m/$_/sxmi } @administrator_pns ) {
     $roles .= ",administrator";
  }

  $roles =~ s/^,//sxm;
  return $roles;
}

##############################################################################
#
# Function: get_targets
#
# Parameter: file name, target, text first line, path of chunk, ccl and ch
# Determines the targets
##############################################################################

sub get_targets {

   my $filebeingscanned = shift;
   my $targetid = shift;
   my $texttoscan = shift;
   my $beginingline = shift;
   my $cpath = shift;
   my $ccl = shift;
   my $ch = shift;
   $texttoscan = $beginingline.$texttoscan;
   my $subtext = &scanify ($texttoscan);
   my $targetpath = $filebeingscanned."#".$targetid;
   if(exists $targets{$targetpath}){
   }
   else{
       $targets{$targetpath}{file} = $filebeingscanned;
       $targets{$targetpath}{target} = $targetid;
       $targets{$targetpath}{title} = $subtext;
       $targets{$targetpath}{equivalence} = join (" ", @allids);
       $targets{$targetpath}{path} = $cpath;
       $targets{$targetpath}{class} = $ccl;
       $targets{$targetpath}{head} = $ch;
   }

}

sub elastic_search_mapping {
    my $mapping_type = shift;
    $mapping_type =~ s/_(.*)//sxm;
    $mapping_type =~ s/-(.*)//sxm;
    if ( $mapping_type =~ m/([[:alpha:]](\d){5}([_-](\d){1,2})*)\//sxm ) {
       $mapping_type = lc $mapping_type;
    }
    if (exists $index_type_mappings{$mapping_type}) {
      return;
    }
    my $elastic_search_mapping = <<"TOP";
curl -XPUT 'http://${server_elastic_search}:9200/${context}/${mapping_type}/_mapping' -d '
{
    "${mapping_type}" : {
        "properties" : {
            "id" : {"type" : "string", "index" : "not_analyzed"},
            "title" : {"type" : "string"},
            "summary" : {"type" : "string"},
            "book_title" : {"type" : "string"},
            "part_number" : {"type" : "string", "index" : "not_analyzed"},
            "book_release" : {"type" : "string"},
            "ul" : {"type" : "string"},
            "book_url" : {"type" : "string"},
            "topic_type" : {"type" : "string"},
            "roles" : {"type" : "string"},
            "lang" : {"type" : "string"},
            "page_title" : {"type" : "string"},
            "release_date" : {"type" : "date", "format" : "yyyy-mm-dd" },
            "formats" : {"type" : "string"},
            "libraryurl" : {"type" : "string", "index" : "not_analyzed"},
            "url" : {"type" : "string", "index" : "not_analyzed"},
            "description" : {"type" : "string"},
            "mainpn" : {"type" : "string", "index" : "not_analyzed"},
            "librarypn" : {"type" : "string", "index" : "not_analyzed"},
            "category_name" : {"type" : "string", "index" : "not_analyzed"},
            "release_number" : {"type" : "string", "index" : "not_analyzed"},
            "computedtype" : {"type" : "string"},
            "keywords" : {"type" : "string"}
        }
    }
}
'
TOP
    $index_type_mappings{$mapping_type} = q();

    open my $CHUNK, '>>', $jsonoutputfolder . q(/) . $library_partnumber . q(-mapping.txt)
        or croak "Couldnt write to json output mapping file: $OS_ERROR\n";
    binmode $CHUNK, ':encoding(UTF-8)';
    print {$CHUNK} $elastic_search_mapping . "\n\n";
    print {$CHUNK} q(VIVEK-SAMPLE) . "\n\n";
    close $CHUNK
        or carp "Cant close: $OS_ERROR";

  return;

}

sub elastic_search_mapping_old {
    my $mapping_type = shift;
    $mapping_type =~ s/_(.*)//sxm;
    $mapping_type =~ s/-(.*)//sxm;
    if ( $mapping_type =~ m/([[:alpha:]](\d){5}([_-](\d){1,2})*)\//sxm ) {
       $mapping_type = lc $mapping_type;
    }
    if (exists $index_type_mappings{$mapping_type}) {
      return;
    }
    my $elastic_search_mapping = <<"TOP";
curl -XPUT 'http://${server_elastic_search}:9200/${context}/${mapping_type}/_mapping' -d '
{
    "${mapping_type}" : {
        "properties" : {
            "id" : {"type" : "string", "store" : true },
            "title" : {"type" : "string", "store" : true },
            "summary" : {"type" : "string", "store" : true },
            "book_title" : {"type" : "string", "store" : true },
            "part_number" : {"type" : "string", "store" : true },
            "book_release" : {"type" : "string", "store" : true },
            "ul" : {"type" : "string", "store" : true },
            "book_url" : {"type" : "string", "store" : true },
            "topic_type" : {"type" : "string", "store" : true },
            "roles" : {"type" : "string", "store" : true },
            "lang" : {"type" : "string", "store" : true },
            "page_title" : {"type" : "string", "store" : true },
            "release_date" : {"type" : "date", "format" : "yyyy-mm-dd" },
            "url" : {"type" : "string", "store" : true }
        }
    }
}
'
TOP
    $index_type_mappings{$mapping_type} = q();

    open my $CHUNK, '>>', $jsonoutputfolder . q(/) . $library_partnumber . q(-mapping.txt)
        or croak "Couldnt write to json output mapping file: $OS_ERROR\n";
    binmode $CHUNK, ':encoding(UTF-8)';
    print {$CHUNK} $elastic_search_mapping . "\n\n";
    print {$CHUNK} q(VIVEK-SAMPLE) . "\n\n";
    close $CHUNK
        or carp "Cant close: $OS_ERROR";

  return;

}


sub get_correct_date_format {
  my $raw_date = shift;
  $raw_date =~ s/\s(.*)//gis;
  $raw_date =~ s/T(.*)//gis;
  my $reldate = "";
  if ($raw_date =~ m/[0-9]{4}[-|\/][0-9]{1,2}[-|\/][0-9]{1,2}/) {
    $raw_date =~ s/\//-/gis;
    my @date_parts = split (/-/, $raw_date);
    if (scalar @date_parts == 3) {
        if (length $date_parts[1] < 2) {
           # its a month
           $date_parts[1] = "0".$date_parts[1];
        }
        if (length $date_parts[2] < 2) {
           # its a month
           $date_parts[2] = "0".$date_parts[2];
        }
        $reldate = $date_parts[0]."-".$date_parts[1]."-".$date_parts[2];
    }
  }
  return $reldate;
}

sub snapshot_time {
   (my $sec, my $min, my $hour, my $mday, my $mon, my $year, my $wday, my $yday, my $isdst) = gmtime();
   $mon=$mon+1; $year = 1900+$year;
   if($sec<10){
      $sec = "0".$sec;
   }
   if($min<10){
      $min = "0".$min;
   }
   if($hour<10){
      $hour = "0".$hour;
   }
   if($mon<10){
      $mon = "0".$mon;
   }

   my $todays_date = "DATE_".$year."_".$mon."_".$mday."_TIME_".$hour."_".$min."_".$sec;
   return($todays_date);

}

sub calculate_path {
   my $folderfrom = shift;
   my $fileto = shift;
   my $pcomputed = File::Spec->catfile($folderfrom, $fileto);
   if (-f $pcomputed) {
     $pcomputed = abs_path($pcomputed);
     $pcomputed = File::Spec->abs2rel($pcomputed);
   }
   else{
      $pcomputed = q();
   }
   return $pcomputed;
}

sub tidyhtml {
    my $rawfile = shift;
    if (-f "c:\\dev5\\src\\tools\\tidy.exe") {
      system ("c:\\dev5\\src\\tools\\tidy.exe -q  -raw -o ".$rawfile." -wrap 0 -ashtml -quiet --show-warnings no --show-errors 0 --preserve-entities y --tidy-mark 0 ".$rawfile);
    }
    else {
      system ("/usr/bin/tidy -q -raw -o ".$rawfile." -wrap 0 -ashtml -quiet --show-warnings no --show-errors 0 --preserve-entities y --tidy-mark 0 ".$rawfile);
    }
    return ;
}

sub encode_entities_html {
   my $string = shift;
   #return HTML::Entities::encode_entities($string);
   return $string;
}

sub getFolderMappings {
        my $mainPN = shift;
        $mainPN =~ s/_/-/gis;
        $mainPN = uc ($mainPN);
        my $url = 'http://pdb-stage.us.oracle.com/subFolderMappings.jsp?pn='.$mainPN;
        my $summaries4perl = '';

        my $ua = new LWP::UserAgent();
        my $resp = $ua->get($url);
        if ( $resp->is_success() ) {
            $summaries4perl = $resp->content();
        }

        $summaries4perl = Encode::decode('UTF-8', $summaries4perl);
        eval $summaries4perl;
        if ( $@ ) {
            print $@;
            %subfolder_mappings = ();
        }
}

sub standardize_release_number {
   my $rn = shift;
   if (length $rn < 1) {
     $rn = "0";
   }
   my $numberOfDotPlaces = 5;
   $rn =~ s/^[.]//gis;
   $rn =~ s/[.]$//gis;

   my @parts = split(/\./, $rn);
   for (my $dots = 0; $dots < ($numberOfDotPlaces - scalar @parts); $dots++) {
     $rn .= ".0";
   }
   return $rn;
}

__END__

=head1 NAME

analyzer_elastic_json.pl - Generates the JSON data file

=head1 USAGE

 analyzer_elastic_json.pl -f ./ -l E11882-091 --ctx db112

=head1 DESCRIPTION

Scans a directory of HTML, writes the JSON data file for elastic search.

=head1 REQUIRED ARGUMENTS

none

=head1 OPTIONS

=over 4

=item -f --folder

The directory where the HTML files for which the JSOPN data
is to be created is present.

=item -l --lib

The library part number

=item -c --ctx

The library context value.

=item -p --prod

The product shortname using which the index will be created.

=back

=head1 EXAMPLES

  analyzer_elastic_json.pl -f ./ -l E11882-091 --ctx db112
 

=head1 DIAGNOSTICS

n/a

=head1 EXIT STATUS

n/a

=head1 CONFIGURATION

n/a

=head1 DEPENDENCIES

n/a

=head1 INCOMPATIBILITIES

n/a

=head1 BUGS AND LIMITATIONS

n/a

=head1 AUTHOR

Vivek Kumar <vivek.kumar@oracle.com>

=head1 LICENSE AND COPYRIGHT

Copyright 2014, Oracle and/or its affiliates. All rights reserved.
